// (C) Copyright Cryolite 2004, 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This header provides some algorithms missed in the C++ standard, such as
// `copy_if'.
//
// TODO: Implement some algorithm defined in SGI STL but not in the C++
//       standard.
// TODO: Reimplement STL algorithms which take a random number generator (RNG).
//       They should adopt the new RNG concepts proposed in TR1.

#if !defined(CRY_ALGORITHM_HPP_INCLUDE_GUARD)
#define CRY_ALGORITHM_HPP_INCLUDE_GUARD

#define CRY_ALGORITHM_ENABLE_SIG

#include <functional>

#include <boost/config.hpp>

#include <boost/utility/result_of.hpp>

#include <boost/iterator/iterator_categories.hpp>
#include <boost/iterator/iterator_traits.hpp>

#include <boost/random/uniform_smallint.hpp>
#include <boost/random/uniform_01.hpp>

#include <cry/functional.hpp>
#include <cry/random/crand.hpp>
#include <cry/lambda/enable_sig.hpp>

#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable:4675)
#endif


namespace cry{

////////////////////////////////////////////////////////////////////////////////
// Iterator-oriented algorithms missed in the C++ standard.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// copy_if
////////////////////////////////////////////////////////////////////////////////

template<
  class SinglePassReadableIterator
, class IncrementableWritableIterator, class UnaryPredicate
>
inline
IncrementableWritableIterator
copy_if(
  SinglePassReadableIterator first, SinglePassReadableIterator last
, IncrementableWritableIterator dest, UnaryPredicate pred
)
{
  for(; first != last; ++first)
  {
    if(pred(*first))
    {
      *dest++ = *first;
    }
  }
  return dest;
}



////////////////////////////////////////////////////////////////////////////////
// is_sorted
////////////////////////////////////////////////////////////////////////////////

namespace algorithm_detail{

template<class SinglePassTraversalReadableIterator, class StrictWeakOrdering>
bool is_sorted_dispatch(
  SinglePassTraversalReadableIterator first
, SinglePassTraversalReadableIterator last
, StrictWeakOrdering comp
, boost::single_pass_traversal_tag
)
{
  if(first == last) return true;

  // SinglePassTraversalReadableIterator cannot be derefered twice.
  //
  // It should be assumed that `value_type' meets CopyConstructible and
  // Assignable rather than DefaultConstructible and Assignable. See
  // http://www.open-std.org/jtc1/sc22/wg21/docs/lwg-defects.html#227
  typedef typename boost::iterator_value<
      SinglePassTraversalReadableIterator
    >::type value_type;
  value_type prev_value(*first);
  for(++first; first != last; ++first){
    value_type value(*first);
    if(comp(value, prev_value)){
      return false;
    }
    prev_value = value;
  }
  return true;
}

template<class ForwardTraversalReadableIterator, class StrictWeakOrdering>
bool is_sorted_dispatch(
  ForwardTraversalReadableIterator first
, ForwardTraversalReadableIterator last
, StrictWeakOrdering comp
, boost::forward_traversal_tag
)
{
  if(first == last) return true;
	for (ForwardTraversalReadableIterator prev(first++);
       first != last;
       ++first, ++prev)
  {
    if(comp(*first, *prev)) return false;
  }
	return true;
}

} // namespace algorithm_detail

template<class SinglePassTraversalReadableIterator, class StrictWeakOrdering>
bool is_sorted(
  SinglePassTraversalReadableIterator first
, SinglePassTraversalReadableIterator last
, StrictWeakOrdering comp
)
{
  typedef typename boost::iterator_traversal<
      SinglePassTraversalReadableIterator
    >::type traversal_tag;
  return algorithm_detail::is_sorted_dispatch(
      first, last, comp, traversal_tag()
    );
}

template<class SinglePassReadableIterator>
bool is_sorted(
  SinglePassReadableIterator first, SinglePassReadableIterator last
)
{
  typedef typename boost::iterator_value<SinglePassReadableIterator>::type
    value_type;
  return is_sorted(first, last, cry::less());
}



////////////////////////////////////////////////////////////////////////////////
// random_shuffle
////////////////////////////////////////////////////////////////////////////////

template<class RandomAccessIterator, class UniformRNG>
void algorithm_random_shuffle(
  RandomAccessIterator first, RandomAccessIterator last, UniformRNG &rng
)
{
  // See the section 3.4.2 of D. E. Knuth, "The Art of Computer Programming.
  // Volume 2: Seminumerical Algorithms," Third Edition. Addison-Wesley, 1998.
  using std::swap;
  typedef typename boost::iterator_difference<RandomAccessIterator>::type
    difference_type;
  typedef boost::uniform_smallint<difference_type> dist_type;
  
  difference_type n = last - first;
  for(difference_type i = n - 1; i > 0; --i){
    dist_type dist(0, i);
    swap(*(first + dist(rng)), *(first + i));
  }
}

struct random_shuffle_
  : public cry::lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<class RandomIteratorFirst, class RandomIteratorLast>
  struct result<
    random_shuffle_ const(RandomIteratorFirst, RandomIteratorLast)
  >
  {
    typedef void type;
  };

  template<class RandomIteratorFirst, class RandomIteratorLast>
  struct result< random_shuffle_(RandomIteratorFirst, RandomIteratorLast) >
  {
    typedef void type;
  };
  
  template<
    class RandomIteratorFirst, class RandomIteratorLast, class UniformRNG
  >
  struct result<
    random_shuffle_ const(RandomIteratorFirst, RandomIteratorLast, UniformRNG)
  >
  {
    typedef void type;
  };

  template<
    class RandomIteratorFirst, class RandomIteratorLast, class UniformRNG
  >
  struct result<
    random_shuffle_(RandomIteratorFirst, RandomIteratorLast, UniformRNG)
  >
  {
    typedef void type;
  };



  // operator()
  template<class RandomAccessIterator>
  void operator()(RandomAccessIterator first, RandomAccessIterator last) const
  {
    crand rng;
    algorithm_random_shuffle(first, last, rng);
  }
  
  template<class RandomAccessIterator, class UniformRNG>
  typename boost::result_of<
    random_shuffle_ const(
      RandomAccessIterator, RandomAccessIterator, UniformRNG &
    )
  >::type
  operator()(
    RandomAccessIterator first, RandomAccessIterator last, UniformRNG &rng
  ) const
  {
    algorithm_random_shuffle(first, last, rng);
  }
}; // struct random_shuffle_

random_shuffle_ const random_shuffle = random_shuffle_();



////////////////////////////////////////////////////////////////////////////////
// random_sample
////////////////////////////////////////////////////////////////////////////////

// In the description of SGI STL, the input is required to be single-pass,
// but in the Knuth's book seems to require it to be forward-traversal.

// ADL hook point.
template<class InputIterator, class RandomAccessIterator, class UniformRNG>
void algorithm_random_sample(
  InputIterator ifirst, InputIterator ilast
, RandomAccessIterator ofirst, RandomAccessIterator olast, UniformRNG &rng
)
{
  // See the section 3.4.2 of D. E. Knuth, "The Art of Computer Programming.
  // Volume 2: Seminumerical Algorithms," Third Edition. Addison-Wesley, 1998.
  typedef typename boost::iterator_difference<InputIterator>::type
    i_diff_t;
  typedef typename boost::iterator_difference<RandomAccessIterator>::type
    o_diff_t;
  typedef boost::uniform_01<UniformRNG, double> dist_type;

  i_diff_t m = std::distance(ifirst, ilast);
  o_diff_t n = olast - ofirst;
  i_diff_t i = 0;
  o_diff_t j = 0;
  dist_type dist(rng);
  while(j < n){
    if((m - i) * dist() >= n - j){
      ++ifirst;
      ++i;
    }
    else{
      *ofirst++ = *ifirst++;
      ++i;
      ++j;
    }
  }
}

struct random_sample_
  : public cry::lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<
    class InputIteratorFirst, class InputIteratorLast
  , class RandomIteratorFirst, class RandomIteratorLast, class UniformRNG
  >
  struct result<
    random_sample_ const(
      InputIteratorFirst, InputIteratorLast
    , RandomIteratorFirst, RandomIteratorLast, UniformRNG
    )
  >
  {
    typedef void type;
  };

  template<
    class InputIteratorFirst, class InputIteratorLast
  , class RandomIteratorFirst, class RandomIteratorLast, class UniformRNG
  >
  struct result<
    random_sample_(
      InputIteratorFirst, InputIteratorLast
    , RandomIteratorFirst, RandomIteratorLast, UniformRNG
    )
  >
    : public result<
        random_sample_ const(
          InputIteratorFirst, InputIteratorLast
        , RandomIteratorFirst, RandomIteratorLast, UniformRNG
        )
      >
  {};



  // operator()
  template<class InputIterator, class RandomAccessIterator, class UniformRNG>
  typename boost::result_of<
    random_sample_ const(InputIterator, RandomAccessIterator, UniformRNG &)
  >::type
  operator()(
    InputIterator ifirst, InputIterator ilast
  , RandomAccessIterator ofirst, RandomAccessIterator olast, UniformRNG &rng
  ) const
  {
    algorithm_random_sample(ifirst, ilast, ofirst, olast, rng);
  }
}; // struct random_sample_

random_sample_ const random_sample = random_sample_();



////////////////////////////////////////////////////////////////////////////////
// random_sample_n
////////////////////////////////////////////////////////////////////////////////

// ADL hook point.
template<
  class ForwardIterator, class OutputIterator, class Distance, class UniformRNG
>
void algorithm_random_sample_n(
  ForwardIterator first, ForwardIterator last, OutputIterator dest, Distance n
, UniformRNG &rng
)
{
  typedef typename boost::iterator_difference<ForwardIterator>::type diff_t;
  typedef boost::uniform_01<UniformRNG, double> dist_t;

  diff_t m = std::distance(first, last);
  dist_t dist(rng);
  diff_t i = 0;
  Distance j = 0;
  while(j < n){
    if((m - i) * dist() >= n - j){
      ++first;
      ++i;
    }
    else{
      *dest++ = *first++;
      ++i;
      ++j;
    }
  }
}

struct random_sample_n_
  : public cry::lambda::enable_sig
{
  // Result type calculation
  template<class FArgs>
  struct result;

  template<
    class ForwardIteratorFirst, class ForwardIteratorLast
  , class OutputIterator, class Distance, class UniformRNG
  >
  struct result<
    random_sample_n_ const(
      ForwardIteratorFirst, ForwardIteratorLast
    , OutputIterator, Distance, UniformRNG
    )
  >
  {
    typedef void type;
  };

  template<
    class ForwardIteratorFirst, class ForwardIteratorLast
  , class OutputIterator, class Distance, class UniformRNG
  >
  struct result<
    random_sample_n_(
      ForwardIteratorFirst, ForwardIteratorLast
    , OutputIterator, Distance, UniformRNG
    )
  >
    : public result<
        random_sample_n_ const(
          ForwardIteratorFirst, ForwardIteratorLast
        , OutputIterator, Distance, UniformRNG
        )
      >
  {};



  // operator()
  template<
    class ForwardIterator, class OutputIterator, class Distance
  , class UniformRNG
  >
  typename boost::result_of<
    random_sample_n_ const(
      ForwardIterator, ForwardIterator, OutputIterator, Distance, UniformRNG &
    )
  >::type
  operator()(
    ForwardIterator first, ForwardIterator last, OutputIterator dest, Distance n
  , UniformRNG &rng
  ) const
  {
    algorithm_random_sample_n(first, last, dest, n, rng);
  }
}; // struct random_sample_n_

random_sample_n_ const random_sample_n = random_sample_n_();

} // namespace cry


#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // !defiend(CRY_ALGORITHM_HPP_INCLUDE_GUARD)
