// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_CONTAINER_CONTAINER_TRAITS_HPP_INCLUDE_GUARD)
#define CRY_CONTAINER_CONTAINER_TRAITS_HPP_INCLUDE_GUARD

#include <boost/config.hpp>
#include <boost/iterator/iterator_categories.hpp>


namespace std{

template<class T, class Allocator>
class vector;

template<class T, class Allocator>
class deque;

template<class T, class Allocator>
class list;

#if defined(BOOST_HAS_SLIST)
template<class T, class Allocator>
class slist;
#endif

template<class T, class Compare, class Allocator>
class set;

template<class T, class Compare, class Allocator>
class multiset;

template<class Key, class T, class Compare, class Allocator>
class map;

template<class Key, class T, class Compare, class Allocator>
class multimap;

} // namespace std



namespace cry{namespace container{

struct container_tag
{ };

struct sequential_tag
  : virtual public container_tag
{ };

struct back_insertable_sequential_tag
  : virtual public sequential_tag
{ };

struct front_insertable_sequential_tag
  : virtual public sequential_tag
{ };

struct front_and_back_insertable_sequential_tag
  : virtual public front_insertable_sequential_tag
  , virtual public back_insertable_sequential_tag
{ };

struct associative_tag
  : virtual public container_tag
{ };

struct sorted_associative_tag
  : virtual public associative_tag
{ };

struct hashed_associative_tag
  : virtual public associative_tag
{ };

struct unique_key_tag
{ };

struct multiple_key_tag
{ };

struct simple_value_form_tag
{ };

struct pair_value_form_tag
{ };

struct stable_iterator_tag
{ };

struct unstable_iterator_tag
{ };



// Default `container_traits'.
template<class Container = void>
struct container_traits
{
  typedef boost::mpl::false_ is_specialized;
  typedef void               container_category;
  typedef void               traversal_tag;
  typedef void               key_uniqueness;
  typedef void               value_form;
  typedef void               iterator_stability;
  typedef boost::mpl::false_ has_reserve;
  typedef boost::mpl::false_ has_capacity;
  typedef boost::mpl::false_ has_splice;
  typedef boost::mpl::false_ has_remove;
  typedef boost::mpl::false_ has_unique;
  typedef boost::mpl::false_ has_merge;
  typedef boost::mpl::false_ has_reverse;
  typedef boost::mpl::false_ has_sort;
};

// specializatin for `std::vector'.
template<class T, class Allocator>
struct container_traits<std::vector<T, Allocator> >
  : public container_traits<>
{
  typedef back_insertable_sequential_tag     container_category;
  typedef boost::random_access_traversal_tag traversal_tag;
  typedef unstable_iterator_tag              iterator_stability;
  typedef boost::mpl::true_                  has_reserve;
  typedef boost::mpl::true_                  has_capacity;
};

// specialization for std::deque
template<class T, class Allocator>
struct container_traits<std::deque<T, Allocator> >
  : public container_traits<>
{
  typedef front_and_back_insertable_sequential_tag container_category;
  typedef boost::random_access_traversal_tag       traversal_tag;
  typedef unstable_iterator_tag                    iterator_stabiliity;
};

// specialization for std::list
template<class T, class Allocator>
struct container_traits<std::list<T, Allocator> >
  : public container_traits<>
{
  typedef front_and_back_insertable_sequential_tag container_category;
  typedef boost::bidirectional_traversal_tag       traversal_tag;
  typedef stable_iterator_tag                      iterator_stability;
  typedef boost::mpl::true_                        has_splice;
  typedef boost::mpl::true_                        has_remove;
  typedef boost::mpl::true_                        has_unique;
  typedef boost::mpl::true_                        has_merge;
  typedef boost::mpl::true_                        has_reverse;
  typedef boost::mpl::true_                        has_sort;
};

#if defined(BOOST_HAS_SLIST)
// specialization for std::slist
template<class T, class Allocator>
struct container_traits<std::slist<T, Allocator> >
  : public container_traits<>
{
  typedef front_insertable_sequentail_tag container_category;
  typedef boost::forward_traversal_tag    traversal_tag;
  typedef stable_iterator_tag             iterator_stability;
  typedef boost::mpl::true_               has_splice;
  typedef boost::mpl::true_               has_remove;
  typedef boost::mpl::true_               has_unique;
  typedef boost::mpl::true_               has_merge;
  typedef boost::mpl::true_               has_reverse;
  typedef boost::mpl::true_               has_sort;
};
#endif // defined(BOOST_HAS_SLIST)

// specialization for std::set
template<class T, class Compare, class Allocator>
struct container_traits<std::set<T, Compare, Allocator> >
  : public container_traits<>
{
  typedef sorted_associative_tag             container_category;
  typedef boost::bidirectional_traversal_tag traversal_tag;
  typedef unique_key_tag                     key_uniqueness;
  typedef simple_value_form_tag              value_form;
  typedef stable_iterator_tag                iterator_stability;
};

// specialization for std::multiset
template<class T, class Compare, class Allocator>
struct container_traits<std::multiset<T, Compare, Allocator> >
  : public container_traits<>
{
  typedef sorted_associative_tag             container_category;
  typedef boost::bidirectional_traversal_tag traversal_tag;
  typedef multiple_key_tag                   key_uniqueness;
  typedef simple_value_form_tag              value_form;
  typedef stable_iterator_tag                iterator_stability;
};

// specialization for std::map
template<class Key, class T, class Compare, class Allocator>
struct container_traits<std::map<Key, T, Compare, Allocator> >
  : public container_traits<>
{
  typedef sorted_associative_tag             container_category;
  typedef boost::bidirectional_traversal_tag traversal_tag;
  typedef unique_key_tag                     key_uniqueness;
  typedef pair_value_form_tag                value_form;
  typedef stable_iterator_tag                iterator_stability;
};

// specialization for std::multimap
template<class Key, class T, class Compare, class Allocator>
struct container_traits<std::multimap<Key, T, Compare, Allocator> >
  : public container_traits<>
{
  typedef sorted_associative_tag             container_category;
  typedef boost::bidirectional_traversal_tag traversal_tag;
  typedef multiple_key_tag                   key_uniqueness;
  typedef pair_value_form_tag                value_form;
  typedef stable_iterator_tag                iterator_stability;
};

// Meta-functions with mpl-compatible form.
template<class Container>
struct category
{
  typedef typename container_traits<Container>::container_category type;
};

template<class Container>
struct traversal
{
  typedef typename container_traits<Container>::traversal_tag type;
};

template<class Container>
struct key_uniqueness
{
  typedef typename container_traits<Container>::key_uniqueness type;
};

template<class Container>
struct value_form
{
  typedef typename container_traits<Container>::value_form type;
};

template<class Container>
struct iterator_stability
{
  typedef typename container_traits<Container>::iterator_stability type;
};

template<class Container>
struct has_reserve
{
  typedef typename container_traits<Container>::has_reserve type;
};

template<class Container>
struct has_capacity
{
  typedef typename container_traits<Container>::has_capacity type;
};

template<class Container>
struct has_splice
{
  typedef typename container_traits<Container>::has_splice type;
};

template<class Container>
struct has_remove
{
  typedef typename container_traits<Container>::has_remove type;
};

template<class Container>
struct has_unique
{
  typedef typename container_traits<Container>::has_unique type;
};

template<class Container>
struct has_merge
{
  typedef typename container_traits<Container>::has_merge type;
};

template<class Container>
struct has_reverse
{
  typedef typename container_traits<Container>::has_reverse type;
};

template<class Container>
struct has_sort
{
  typedef typename container_traits<Container>::has_sort type;
};

}} // namespace cry::container

#endif // !defined(CRY_CONTAINER_CONTAINER_TRAITS_HPP_INCLUDE_GUARD)
