// (C) Jonathan Turkanis.
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt.)

// Contains the definition of move_ptrs::default_deleter, the default
// Deleter template argument to move_ptr. Uses a technique of Daniel
// Wallin to capture the type of a pointer at the time the deleter 
// is constructed, so that move_ptrs can delete objects of incomplete 
// type by default.

#if !defined(CRY_DELETER_DEFAULT_DELETER_HPP_INCLUDE_GUARD)
#define CRY_DELETER_DEFAULT_DELETER_HPP_INCLUDE_GUARD

#include <boost/mpl/if.hpp>
#include <boost/type_traits/is_array.hpp>
#include <boost/type_traits/remove_bounds.hpp>
#include <boost/checked_delete.hpp>


namespace cry{

namespace deleter_detail{

template<typename T>
struct deleter_base{
  typedef void (*deleter)(T*);
  deleter_base(deleter d) { delete_ = d; }
  void operator() (T *p) const { delete_(p); }
  static deleter delete_;
};

template<class T>
typename deleter_base<T>::deleter 
deleter_base<T>::delete_;

template<typename T>
struct scalar_deleter : deleter_base<T>{
  typedef deleter_base<T> base;
  scalar_deleter() : base(do_delete) { }
  static void do_delete(T *p) { boost::checked_delete(p); }
};

template<typename T>
struct array_deleter 
  : deleter_base<typename boost::remove_bounds<T>::type>
{
  typedef typename boost::remove_bounds<T>::type element_type;
  typedef deleter_base<element_type> base;
  array_deleter() : base(do_delete) { }
  static void do_delete(element_type *p) { boost::checked_array_delete(p); }
};

} // namespace deleter_detail

template<typename T>
struct default_deleter
  : boost::mpl::if_<
      boost::is_array<T>,
      deleter_detail::array_deleter<T>,
      deleter_detail::scalar_deleter<T>
    >::type
{
  default_deleter() { }
  template<typename AnotherDeleter>
  default_deleter(default_deleter<AnotherDeleter>) { }
};

} // namespace cry


#endif // !defined(CRY_DELETER_DEFAULT_DELETER_HPP_INCLUDED)
