// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//  This header provides replacements for standard functors. The functors
// defined here have some differences from standard version.
//  Taking a look at `boost::result_of' or TR1, `result_of' can be used to
// deduct the return type of functors, even if their `operator()' are overloaded
// or templatized. So, It does not seem that standard functors need to be
// templates any more. Instead, this header defines functors as non-template
// struct, and they have `operator()' member templates.
//  This header also implements result type calculation for fundamental types.
// Result types of all built-in operations can be deduced. If you want to extend
// type deduction ability of these functors for user-defined classes, you can do
// it by specializations of `result' member struct templates of each functor, or
// preparing `boost::typeof' for the classes and define
// `CRY_FUNCTIONAL_USE_TYPEOF' before including this header.

// TODO: Work-around for typeof with VC++7.1.

#if !defined(CRY_FUNCTIONAL_HPP_INCLUDE_GUARD)
#define CRY_FUNCTIONAL_HPP_INCLUDE_GUARD

#define CRY_FUNCTIONAL_ENABLE_SIG

#include <cstddef>

#include <boost/config.hpp>

#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
#include <boost/typeof/typeof.hpp>
#endif

#include <boost/mpl/and.hpp>
#include <boost/mpl/or.hpp>

#include <boost/type_traits/is_pointer.hpp>
#include <boost/type_traits/is_array.hpp>
#include <boost/type_traits/is_fundamental.hpp>
#include <boost/type_traits/add_pointer.hpp>
#include <boost/type_traits/add_reference.hpp>
#include <boost/type_traits/remove_pointer.hpp>

#include <boost/utility/enable_if.hpp>
#include <boost/utility/result_of.hpp>

#include <cry/type_traits/add_const_reference.hpp>
#include <cry/type_traits/plain.hpp>
#include <cry/lambda/enable_sig.hpp>


namespace cry{

namespace functional_detail{

template<std::size_t code>
struct decode_fundamental_type_code;

#define CRY_REGISTER_FUNDAMENTAL_TYPE(TYPE, CODE)      \
char (&descriminate_fundamental_type_impl(TYPE))[CODE]; \
                                                        \
template<> struct decode_fundamental_type_code<CODE>    \
{                                                       \
  typedef TYPE type;                                    \
};

CRY_REGISTER_FUNDAMENTAL_TYPE(char              , 1)
CRY_REGISTER_FUNDAMENTAL_TYPE(signed char       , 2)
CRY_REGISTER_FUNDAMENTAL_TYPE(unsigned char     , 3)
CRY_REGISTER_FUNDAMENTAL_TYPE(short int         , 4)
CRY_REGISTER_FUNDAMENTAL_TYPE(unsigned short int, 5)
CRY_REGISTER_FUNDAMENTAL_TYPE(int               , 6)
CRY_REGISTER_FUNDAMENTAL_TYPE(unsigned int      , 7)
CRY_REGISTER_FUNDAMENTAL_TYPE(long int          , 8)
CRY_REGISTER_FUNDAMENTAL_TYPE(unsigned long int , 9)
CRY_REGISTER_FUNDAMENTAL_TYPE(float             , 10)
CRY_REGISTER_FUNDAMENTAL_TYPE(double            , 11)
CRY_REGISTER_FUNDAMENTAL_TYPE(long double       , 12)
CRY_REGISTER_FUNDAMENTAL_TYPE(bool              , 13)
#if !defined(BOOST_NO_INTRINSIC_WCHAR_T)
CRY_REGISTER_FUNDAMENTAL_TYPE(wchar_t           , 14)
#endif

#define CRY_DEDUCE_FUNDAMENTAL_TYPE(EXPR)                              \
::cry::functional_detail::decode_fundamental_type_code<                \
  sizeof(                                                               \
    ::cry::functional_detail::descriminate_fundamental_type_impl(EXPR) \
  )                                                                     \
>::type



// From (5.2.1)
////////// subscription //////////
template<class LHS, class RHS, class = void>
struct subscription_result
#if defined(CRY_FUNCTIONNAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs()[rhs()] ) type;
}
#endif
;

template<class LHS, class RHS>
struct subscription_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::or_<
      boost::is_array<typename plain<LHS>::type>
    , boost::is_pointer<typename plain<LHS>::type>
    >
  >::type
>
  : public boost::add_reference<typename plain<LHS>::type>
{};



// From (5.2.6).
////////// post-increment //////////
template<class T, class = void>
struct post_increment_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( (x())++ ) type;
}
#endif
;

template<class T>
struct post_increment_result<
  T
, typename boost::enable_if<
    boost::mpl::or_<
      boost::is_fundamental<typename plain<T>::type>
    , boost::is_pointer<typename plain<T>::type>
    >
  >::type
>
  : public plain<T>
{};



////////// post-decrement //////////
template<class T, class = void>
struct post_decrement_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( (x())-- ) type;
}
#endif
;

template<class T>
struct post_decrement_result<
  T
, typename boost::enable_if<
    boost::mpl::or_<
      boost::is_fundamental<typename plain<T>::type>
    , boost::is_pointer<typename plain<T>::type>
    >
  >::type
>
  : public plain<T>
{};



// From (5.3.1/1).
////////// indirection //////////
template<class T, class = void>
struct indirection_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( *(x()) ) type;
}
#endif
;

template<class T>
struct indirection_result<
  T
, typename boost::enable_if<
    boost::is_pointer<typename plain<T>::type>
  >::type
>
  : public boost::add_reference<
      typename boost::remove_pointer<typename plain<T>::type>::type
    >
{};



// From (5.3.1/2).
////////// address-of //////////
template<class T>
struct address_of_result
  : public boost::add_pointer<T>
{};



// From (5.3.1/6).
////////// unary plus //////////
template<class T, class = void>
struct unary_plus_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( +(x()) ) type;
}
#endif
;

template<class T>
struct unary_plus_result<
  T
, typename boost::enable_if<
    boost::is_fundamental<typename plain<T>::type>
  >::type
>
{
  static T x();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( +(x()) ) type;
};

template<class T>
struct unary_plus_result<
  T
, typename boost::enable_if<
    boost::is_pointer<typename plain<T>::type>
  >::type
>
{
  typedef T type;
};



// From (5.3.1/7).
////////// negation //////////
template<class T, class = void>
struct negation_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( -(x()) ) type;
}
#endif
;

template<class T>
struct negation_result<
  T
, typename boost::enable_if<
    boost::is_fundamental<typename plain<T>::type>
  >::type
>
{
  static T x();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( -(x()) ) type;
};



// From (5.3.2).
////////// pre-increment //////////
template<class T, class = void>
struct pre_increment_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( ++(x()) ) type;
}
#endif
;

template<class T>
struct pre_increment_result<
  T
, typename boost::enable_if<
    boost::mpl::or_<
      boost::is_fundamental<typename plain<T>::type>
    , boost::is_pointer<typename plain<T>::type>
    >
  >::type
>
  : public boost::add_reference<typename plain<T>::type>
{};



////////// pre-decrement //////////
template<class T, class = void>
struct pre_decrement_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static T x();
  typedef BOOST_TYPEOF_TPL( --(x()) ) type;
}
#endif
;

template<class T>
struct pre_decrement_result<
  T
, typename boost::enable_if<
    boost::mpl::or_<
      boost::is_fundamental<T>
    , boost::is_pointer<T>
    >
  >::type
>
  : public boost::add_reference<typename plain<T>::type>
{};



// From (5.6).
////////// multiplication //////////
template<class LHS, class RHS, class = void>
struct multiplication_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() * rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct multiplication_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<LHS>
    , boost::is_fundamental<RHS>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() * rhs() ) type;
};



////////// division //////////
template<class LHS, class RHS, class = void>
struct division_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( rhs() / lhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct division_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<LHS>
    , boost::is_fundamental<RHS>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() / rhs() ) type;
};



////////// modulo //////////
template<class LHS, class RHS, class = void>
struct modulo_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() % rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct modulo_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<LHS>
    , boost::is_fundamental<RHS>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() % rhs() ) type;
};



// From (5.7).
////////// addition //////////
template<class LHS, class RHS, class = void>
struct addition_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() + rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct addition_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() + rhs() ) type;
};

template<class LHS, class RHS>
struct addition_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_pointer<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
  : public plain<LHS>
{};

template<class LHS, class RHS>
struct addition_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_pointer<typename plain<RHS>::type>
    >
  >::type
>
  : public plain<RHS>
{};



////////// subtraction //////////
template<class LHS, class RHS, class = void>
struct subtraction_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() - rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct subtraction_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() - rhs() ) type;
};

template<class LHS, class RHS>
struct subtraction_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_pointer<typename plain<LHS>::type>
    , boost::is_pointer<typename plain<RHS>::type>
    >
  >::type
>
{
  typedef std::ptrdiff_t type;
};

template<class LHS, class RHS>
struct subtraction_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_pointer<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
  : plain<LHS>
{};



// From (5.8).
////////// left shift ///////////
template<class LHS, class RHS, class = void>
struct left_shift_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() << rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct left_shift_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() << rhs() ) type;
};



////////// right shift //////////
template<class LHS, class RHS, class = void>
struct right_shift_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() >> rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct right_shift_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() >> rhs() ) type;
};



// From (5.11).
////////// bitwise-and //////////
template<class LHS, class RHS, class = void>
struct bitwise_and_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() & rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct bitwise_and_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() & rhs() ) type;
};



// From (5.12).
/////////// bitwise-exclusive-or //////////
template<class LHS, class RHS, class = void>
struct bitwise_exclusive_or_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() ^ rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct bitwise_exclusive_or_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() ^ rhs() ) type;
};



// From (5.13).
////////// bitwise-inclusive-or //////////
template<class LHS, class RHS, class = void>
struct bitwise_inclusive_or_result
#if defined(CRY_FUNCTIONAL_USE_TYPEOF)
{
  static LHS lhs();
  static RHS rhs();
  typedef BOOST_TYPEOF_TPL( lhs() | rhs() ) type;
}
#endif
;

template<class LHS, class RHS>
struct bitwise_inclusive_or_result<
  LHS
, RHS
, typename boost::enable_if<
    boost::mpl::and_<
      boost::is_fundamental<typename plain<LHS>::type>
    , boost::is_fundamental<typename plain<RHS>::type>
    >
  >::type
>
{
  static LHS lhs();
  static RHS rhs();
  typedef typename CRY_DEDUCE_FUNDAMENTAL_TYPE( lhs() | rhs() ) type;
};

} // namespace functional_detail;





struct plus
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< plus const(LHS, RHS) >
    : public functional_detail::addition_result<LHS, RHS>
  {};

  template<class LHS, class RHS>
  struct result< plus(LHS, RHS) >
    : public result< plus const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    plus const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs + rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    plus const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs + rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    plus const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs + rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    plus const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs + rhs;
  }
}; // struct plus



////////// minus //////////
struct minus
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< minus const(LHS, RHS) >
    : public functional_detail::subtraction_result<LHS, RHS>
  { };

  template<class LHS, class RHS>
  struct result< minus(LHS, RHS) >
    : public result< minus const(LHS, RHS) >
  {};

  

  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    minus const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs - rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    minus const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs - rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    minus const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs - rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    minus const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs - rhs;
  }
}; // struct minus



////////// multiplies //////////
struct multiplies
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< multiplies const(LHS, RHS) >
    : public functional_detail::multiplication_result<LHS, RHS>
  {};

  template<class LHS, class RHS>
  struct result< multiplies(LHS, RHS) >
    : public result< multiplies const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    multiplies const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs * rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    multiplies const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs * rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    multiplies const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs * rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    multiplies const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs * rhs;
  }
}; // struct multiplies



////////// divides //////////
struct divides
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< divides const(LHS, RHS) >
    : public functional_detail::division_result<LHS, RHS>
  {};

  template<class LHS, class RHS>
  struct result< divides(LHS, RHS) >
    : public result< divides const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    divides const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs / rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    divides const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs / rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    divides const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs / rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    divides const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs / rhs;
  }
}; // struct divides



////////// modulos //////////
struct modulos
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< modulos const(LHS, RHS) >
    : public functional_detail::modulo_result<LHS, RHS>
  { };

  template<class LHS, class RHS>
  struct result< modulos(LHS, RHS) >
    : public result< modulos const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    modulos const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs % rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    modulos const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs % rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    modulos const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs % rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    modulos const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs % rhs;
  }
}; // struct modulos



////////// negate //////////
struct negate
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class T>
  struct result< negate const(T) >
    : public functional_detail::negation_result<T>
  {};

  template<class T>
  struct result< negate(T) >
    : public result< negate const(T) >
  {};



  ////////// operator() //////////
  template<class T>
  typename boost::result_of<
    negate const(typename boost::add_reference<T>::type)
  >::type
  operator()(T &x) const
  {
    return -x;
  }

  template<class T>
  typename boost::result_of<
    negate const(typename add_const_reference<T>::type)
  >::type
  operator()(T const &x) const
  {
    return -x;
  }
}; // struct negate



////////// equal_to //////////
struct equal_to
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< equal_to const(LHS, RHS) >
  {
    // TODO: Use `BOOST_TYPEOF' when `CRY_FUNCTIONAL_USE_TYPEOF' is specified.
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< equal_to(LHS, RHS) >
    : public result< equal_to const(LHS, RHS) >
  {};
  


  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    equal_to const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs == rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    equal_to const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs == rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    equal_to const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs == rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    equal_to const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs == rhs;
  }
}; // struct equal_to



////////// not_equal_to //////////
struct not_equal_to
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< not_equal_to const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< not_equal_to(LHS, RHS) >
    : public result< not_equal_to const(LHS, RHS) >
  {};

  

  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    not_equal_to const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs != rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    not_equal_to const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs != rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    not_equal_to const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs != rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    not_equal_to const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs != rhs;
  }
}; // struct not_equal_to



////////// less //////////
struct less
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< less const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< less(LHS, RHS) >
    : public result< less const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    less const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs < rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs < rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs < rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs < rhs;
  }
}; // struct less



////////// greater //////////
struct greater
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< greater const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< greater(LHS, RHS) >
    : public result< greater const(LHS, RHS) >
  {};
  


  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    greater const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs)
  {
    return lhs > rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs)
  {
    return lhs > rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs)
  {
    return lhs > rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs)
  {
    return lhs > rhs;
  }
}; // struct greater_



////////// less_equal //////////
struct less_equal
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< less_equal const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< less_equal(LHS, RHS) >
    : public result< less_equal const(LHS, RHS) >
  {};
  


  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    less_equal const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs)
  {
    return lhs <= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less_equal const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs)
  {
    return lhs <= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less_equal const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs)
  {
    return lhs <= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    less_equal const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs)
  {
    return lhs <= rhs;
  }
}; // struct less_equal



////////// greater_equal //////////
struct greater_equal
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< greater_equal const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< greater_equal(LHS, RHS) >
    : public result< greater const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    greater_equal const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs)
  {
    return lhs >= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater_equal const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs)
  {
    return lhs >= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater_equal const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs)
  {
    return lhs >= rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    greater_equal const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs)
  {
    return lhs >= rhs;
  }
}; // struct greater_equal_



////////// logical_and //////////
struct logical_and
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< logical_and const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< logical_and(LHS, RHS) >
    : public result< logical_and const(LHS, RHS) >
  {};
  


  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    logical_and const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs && rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_and const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs && rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_and const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs && rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_and const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs && rhs;
  }
}; // struct logical_and



////////// logical_or //////////
struct logical_or
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class LHS, class RHS>
  struct result< logical_or const(LHS, RHS) >
  {
    typedef bool type;
  };

  template<class LHS, class RHS>
  struct result< logical_or(LHS, RHS) >
    : public result< logical_or const(LHS, RHS) >
  {};



  ////////// operator() //////////
  template<class LHS, class RHS>
  typename boost::result_of<
    logical_or const(
      typename boost::add_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS &rhs) const
  {
    return lhs || rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_or const(
      typename boost::add_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS &lhs, RHS const &rhs) const
  {
    return lhs || rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_or const(
      typename add_const_reference<LHS>::type
    , typename boost::add_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS &rhs) const
  {
    return lhs || rhs;
  }

  template<class LHS, class RHS>
  typename boost::result_of<
    logical_or const(
      typename add_const_reference<LHS>::type
    , typename add_const_reference<RHS>::type
    )
  >::type
  operator()(LHS const &lhs, RHS const &rhs) const
  {
    return lhs || rhs;
  }
}; // struct logical_or



////////// logical_not //////////
struct logical_not
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class T>
  struct result< logical_not const(T) >
  {
    typedef bool type;
  };

  template<class T>
  struct result< logical_not(T) >
    : public result< logical_not const(T) >
  {};



  ////////// operator() //////////
  template<class T>
  typename boost::result_of<
    logical_not const(typename boost::add_reference<T>::type)
  >::type
  operator()(T &x) const
  {
    return !x;
  }

  template<class T>
  typename boost::result_of<
    logical_not const(typename add_const_reference<T>::type)
  >::type
  operator()(T const &x) const
  {
    return !x;
  }
}; // struct logical_not



// struct assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<assign_, LHS, RHS> >
//     : public result< assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs = rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs = rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs = rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs = rhs;
//   }
// }; // struct assign_

// assign_ const assign = assign_();

// struct multiplies_assign_
// {
//   template<class FArgs>
//   struct result;
  
//   template<class LHS, class RHS>
//   struct result< multiplies_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<multiplies_assign_, LHS, RHS> >
//     : public result< multiplies_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< multiplies_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs *= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< multiplies_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs *= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< multiplies_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs *= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< multiplies_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs *= rhs;
//   }
// }; // struct multiplies_assign_

// multiplies_assign_ const multiplies_assign = multiplies_assign_();

// struct divides_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< divides_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<divides_assign_, LHS, RHS> >
//     : public result< divides_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< divides_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs /= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< divides_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs /= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< divides_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs /= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< divides_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs /= rhs;
//   }
// }; // struct divides_assign_

// divides_assign_ const divides_assign = divides_assign_();

// struct modulo_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< modulo_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<modulo_assign_, LHS, RHS> >
//     : public result< modulo_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< modulo_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs %= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< modulo_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs %= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< modulo_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs %= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< modulo_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs %= rhs;
//   }
// }; // struct modulo_assign_

// modulo_assign_ const modulo_assign = modulo_assign_();

// struct plus_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< plus_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<plus_assign_, LHS, RHS> >
//     : public result< plus_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< plus_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs += rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< plus_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs += rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< plus_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs += rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< plus_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs += rhs;
//   }
// }; // struct plus_assign_

// plus_assign_ const plus_assign = plus_assign_();

// struct minus_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< minus_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<minus_assign_, LHS, RHS> >
//     : public result< minus_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< minus_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs -= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< minus_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs -= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< minus_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs -= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< minus_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs -= rhs;
//   }
// }; // struct minus_assign_

// minus_assign_ const minus_assign = minus_assign_();

// struct right_shift_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< right_shift_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<right_shift_assign_, LHS, RHS> >
//     : public result< right_shift_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< right_shift_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs >>= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< right_shift_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs >>= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< right_shift_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs >>= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< right_shift_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs >>= rhs;
//   }
// }; // struct right_shift_assign_

// right_shift_assign_ const right_shift_assign = right_shift_assign_();

// struct left_shift_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< left_shift_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<left_shift_assign_, LHS, RHS> >
//     : public result< left_shift_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< left_shift_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs <<= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< left_shift_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs <<= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< left_shift_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs <<= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< left_shift_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs <<= rhs;
//   }
// }; // struct left_shift_assign_

// left_shift_assign_ const left_shift_assign = left_shift_assign_();

// struct and_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< and_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<and_assign_, LHS, RHS> >
//     : public result< and_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< and_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs &= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< and_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs &= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< and_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs &= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< and_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs &= rhs;
//   }
// }; // struct and_assign_

// and_assign_ const and_assign = and_assign_();

// struct xor_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< xor_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<xor_assign_, LHS, RHS> >
//     : public result< xor_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< xor_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs ^= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< xor_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs ^= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< xor_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs ^= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< xor_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs ^= rhs;
//   }
// }; // struct xor_assign_

// xor_assign_ const xor_assign = xor_assign_();

// struct or_assign_
// {
//   template<class FArgs>
//   struct result;

//   template<class LHS, class RHS>
//   struct result< or_assign_(LHS &, RHS) >
//   {
//     typedef LHS &type;
//   };

// #if defined(CRY_FUNCTIONAL_ENABLE_SIG)
//   template<class FArgs>
//   struct sig;

//   template<class LHS, class RHS>
//   struct sig< boost::tuples::tuple<or_assign_, LHS, RHS> >
//     : public result< or_assign_(LHS &, RHS &) >
//   {};
// #endif
  
//   template<class LHS, class RHS>
//   typename result< or_assign_(LHS &, RHS &) >::type
//   operator()(LHS &lhs, RHS &rhs) const
//   {
//     return lhs |= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< or_assign_(LHS &, RHS const &) >::type
//   operator()(LHS &lhs, RHS const &rhs) const
//   {
//     return lhs |= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< or_assign_(LHS const &, RHS &) >::type
//   operator()(LHS const &lhs, RHS &rhs) const
//   {
//     return lhs |= rhs;
//   }

//   template<class LHS, class RHS>
//   typename result< or_assign_(LHS const &, RHS const &) >::type
//   operator()(LHS const &lhs, RHS const &rhs) const
//   {
//     return lhs |= rhs;
//   }
// }; // struct or_assign_

// or_assign_ const or_assign = or_assign_();



////////// select1st //////////
struct select1st
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class Pair>
  struct result< select1st const(Pair &) >
  {
    typedef typename Pair::first_type &type;
  };

  template<class Pair>
  struct result< select1st(Pair &) >
    : public result< select1st const(Pair &) >
  {};

  template<class Pair>
  struct result< select1st const(Pair const &) >
  {
    typedef typename Pair::first_type const &type;
  };

  template<class Pair>
  struct result< select1st(Pair const &) >
    : public result< select1st const(Pair const &) >
  {};
  
  template<class Pair>
  struct result< select1st const(Pair) >
  {
    // CONSIDER: Plain value or const reference?
    typedef typename Pair::first_type type;
  };

  template<class Pair>
  struct result< select1st(Pair) >
    : public result< select1st const(Pair) >
  {};



  ////////// operator() //////////
  template<class Pair>
  typename boost::result_of<
    select1st const(typename boost::add_reference<Pair>::type)
  >::type
  operator()(Pair &p) const
  {
    return p.first;
  }

  template<class Pair>
  typename boost::result_of<
    select1st const(typename add_const_reference<Pair>::type)
  >::type
  operator()(Pair const &p) const
  {
    return p.first;
  }
}; // struct select1st



////////// select2nd //////////
struct select2nd
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class Pair>
  struct result< select2nd const(Pair &) >
  {
    typedef typename Pair::second_type &type;
  };

  template<class Pair>
  struct result< select2nd(Pair &) >
    : public result< select2nd const(Pair &) >
  {};

  template<class Pair>
  struct result< select2nd const(Pair const &) >
  {
    typedef typename Pair::second_type const &type;
  };

  template<class Pair>
  struct result< select2nd(Pair const &) >
    : public result< select2nd const(Pair const &) >
  {};
  
  template<class Pair>
  struct result< select2nd const(Pair) >
  {
    // CONSIDER: Plain value or const reference?
    typedef typename Pair::second_type type;
  };

  template<class Pair>
  struct result< select2nd(Pair) >
    : public result< select2nd const(Pair) >
  {};



  ////////// operator() //////////
  template<class Pair>
  typename boost::result_of<
    select2nd const(typename boost::add_reference<Pair>::type)
  >::type
  operator()(Pair &p) const
  {
    return p.second;
  }

  template<class Pair>
  typename boost::result_of<
    select2nd const(typename add_const_reference<Pair>::type)
  >::type
  operator()(Pair const &p) const
  {
    return p.second;
  }
}; // struct select2nd

} // namespace cry


#endif // !defined(CRY_FUNCTIONAL_HPP_INCLUDE_GUARD)
