// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_MEMORY_HPP_INCLUDE_GUARD)
#define CRY_MEMORY_HPP_INCLUDE_GUARD

#include <cstddef>
#include <memory>
#include <iterator>

#include <boost/preprocessor.hpp> // TODO: Limit Boost.PP headers.
#include <boost/type_traits/remove_reference.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/move.hpp>
#include <boost/iterator/iterator_traits.hpp>

#include <cry/type_traits/plain.hpp>
#include <cry/utility/enable_if_same.hpp>
#include <cry/utility/enable_if_convertible.hpp>
#include <cry/move/is_move_convertible.hpp>
#include <cry/move/enable_if_movable.hpp>
#include <cry/lambda/enable_sig.hpp>
#include <cry/scope_guard.hpp>


namespace cry{

#define CRY_MEMORY_DETAIL_OVERLOAD_AUTO_PTR_NEW(z, ARITY, _)                  \
template<                                                                      \
  class T                                                                      \
  BOOST_PP_ENUM_TRAILING_BINARY_PARAMS(ARITY, class BOOST_PP_INTERCEPT, A)     \
>                                                                              \
inline                                                                         \
std::auto_ptr<T> auto_ptr_new(BOOST_PP_ENUM_BINARY_PARAMS(ARITY, A, const &a)) \
{                                                                              \
  return std::auto_ptr<T>(new T(BOOST_PP_ENUM_PARAMS(ARITY, a)));              \
}

BOOST_PP_REPEAT(11, CRY_MEMORY_DETAIL_OVERLOAD_AUTO_PTR_NEW, _)

#undef CRY_MEMORY_DETAIL_OVERLOAD_AUTO_PTR_NEW



// Move-aware allocator. Default allocator with move construct.
template<class T>
class allocator
{
public:
  typedef std::size_t size_type;
  typedef std::ptrdiff_t difference_type;
  typedef T *pointer;
  typedef T const *const_pointer;
  typedef T &reference;
  typedef T const &const_reference;
  typedef T value_type;

  template<class U>
  struct rebind
  {
    typedef allocator<U> other;
  };

  allocator() // throw()
  { }

  allocator(allocator const &) // throw()
  { }

  template<class OtherT>
  allocator(allocator<OtherT> const &) // throw()
  { }

  ~allocator() // throw()
  { }

  pointer address(reference x) const
  {
    return std::allocator<T>().address(x);
  }

  const_pointer address(const_reference x) const
  {
    return std::allocator<T>().address(x);
  }

  pointer allocate(size_type count, void const *hint = 0)
  {
    return std::allocator<T>().allocate(count, hint);
  }

  void deallocate(pointer p, size_type n)
  {
    std::allocator<T>().deallocate(p, n);
  }

  size_type max_size() const // throw()
  {
    return std::allocator<T>().max_size();
  }

  // For non-movable type.
  template<class U>
  void construct(
    pointer p, U const &val
  , typename disable_if_movable<value_type, U>::type * = 0
  //, typename enable_if_convertible<U, value_type>::type * = 0
  // TODO: The above SFINAE cannot work when `U = boost::move_from<value_type>'
  //       on GCC. Workaround? Or is `mpl::and' with short circuit evaluation
  //       needed?
  )
  {
    std::allocator<T>().construct(p, val);
  }

  // For non-const lvalue.
  void construct(pointer p, value_type &val)
  {
    std::allocator<T>().construct(p, val);
  }

  // For const lvalue.
  template<class U>
  void construct(
    pointer p, U &val
  , typename enable_if_movable<value_type, U>::type * = 0
  , typename enable_if_same<U, value_type const>::type * = 0
  )
  {
    std::allocator<T>().construct(p, val);
  }

  // For different type.
  template<class U>
  void construct(
    pointer p, U const &val
  , typename enable_if_movable<value_type, U>::type * = 0
  , typename disable_if_same<U, value_type>::type * = 0
  )
  {
    std::allocator<T>().construct(p, val);
  }

  // For move source.
  void construct(pointer p, boost::move_from<value_type> src)
  {
    new (p) T(src);
  }

  void destroy(pointer p)
  {
    std::allocator<T>().destroy(p);
  }

  friend void swap(allocator &, allocator &) // throw()
  { }
};

template<class T, class OtherT>
inline
bool operator==(allocator<T> const &, allocator<OtherT> const &)
{
  return true;
}

template<class T, class OtherT>
inline
bool operator!=(allocator<T> const &, allocator<OtherT>&)
{
  return false;
}

template<>
class allocator<void>
{
public:
  typedef void *pointer;
  typedef void const *const_pointer;
  typedef void value_type;

  template<class OtherT>
  struct rebind
  {
    typedef allocator<OtherT> other;
  };

  allocator() // throw()
  { }

  allocator(allocator<void> const &) // throw()
  { }

  template<class OtherT>
  allocator(allocator<OtherT> const &) // throw()
  { }

  template<class OtherT>
  allocator<void>& operator=(allocator<OtherT> const &) // throw()
  {
    return *this;
  }
};





struct destroy_
  : public lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class TrivialIterator, class Allocator>
  struct result< destroy_ const(TrivialIterator, Allocator) >
  {
    typedef void type;
  };

  template<class TrivialIterator, class Allocator>
  struct result< destroy_(TrivialIterator, Allocator) >
    : public result< destroy_ const(TrivialIterator, Allocator) >
  {};

  template<class InputIterator, class Allocator>
  struct result< destroy_ const(InputIterator, InputIterator, Allocator) >
    : public plain<InputIterator>
  {};

  template<class InputIterator, class Allocator>
  struct result< destroy_(InputIterator, InputIterator, Allocator) >
    : public result< destroy_ const(InputIterator, InputIterator, Allocator) >
  {};



  template<class TrivialIterator, class Allocator>
  typename result< destroy_ const(TrivialIterator, Allocator) >::type
  operator()(TrivialIterator pos, Allocator a) const
  {
    a.destroy(&*pos);
  }

  template<class InputIterator, class Allocator>
  typename result<
    destroy_ const(InputIterator, InputIterator, Allocator)
  >::type
  operator()(InputIterator first, InputIterator last, Allocator a) const
  {
    for(; first != last; ++ first){
      a.destroy(&*first);
    }
    return first;
  }
}; // struct destroy_

destroy_ const destroy = destroy_();





namespace detail{ // Avoids the conflict with boost::move

struct move_
  : public lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<class InputIterator, class OutputIterator>
  struct result< move_ const(InputIterator, InputIterator, OutputIterator) >
    : public plain<OutputIterator>
  {};

  template<class InputIterator, class OutputIterator>
  struct result< move_(InputIterator, InputIterator, OutputIterator) >
    : result< move_ const(InputIterator, InputIterator, OutputIterator) >
  {};



  // operator()
  template<class InputIterator, class OutputIterator>
  typename result<
    move_ const(InputIterator, InputIterator, OutputIterator)
  >::type
  operator()(InputIterator first, InputIterator last, OutputIterator dest) const
  {
    // CONSIDER: ADL hook problem.
    return boost::move(first, last, dest);
  }
}; // struct move_

move_ const move = move_();

} // namespace detail





namespace memory_detail{

template<class SourceIterator, class TargetIterator, class R = void>
struct enable_if_movable
  : public boost::enable_if<
      is_move_convertible<
        typename boost::iterator_value<SourceIterator>::type
      , typename boost::iterator_value<TargetIterator>::type
      >
    , R
    >
{};

template<class SourceIterator, class TargetIterator, class R = void>
struct disable_if_movable
  : public boost::disable_if<
      is_move_convertible<
        typename boost::iterator_value<SourceIterator>::type
      , typename boost::iterator_value<TargetIterator>::type
      >
    , R
    >
{};

} // namespace memory_detail





struct move_rollback_
  : public lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<class TrivialIterator, class OutputIterator>
  struct result< move_rollback_ const(TrivialIterator, OutputIterator) >
  {
    typedef void type;
  };

  template<class TrivialIterator, class OutputIterator>
  struct result< move_rollback_(TrivialIterator, OutputIterator) >
    : public result< move_rollback_ const(TrivialIterator, OutputIterator) >
  {};

  template<class InputIterator, class OutputIterator>
  struct result<
    move_rollback_ const(InputIterator, InputIterator, OutputIterator)
  >
    : public plain<OutputIterator>
  {};

  template<class InputIterator, class OutputIterator>
  struct result< move_rollback_(InputIterator, InputIterator, OutputIterator) >
    : public result<
        move_rollback_ const(InputIterator, InputIterator, OutputIterator)
      >
  {};



  // operator()
  // WORKAROUND: MSVC cannot treat SFINAE in operator() signature. So additional
  //             indirection is needed.
  template<class TrivialIterator, class OutputIterator>
  typename result<
    move_rollback_ const(TrivialIterator, OutputIterator)
  >::type
  operator()(TrivialIterator pos , OutputIterator dest) const // throw()
  {
    do_invoke(pos, dest);
  }

  template<class InputIterator, class OutputIterator>
  typename result<
    move_rollback_ const(InputIterator, InputIterator, OutputIterator)
  >::type
  operator()(InputIterator first, InputIterator last, OutputIterator dest) const
  // throw()
  {
    return do_invoke(first, last, dest);
  }


  
  // do_invoke
  template<class TrivialIterator, class OutputIterator>
  typename result<
    move_rollback_ const(TrivialIterator, OutputIterator)
  >::type
  do_invoke(
    TrivialIterator pos, OutputIterator dest
  , typename memory_detail::enable_if_movable<
      TrivialIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    using boost::move;
    *dest = move(*pos);
  }

  template<class TrivialIterator, class OutputIterator>
  typename result<
    move_rollback_ const(TrivialIterator, OutputIterator)
  >::type
  do_invoke(
    TrivialIterator, OutputIterator
  , typename memory_detail::disable_if_movable<
      TrivialIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    ; // Do nothing.
  }

  template<class InputIterator, class OutputIterator>
  typename result<
    move_rollback_ const(InputIterator, InputIterator, OutputIterator)
  >::type
  do_invoke(
    InputIterator first, InputIterator last, OutputIterator dest
  , typename memory_detail::enable_if_movable<
      InputIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    return boost::move(first, last, dest);
  }

  template<class InputIterator, class OutputIterator>
  typename result<
    move_rollback_ const(InputIterator, InputIterator, OutputIterator)
  >::type
  do_invoke(
    InputIterator first, InputIterator last, OutputIterator dest
  , typename memory_detail::disable_if_movable<
      InputIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    std::advance(dest, std::distance(first, last));
    return dest;
  }
}; // struct move_rollback_

move_rollback_ const move_rollback = move_rollback_();





struct uninitialized_move_rollback_
  : public lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<class TrivialIterator, class OutputIterator, class Allocator>
  struct result<
    uninitialized_move_rollback_ const(
      TrivialIterator, OutputIterator, Allocator
    )
  >
  {
    typedef void type;
  };

  template<class TrivialIterator, class OutputIterator, class Allocator>
  struct result<
    uninitialized_move_rollback_(TrivialIterator, OutputIterator, Allocator)
  >
    : public result<
        uninitialized_move_rollback_ const(
          TrivialIterator, OutputIterator, Allocator
        )
      >
  {};

  template<class InputIterator, class OutputIterator, class Allocator>
  struct result<
    uninitialized_move_rollback_ const(
      InputIterator, InputIterator, OutputIterator, Allocator
    )
  >
    : public plain<OutputIterator>
  {};

  template<class InputIterator, class OutputIterator, class Allocator>
  struct result<
    uninitialized_move_rollback_(
      InputIterator, InputIterator, OutputIterator, Allocator
    )
  >
    : public result<
        uninitialized_move_rollback_ const(
          InputIterator, InputIterator, OutputIterator, Allocator
        )
      >
  {};



  // operator()
  template<class TrivialIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      TrivialIterator, OutputIterator, Allocator
    )
  >::type
  operator()(TrivialIterator pos, OutputIterator dest, Allocator a) const
  // throw()
  {
    do_invoke(pos, dest, a);
  }

  template<class InputIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      InputIterator, InputIterator, OutputIterator, Allocator
    )
  >::type
  operator()(
    InputIterator first, InputIterator last, OutputIterator dest, Allocator a
  ) const // throw()
  {
    return do_invoke(first, last, dest, a);
  }
  


  // do_invoke
  template<class TrivialIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      TrivialIterator, OutputIterator, Allocator
    )
  >::type
  do_invoke(
    TrivialIterator pos, OutputIterator dest, Allocator a
  , typename memory_detail::enable_if_movable<
      TrivialIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    using boost::move;
    *dest = move(*pos);
    a.destroy(&*pos);
  }

  template<class TrivialIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      TrivialIterator, OutputIterator, Allocator
    )
  >::type
  do_invoke(
    TrivialIterator pos, OutputIterator, Allocator a
  , typename memory_detail::disable_if_movable<
      TrivialIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    a.destroy(&*pos);
  }

  template<class InputIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      InputIterator, InputIterator, OutputIterator, Allocator
    )
  >::type
  do_invoke(
    InputIterator first, InputIterator last, OutputIterator dest, Allocator a
  , typename memory_detail::enable_if_movable<
      InputIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    dest = boost::move(first, last, dest);
    destroy(first, last, a);
    return dest;
  }

  template<class InputIterator, class OutputIterator, class Allocator>
  typename result<
    uninitialized_move_rollback_ const(
      InputIterator, InputIterator, OutputIterator, Allocator
    )
  >::type
  do_invoke(
    InputIterator first, InputIterator last, OutputIterator dest, Allocator a
  , typename memory_detail::disable_if_movable<
      InputIterator, OutputIterator
    >::type * = 0
  ) const // throw()
  {
    destroy(first, last, a);
    std::advance(dest, std::distance(first, last));
    return dest;
  }
}; // struct uninitialized_move_rollback_

uninitialized_move_rollback_ const uninitialized_move_rollback
  = uninitialized_move_rollback_();





namespace memory_detail{

template<class ForwardIterator, class Allocator>
class growing_destroyer
{
public:
  growing_destroyer(ForwardIterator first, ForwardIterator &last, Allocator a)
    : first_(first)
    , p_last_(&last)
    , a_(a)
  { }

  void operator()() const // throw()
  {
    destroy(first_, *p_last_, a_);
  }

private:
  ForwardIterator first_;
  ForwardIterator *p_last_;
  Allocator a_;
};

template<class FwdRWIterator, class Allocator>
inline
growing_destroyer<FwdRWIterator, Allocator> make_growing_destroyer(
  FwdRWIterator first, FwdRWIterator &last, Allocator a
)
{
  return growing_destroyer<FwdRWIterator, Allocator>(first, last, a);
}

} // namespace memory_detail

template<class InputIterator, class ForwardIterator, class Allocator>
inline
ForwardIterator uninitialized_copy(
  InputIterator first, InputIterator last, ForwardIterator dest, Allocator a
)
{
  using namespace memory_detail;

  scope_guard sg = make_guard(
      memory_detail::make_growing_destroyer(dest, dest, a)
    );
  for(; first != last; ++first, ++dest){
    a.construct(&*dest, *first);
  }

  // Commit
  sg.dismiss();
  return dest;
}

template<class ForwardIterator, class Value, class Allocator>
inline
ForwardIterator uninitialized_fill(
  ForwardIterator first, ForwardIterator last, Value const &value, Allocator a
)
{
  scope_guard sg = make_guard(
      memory_detail::make_growing_destroyer(first, first, a)
    );
  for(; first != last; ++first){
    a.construct(&*first, value);
  }

  // Commit
  sg.dismiss();
  return first;
}





namespace memory_detail{

struct movable_tag{ };
struct non_movable_tag{ };

template<class T>
inline
typename boost::mpl::if_<
  boost::is_movable<T>, movable_tag, non_movable_tag
>::type
move_category(T const &)
{
  return typename boost::mpl::if_<
      boost::is_movable<T>, movable_tag, non_movable_tag
    >::type();
}

} // namespace memory_detail

template<class InputIterator, class OutputIterator, class Allocator>
inline
OutputIterator uninitialized_move_dispatch(
  InputIterator first, InputIterator last, OutputIterator dest, Allocator a
  // The constness of `iterator_reference<InputIterator>::type' affects
  // movability. Say, if `InputIterator' is immutable iterator type, then
  // I cannot move the resource from it.
, typename boost::enable_if<
    boost::is_movable<
      typename boost::remove_reference<
        typename boost::iterator_reference<InputIterator>::type
      >::type
    >
  >::type * = 0
)
{
  using boost::move;
  
  // It is presumed that move operations must not throw.
  for(; first != last; ++first, ++dest){
    a.construct(&*dest, move(*first));
  }

  return dest;
}

template<class InputIterator, class OutputIterator, class Allocator>
inline
OutputIterator uninitialized_move_dispatch(
  InputIterator first, InputIterator last, OutputIterator dest, Allocator a
, typename boost::disable_if<
    boost::is_movable<
      typename boost::remove_reference<
        typename boost::iterator_reference<InputIterator>::type
      >::type
    >
  >::type * = 0
)
{
  return uninitialized_copy(first, last, dest, a);
}

template<class InputIterator, class OutputIterator, class Allocator>
inline
OutputIterator uninitialized_move(
  InputIterator first, InputIterator last, OutputIterator dest, Allocator a
)
{
  return uninitialized_move_dispatch(first, last, dest, a);
}

template<class InputIterator, class OutputIterator>
inline
OutputIterator uninitialized_move(
  InputIterator first, InputIterator last, OutputIterator dest
)
{
  typedef typename boost::iterator_value<InputIterator>::type value_type;
  return uninitialized_move(first, last, dest, allocator<value_type>());
}

} // namespace cry

#endif // !defined(CRY_MEMORY_HPP_INCLUDE_GUARD)
