// (C) Copyright Cryolite 2005.
// (C) Copyright Jonathan Turkanis 2004.
// (C) Copyright Daniel Wallin 2004.
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt.)

// Implementation of the move_ptr from the "Move Proposal" 
// (http://std.dkuug.dk/jtc1/sc22/wg21/docs/papers/2002/n1377.htm) 
// enhanced to support custom deleters and safe boolean conversions.
//
// The implementation is based on an implementation by Daniel Wallin, at
// "http://aspn.activestate.com/ASPN/Mail/Message/Attachments/boost/
// 400DC271.1060903@student.umu.se/move_ptr.hpp". The current was adapted 
// by Jonathan Turkanis to incorporating ideas of Howard Hinnant and 
// Rani Sharoni. 

// Note: This implementation is borrowed from
// http://www.kangaroologic.com/move_ptr/libs/move_ptr/doc/index.html
// And I have added some changes.
//
//                                                    2005 Cryolite

#ifndef CRY_STATIC_MOVE_PTR_HPP_INCLUDE_GUARD
#define CRY_STATIC_MOVE_PTR_HPP_INCLUDE_GUARD

#include <cstddef>

#include <boost/config.hpp>

#include <boost/mpl/bool.hpp>

#include <boost/type_traits/is_const.hpp>
#include <boost/type_traits/is_array.hpp>
#include <boost/type_traits/is_class.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/add_reference.hpp>
#include <boost/type_traits/remove_bounds.hpp>

#include <boost/static_assert.hpp>

#include <boost/utility/enable_if.hpp>

#include <boost/move.hpp>

#include <boost/compressed_pair.hpp>

#include <cry/type_traits/is_incomplete_array.hpp>
#include <cry/move_ptr/is_convertible.hpp>
#include <cry/deleter/default_deleter.hpp>

#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable:4180)
#pragma warning(disable:4348)
#pragma warning(disable:4521)
#pragma warning(disable:4522)
#endif


namespace cry{

template<
  class T, class Deleter = default_deleter<T>
>
class static_move_ptr
{
public:
  typedef typename boost::remove_bounds<T>::type element_type;
  typedef Deleter                                deleter_type;
  
private:
  BOOST_STATIC_CONSTANT(bool, is_array = boost::is_array<T>::value);
  BOOST_STATIC_ASSERT(!is_array || is_incomplete_array<T>::value);
  struct safe_bool_helper { int x; };
  typedef int safe_bool_helper::* safe_bool;
  typedef boost::compressed_pair<element_type *, Deleter> impl_type;
  
public:
  typedef typename impl_type::second_reference        deleter_reference;
  typedef typename impl_type::second_const_reference  deleter_const_reference;

  // Constructors
  
  // Move constructor from rvalue.
  static_move_ptr(static_move_ptr const &p)
    : impl_(p.get(), p.get_deleter())
  {
    const_cast<static_move_ptr &>(p).release();
  }

  // Move constructor from rvalue with derived-to-base conversion.
  template<class AnotherT, class AnotherDeleter>
  static_move_ptr(
    static_move_ptr<AnotherT, AnotherDeleter> const &p
  , typename move_ptrs::enable_if_convertible<AnotherT, T>::type * = 0
  )
    : impl_(p.get(), p.get_deleter())
  {
    BOOST_STATIC_ASSERT((
        move_ptrs::is_convertible<AnotherT, T>::value
      ));
    const_cast<static_move_ptr<AnotherT, AnotherDeleter> &>(p).release();
  }

  // Move constructor from explicit move source.
  template<class AnotherT, class AnotherDeleter>
  static_move_ptr(
    boost::move_from<static_move_ptr<AnotherT, AnotherDeleter> > src
  )
    : impl_(src->get(), src->get_deleter())
  {
    BOOST_STATIC_ASSERT((
        move_ptrs::is_convertible<AnotherT, T>::value
      ));
    src->release();
  }
  
  // Constructor from the ordinal pointer
  // This is for  conversion from literal zero.
  // CONSIDER: Is `explicit' needed? Currently I follow the designs in
  //           Boost's smart pointer classes and add it.
  explicit static_move_ptr(element_type *p = 0)
    : impl_(p, Deleter())
  {
    BOOST_STATIC_ASSERT((boost::is_class<Deleter>::value));
  }
  
  template<typename AnotherT>
  explicit static_move_ptr(AnotherT *p) 
    : impl_(p, Deleter()) 
  {
    BOOST_STATIC_ASSERT((boost::is_class<Deleter>::value));
  }

  // This is for conversion from literal zero.
  template<class AnotherDeleter>
  static_move_ptr(element_type *p, AnotherDeleter d)
    : impl_(p, d)
  { }
  
  template<typename AnotherT, typename AnotherDeleter>
  static_move_ptr(AnotherT *p, AnotherDeleter d)
    : impl_(p, d)
  { }

  operator boost::const_lvalue<static_move_ptr>() const
  {
    return boost::const_lvalue<static_move_ptr>(this);
  }

  operator boost::move_from<static_move_ptr>()
  {
    return boost::move_from<static_move_ptr>(this);
  }
  
  // Destructor
  ~static_move_ptr(){
    if(get()){
      get_deleter()(get());
    }
  }

  // Move assignment
  static_move_ptr &operator=(static_move_ptr const &src)
  {
    using boost::move;

    static_move_ptr tmp(move(const_cast<static_move_ptr &>(src)));
    tmp.swap(*this);
    return *this;
  }
  
  // Move assignment with derived-to-base conversion.
  template<class AnotherT, class AnotherDeleter>
  typename move_ptrs::enable_if_convertible<
    AnotherT, T, static_move_ptr &
  >::type
  operator=(static_move_ptr<AnotherT, AnotherDeleter> const &src)
  {
    using boost::move;
    
    static_move_ptr tmp(
        move(const_cast<static_move_ptr<AnotherT, AnotherDeleter> &>(src))
      );
    tmp.swap(*this);
    return *this;
  }
  
  // Assignment from explicit move source, including the case where
  // derived-to-base conversion is applied.
  template<typename AnotherT, typename AnotherDeleter>
  typename move_ptrs::enable_if_convertible<
    AnotherT, T, static_move_ptr &
  >::type
  operator=(boost::move_from<static_move_ptr<AnotherT, AnotherDeleter> > src)
  {
    static_move_ptr<T, Deleter> tmp(src);
    tmp.swap(*this);
    return *this;
  }

  // Smart pointer interface
  element_type *get() const { return impl_.first(); }

  element_type &operator*() const 
  { 
    BOOST_STATIC_ASSERT(!is_array);
    return *get(); 
  }

  element_type *operator->() const 
  { 
    BOOST_STATIC_ASSERT(!is_array);
    return get();
  }

  element_type &operator[](std::size_t i) const 
  { 
    BOOST_STATIC_ASSERT(is_array);
    return get()[i];
  }

  element_type *release()
  {
    element_type *result = get();
    impl_.first() = 0;
    return result;
  }

  void reset()
  {
    if(get()){
      get_deleter()(get());
    }
    impl_.first() = 0;
  }
  
  template<typename AnotherT>
  void reset(AnotherT *p) 
  {
    static_move_ptr(p).swap(*this);
  }

  template<typename AnotherT, typename AnotherDeleter>
  void reset(AnotherT *p, AnotherDeleter d)
  {
    static_move_ptr(p, d).swap(*this);
  }

  operator safe_bool() const { return get() ? &safe_bool_helper::x : 0; }
  
  void swap(static_move_ptr &r)
  {
    using std::swap;
    
#if defined(__GNUC__)
    // WORKAROUND: GCC wrongly looks up boost::move_::swap via ADL.
    //             This workaround disables ADL customizations for user-defined
    //             types.
    std::swap(impl_.first(), r.impl_.first());
    std::swap(impl_.second(), r.impl_.second());
#else
    impl_.swap(r.impl_);
#endif
  }

  friend void swap(static_move_ptr &p, static_move_ptr &q)
  {
    p.swap(q);
  }
  
  deleter_type &get_deleter() { return impl_.second(); }

  deleter_type const &get_deleter() const { return impl_.second(); }

private:
//   template<class TT>
//   struct is_const_static_move_ptr
//     : boost::mpl::false_
//   {};

//   template<class TT, class DD>
//   struct is_const_static_move_ptr<static_move_ptr<TT, DD> const>
//     : boost::mpl::true_
//   {};
  
  template<class Ptr, class Enabler = void>
  struct cannot_move_from_const;

  template<class Ptr>
  struct cannot_move_from_const<
    Ptr
  , typename boost::enable_if<
      boost::is_const<Ptr>
    >::type
  >
  {
    typedef typename Ptr::error type; // TODO: Be more safe.
  };

  // Non-const lvalue trap
  static_move_ptr(static_move_ptr &);

  // Const lvalue trap
  template<class Ptr>
  static_move_ptr(Ptr &, typename cannot_move_from_const<Ptr>::type * = 0);

  // Non-const lvalue with conversion trap
  template<class AnotherT, class AnotherDeleter>
  static_move_ptr(
    static_move_ptr<AnotherT, AnotherDeleter> &
  , typename move_ptrs::enable_if_convertible<
      AnotherT, T
    >::type * = 0
  );

  // Non-const lvalue trap
  static_move_ptr &operator=(static_move_ptr &);

  // Const lvalue trap
  template<class Ptr>
  typename cannot_move_from_const<Ptr>::type
  operator=(Ptr &);

  // Non-const lvalue with convesion trap
  template<class AnotherT, class AnotherDeleter>
  typename move_ptrs::enable_if_convertible<
    AnotherT, T, static_move_ptr &
  >::type
  operator=(static_move_ptr<AnotherT, AnotherDeleter> &);
  
private:
  impl_type impl_;
}; // class static_move_ptr

} // namespace cry


#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif      // #ifndef BOOST_STATIC_MOVE_PTR_HPP_INCLUDED
