// (C) Copyright Cryolite 2004, 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//  This header defines range-oriented STL algorithm. They can accept models of
// RangeConcept. And they are provided as a const object of a functor, not a
// function. Advantages of this design are;
//
//   - to be callbacked easily. Similar idea have been already implemented in
//     Boost.Lambda. See
//     http://tinyurl.com/7sqjo
//   - to avoid unexpected ADL lookup. See
//     http://www.open-std.org/jtc1/sc22/wg21/docs/lwg-defects.html#225
//     (TODO: Confirm that ADL can only look up functions.)
//
// Disadvantages are;
//
//   - cannot provide explicit template arguments. (In fact, it can be done, but
//     syntax is verbose, i.e.
//
//       equal.operator()< vector<int> >(u, v);
//
//     )
//   - cannot specialize.
// 
// However, disadvantages do not seem so critical. Are explicit template
// arguments really needed? Specializations can be done by providing the
// special ADL hooks for user-defined types.
//
//  Semantics of some algorithms, for example `equal', differ from STL's
// (iterator-oriented) ones, and they would be more natural.

// Note: Overloads for const reference are needed in order to bind temporary.

// TODO: Static template method. The current implementation is too loose.
//       It can find user-defined `begin,' `end' and so on via ADL. There may be
//       two ways to call free-standing hooks.
//
//         - call the specialized hooks such as `range_begin.'
//         - stick to call the Boost's original hooks, i.e. always call hooks
//           with the qualification.
//
// TODO: Specify hook dependencies. This is documentation todo.
// TODO: Reimplement all iterator-oriented algorithms to control hook
//       denpencies.

#if !defined(CRY_RANGE_ALGORITHM_HPP_INCLUDE_GUARD)
#define CRY_RANGE_ALGORITHM_HPP_INCLUDE_GUARD

#include <algorithm>

#include <boost/type_traits/add_reference.hpp>
#include <boost/type_traits/remove_reference.hpp>

#include <boost/iterator/iterator_categories.hpp>
#include <boost/iterator/iterator_traits.hpp>

#include <boost/range/value_type.hpp>
#include <boost/range/difference_type.hpp>
#include <boost/range/begin.hpp>

#include <boost/random/mersenne_twister.hpp>

#include <cry/range/end.hpp>
#include <cry/range/size.hpp>

#include <cry/type_traits/plain.hpp>
#include <cry/type_traits/add_const.hpp>
#include <cry/type_traits/add_const_reference.hpp>
#include <cry/lambda/enable_sig.hpp>
#include <cry/range/iterator.hpp>
#include <cry/range/reference.hpp>
#include <cry/functional.hpp>
#include <cry/algorithm.hpp>


namespace cry{ namespace range{

////////////////////////////////////////////////////////////////////////////////
//for_each
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class UnaryFunction>
UnaryFunction range_for_each(SinglePassRange &r, UnaryFunction f)
{
  using boost::begin;
  using boost::end;

  return std::for_each(begin(r), end(r), f);
}

struct for_each_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class UnaryFunction>
  struct result< for_each_ const(SinglePassRange, UnaryFunction) >
    : public plain<UnaryFunction>
  { };

  template<class SinglePassRange, class UnaryFunction>
  struct result< for_each_(SinglePassRange, UnaryFunction) >
    : public result< for_each_ const(SinglePassRange, UnaryFunction) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class UnaryFunction>
  typename result<
    for_each_ const(
      typename boost::add_reference<SinglePassRange>::type
    , UnaryFunction
    )
  >::type operator()(SinglePassRange &r, UnaryFunction f) const
  {
    return range_for_each(r, f);
  }

  template<class SinglePassRange, class UnaryFunction>
  typename result<
    for_each_ const(
      typename add_const_reference<SinglePassRange>::type
    , UnaryFunction
    )
  >::type operator()(SinglePassRange const &r, UnaryFunction f) const
  {
    return range_for_each(r, f);
  }
}; // struct for_each_

for_each_ const for_each = for_each_();



////////////////////////////////////////////////////////////////////////////////
//find
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class T>
typename iterator<SinglePassRange>::type
range_find(SinglePassRange &r, T const &value)
{
  using boost::begin;
  using boost::end;

  return std::find(begin(r), end(r), value);
}

struct find_
  : public cry::lambda::enable_sig
{
  // Return type calculation
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class T>
  struct result< find_ const(SinglePassRange, T) >
    : public iterator<
        typename boost::remove_reference<SinglePassRange>::type
      >
  { };

  template<class SinglePassRange, class T>
  struct result< find_(SinglePassRange, T) >
    : public result< find_ const(SinglePassRange, T) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class T>
  typename result<
    find_ const(
      typename boost::add_reference<SinglePassRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(SinglePassRange &r, T const &value) const
  {
    return range_find(r, value);
  }

  template<class SinglePassRange, class T>
  typename result<
    find_ const(
      typename add_const_reference<SinglePassRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(SinglePassRange const &r, T const &value) const
  {
    return range_find(r, value);
  }
}; // struct find_

find_ const find = find_();



////////////////////////////////////////////////////////////////////////////////
//find_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class UnaryPredicate>
typename iterator<SinglePassRange>::type
range_find_if(SinglePassRange &r, UnaryPredicate pred)
{
  using boost::begin;
  using boost::end;

  return std::find_if(begin(r), end(r), pred);
}

struct find_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class UnaryPredicate>
  struct result< find_if_ const(SinglePassRange, UnaryPredicate) >
    : public iterator<typename boost::remove_reference<SinglePassRange>::type>
  { };

  template<class SinglePassRange, class UnaryPredicate>
  struct result< find_if_(SinglePassRange, UnaryPredicate) >
    : public result< find_if_ const(SinglePassRange, UnaryPredicate) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class UnaryPredicate>
  typename result<
    find_if_ const(
      typename boost::add_reference<SinglePassRange>::type
    , UnaryPredicate
    )
  >::type operator()(SinglePassRange &r, UnaryPredicate pred) const
  {
    return range_find_if(r, pred);
  }

  template<class SinglePassRange, class UnaryPredicate>
  typename result<
    find_if_ const(
      typename add_const_reference<SinglePassRange>::type
    , UnaryPredicate
    )
  >::type operator()(SinglePassRange const &r, UnaryPredicate pred) const
  {
    return range_find_if(r, pred);
  }
}; // struct find_if_

find_if_ const find_if = find_if_();



////////////////////////////////////////////////////////////////////////////////
//find_end
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
typename iterator<ForwardRange0>::type
range_find_end(ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred)
{
  using boost::begin;
  using boost::end;

  return std::find_end(begin(r0), end(r0), begin(r1), end(r1), pred);
}

struct find_end_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange0, class ForwardRange1>
  struct result< find_end_ const(ForwardRange0, ForwardRange1) >
    : public iterator< typename boost::remove_reference<ForwardRange0>::type >
  { };

  template<class ForwardRange0, class ForwardRange1>
  struct result< find_end_(ForwardRange0, ForwardRange1) >
    : public result< find_end_ const(ForwardRange0, ForwardRange1) >
  { };

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result< find_end_(ForwardRange0, ForwardRange1, BinaryPredicate) >
    : public result< find_end_ const(ForwardRange0, ForwardRange1) >
  { };  

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result<
    find_end_ const(ForwardRange0, ForwardRange1, BinaryPredicate)
  >
    : public result< find_end_ const(ForwardRange0, ForwardRange1) >
  { };  



  ////////// operator() //////////
  template<class ForwardRange0, class ForwardRange1>
  typename result<
    find_end_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type operator()(ForwardRange0 &r0, ForwardRange1 const &r1) const
  {
    return range_find_end(r0, r1, equal_to());
  }

  template<class ForwardRange0, class ForwardRange1>
  typename result<
    find_end_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type operator()(ForwardRange0 const &r0, ForwardRange1 const &r1) const
  {
    return range_find_end(r0, r1, equal_to());
  }

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    find_end_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_find_end(r0, r1, pred);
  }

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    find_end_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 const &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_find_end(r0, r1, pred);
  }
}; // struct find_end_

find_end_ const find_end = find_end_();



////////////////////////////////////////////////////////////////////////////////
//find_first_of
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
typename iterator<ForwardRange0>::type
range_find_first_of(
  ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::find_first_of(begin(r0), end(r0), begin(r1), end(r1), pred);
}

struct find_first_of_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange0, class ForwardRange1>
  struct result< find_first_of_ const(ForwardRange0, ForwardRange1) >
    : public iterator< typename boost::remove_reference<ForwardRange0>::type >
  { };

  template<class ForwardRange0, class ForwardRange1>
  struct result< find_first_of_(ForwardRange0, ForwardRange1) >
    : public result< find_first_of_ const(ForwardRange0, ForwardRange1) >
  { };

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result< find_first_of_(ForwardRange0, ForwardRange1, BinaryPredicate) >
    : public result< find_first_of_ const(ForwardRange0, ForwardRange1) >
  { };

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result<
    find_first_of_ const(ForwardRange0, ForwardRange1, BinaryPredicate)
  >
    : public result< find_first_of_ const(ForwardRange0, ForwardRange1) >
  { };



  ////////// operator() //////////
  template<class ForwardRange0, class ForwardRange1>
  typename result<
    find_first_of_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type operator()(ForwardRange0 &r0, ForwardRange1 const &r1) const
  {
    return range_find_first_of(r0, r1, equal_to());
  }

  template<class ForwardRange0, class ForwardRange1>
  typename result<
    find_first_of_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type operator()(ForwardRange0 const &r0, ForwardRange1 const &r1) const
  {
    return range_find_first_of(r0, r1, equal_to());
  }
  
  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    find_first_of_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_find_first_of(r0, r1, pred);
  }

  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    find_first_of_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 const &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_find_first_of(r0, r1, pred);
  }
}; // struct find_first_of_

find_first_of_ const find_first_of = find_first_of_();



////////////////////////////////////////////////////////////////////////////////
//adjacent_find
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class BinaryPredicate>
typename iterator<ForwardRange>::type
range_adjacent_find(ForwardRange &r, BinaryPredicate pred)
{
  using boost::begin;
  using boost::end;
  
  return std::adjacent_find(begin(r), end(r), pred);
}

struct adjacent_find_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange>
  struct result< adjacent_find_ const(ForwardRange) >
    : public iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange>
  struct result< adjacent_find_(ForwardRange) >
    : public result< adjacent_find_ const(ForwardRange) >
  { };

  template<class ForwardRange, class BinaryPredicate>
  struct result< adjacent_find_(ForwardRange, BinaryPredicate) >
    : public result< adjacent_find_ const(ForwardRange) >
  { };

  template<class ForwardRange, class BinaryPredicate>
  struct result< adjacent_find_ const(ForwardRange, BinaryPredicate) >
    : public result< adjacent_find_ const(ForwardRange) >
  { };



  ////////// operator() //////////
  template<class ForwardRange>
  typename result<
    adjacent_find_ const(typename boost::add_reference<ForwardRange>::type)
  >::type operator()(ForwardRange &r) const
  {
    return range_adjacent_find(r, equal_to());
  }
  
  template<class ForwardRange>
  typename result<
    adjacent_find_ const(typename add_const_reference<ForwardRange>::type)
  >::type operator()(ForwardRange const &r) const
  {
    return range_adjacent_find(r, equal_to());
  }
  
  template<class ForwardRange, class BinaryPredicate>
  typename result<
    adjacent_find_ const(
      typename boost::add_reference<ForwardRange>::type
    , BinaryPredicate
    )
  >::type operator()(ForwardRange &r, BinaryPredicate pred) const
  {
    return range_adjacent_find(r, pred);
  }
  
  template<class ForwardRange, class BinaryPredicate>
  typename result<
    adjacent_find_ const(
      typename add_const_reference<ForwardRange>::type
    , BinaryPredicate
    )
  >::type operator()(ForwardRange const &r, BinaryPredicate pred) const
  {
    return range_adjacent_find(r, pred);
  }
}; // struct adjacent_find_

adjacent_find_ const adjacent_find = adjacent_find_();



////////////////////////////////////////////////////////////////////////////////
// count
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class T>
typename boost::range_difference<
  typename add_const<SinglePassRange>::type // CONSIDER: Is add_const necessary?
>::type
range_count(SinglePassRange const &r, T const &value)
{
  using boost::begin;
  using boost::end;
  
  return std::count(begin(r), end(r), value);
}

struct count_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;
  
  template<class SinglePassRange, class T>
  struct result< count_ const(SinglePassRange, T) >
    : public boost::range_difference<
        typename boost::remove_reference<SinglePassRange>::type
      >
  { };

  template<class SinglePassRange, class T>
  struct result< count_(SinglePassRange, T) >
    : public result< count_ const(SinglePassRange, T) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class T>
  typename result<
    count_ const(
      typename add_const_reference<SinglePassRange>::type
    , typename add_const_reference<T>::type
    )
  >::type
  operator()(SinglePassRange const &r, T const &value) const
  {
    return range_count(r, value);
  }
}; // struct count_

count_ const count = count_();



////////////////////////////////////////////////////////////////////////////////
// count_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class UnaryPredicate>
typename boost::range_difference<
  typename add_const<SinglePassRange>::type // Is add_const necessary?
>::type range_count_if(SinglePassRange const &r, UnaryPredicate pred)
{
  using boost::begin;
  using boost::end;
  
  return std::count_if(begin(r), end(r), pred);
}

struct count_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class UnaryPredicate>
  struct result< count_if_ const(SinglePassRange, UnaryPredicate) >
    : public boost::range_difference<
        typename boost::remove_reference<SinglePassRange>::type
      >
  { };

  template<class SinglePassRange, class UnaryPredicate>
  struct result< count_if_(SinglePassRange, UnaryPredicate) >
    : result< count_if_ const(SinglePassRange, UnaryPredicate) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class UnaryPredicate>
  typename result<
    count_if_ const(
      typename add_const_reference<SinglePassRange>::type
    , UnaryPredicate
    )
  >::type operator()(SinglePassRange const &r, UnaryPredicate pred) const
  {
    return range_count_if(r, pred);
  }
}; // struct count_if_

count_if_ const count_if = count_if_();



////////////////////////////////////////////////////////////////////////////////
// mismatch
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange0, class SinglePassRange1, class BinaryPredicate>
std::pair<
  typename iterator<SinglePassRange0>::type
, typename iterator<SinglePassRange1>::type
>
range_mismatch(SinglePassRange0 &r0, SinglePassRange1 &r1, BinaryPredicate pred)
{
  // Here are some design strategies for mismatch. One may put
  // pre-condition that size(r0) < size(r1) (as STL's does). The current
  // implementation is safe for any situation. It differs STL's semantics.
  using boost::begin;
  using boost::end;

  typename iterator<SinglePassRange0>::type it0(begin(r0));
  typename iterator<SinglePassRange0>::type e0(end(r0));
  typename iterator<SinglePassRange1>::type it1(begin(r1));
  typename iterator<SinglePassRange1>::type e1(end(r1));
  for(; it0 != e0 && it1 != e1; ++it0, ++it1)
  {
    if(!pred(*it0, *it1)){
      break;
    }
  }

  return std::make_pair(it0, it1);
}

struct mismatch_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< mismatch_ const(SinglePassRange0, SinglePassRange1) >
  {
    typedef std::pair<
        typename iterator<
          typename boost::remove_reference<SinglePassRange0>::type
        >::type
      , typename iterator<
          typename boost::remove_reference<SinglePassRange1>::type
        >::type
      > type;
  };

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< mismatch_(SinglePassRange0, SinglePassRange1) >
    : public result< mismatch_ const(SinglePassRange0, SinglePassRange1) >
  { };
  
  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  struct result<
    mismatch_(SinglePassRange0, SinglePassRange1, BinaryPredicate)
  >
    : public result< mismatch_ const(SinglePassRange0, SinglePassRange1) >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  struct result<
    mismatch_ const(SinglePassRange0, SinglePassRange1, BinaryPredicate)
  >
    : public result< mismatch_ const(SinglePassRange0, SinglePassRange1) >
  { };
  
  

  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    mismatch_ const(
      typename boost::add_reference<SinglePassRange0>::type
    , typename boost::add_reference<SinglePassRange0>::type
    )
  >::type operator()(
    SinglePassRange0 &r0, SinglePassRange1 &r1
  ) const
  {
    return range_mismatch(r0, r1, equal_to());
  }

  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    mismatch_ const(
      typename boost::add_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 &r0, SinglePassRange1 const &r1
  ) const
  {
    return range_mismatch(r0, r1, equal_to());
  }

  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    mismatch_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename boost::add_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 &r1
  ) const
  {
    return range_mismatch(r0, r1, equal_to());
  }

  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    mismatch_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  ) const
  {
    return range_mismatch(r0, r1, equal_to());
  }



  ////////// operator() with predicate //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  typename result<
    mismatch_ const(
      typename boost::add_reference<SinglePassRange0>::type
    , typename boost::add_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 &r0, SinglePassRange1 &r1, BinaryPredicate pred
  ) const
  {
    return range_mismatch(r0, r1, pred);
  }

  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  typename result<
    mismatch_ const(
      typename boost::add_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    SinglePassRange0 &r0, SinglePassRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_mismatch(r0, r1, pred);
  }

  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  typename result<
    mismatch_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename boost::add_reference<SinglePassRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 &r1, BinaryPredicate pred
  ) const
  {
    return range_mismatch(r0, r1, pred);
  }

  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  typename result<
    mismatch_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_mismatch(r0, r1, pred);
  }
}; // struct mismatch_

mismatch_ const mismatch = mismatch_();



////////////////////////////////////////////////////////////////////////////////
//equal
////////////////////////////////////////////////////////////////////////////////

namespace detail{

template<class SinglePassRange0, class SinglePassRange1, class BinaryPredicate>
bool range_equal_dispatch(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryPredicate pred
, boost::single_pass_traversal_tag, boost::single_pass_traversal_tag
)
{
  using boost::end;
  
  std::pair<
      typename iterator<typename add_const<SinglePassRange0>::type>::type
    , typename iterator<typename add_const<SinglePassRange1>::type>::type
    > iter_pair = mismatch(r0, r1, pred);

  return iter_pair.first == end(r0) && iter_pair.second == end(r1);
}

template<class SinglePassRange0, class SinglePassRange1, class BinaryPredicate>
bool range_equal_dispatch(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryPredicate pred
, boost::forward_traversal_tag, boost::forward_traversal_tag
)
{
  using boost::size;
  
  if(size(r0) != size(r1)){
    return false;
  }
  return range_equal_dispatch(
      r0, r1, pred
    , boost::single_pass_traversal_tag(), boost::single_pass_traversal_tag()
    );
}

} // namespace detail

// ADL hook point
template<class SinglePassRange0, class SinglePassRange1, class BinaryPredicate>
bool range_equal(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryPredicate pred
)
{
  return detail::range_equal_dispatch(
      r0, r1, pred
    , typename boost::iterator_traversal<
        typename iterator<typename add_const<SinglePassRange0>::type>::type
      >::type()
    , typename boost::iterator_traversal<
        typename iterator<typename add_const<SinglePassRange1>::type>::type
      >::type()
    );
}

struct equal_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< equal_ const(SinglePassRange0, SinglePassRange1) >
  {
    typedef bool type;
  };
  
  template<class SinglePassRange0, class SinglePassRange1>
  struct result< equal_(SinglePassRange0, SinglePassRange1) >
    : public result< equal_ const(SinglePassRange0, SinglePassRange1) >
  { };
  
  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  struct result<
    equal_ const(SinglePassRange0, SinglePassRange1, BinaryPredicate)
  >
    : public result< equal_ const(SinglePassRange0, SinglePassRange1) >
  { };
  
  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  struct result< equal_(SinglePassRange0, SinglePassRange1, BinaryPredicate) >
    : public result< equal_ const(SinglePassRange0, SinglePassRange1) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    equal_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  ) const
  {
    return range_equal(r0, r1, equal_to());
  }



  ////////// operator() with predicate //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class BinaryPredicate
  >
  typename result<
    equal_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_equal(r0, r1, pred);
  }
}; // struct equal_

equal_ const equal = equal_();



////////////////////////////////////////////////////////////////////////////////
//search
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
typename iterator<ForwardRange0>::type
range_search(ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred)
{
  using boost::begin;
  using boost::end;
  
  return std::search(begin(r0), end(r0), begin(r1), end(r1), pred);
}

struct search_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange0, class ForwardRange1>
  struct result< search_ const(ForwardRange0, ForwardRange1) >
    : public iterator<typename boost::remove_reference<ForwardRange0>::type>
  { };
  
  template<class ForwardRange0, class ForwardRange1>
  struct result< search_(ForwardRange0, ForwardRange1) >
    : public result< search_ const(ForwardRange0, ForwardRange1) >
  { };
  
  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result< search_(ForwardRange0, ForwardRange1, BinaryPredicate) >
    : public result< search_ const(ForwardRange0, ForwardRange1) >
  { };
  
  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  struct result< search_ const(ForwardRange0, ForwardRange1, BinaryPredicate) >
    : public result< search_ const(ForwardRange0, ForwardRange1) >
  { };



  ////////// operator() //////////
  template<class ForwardRange0, class ForwardRange1>
  typename result<
    search_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type operator()(ForwardRange0 &r0, ForwardRange1 const &r1) const
  {
    return range_search(r0, r1, equal_to());
  }
  
  template<class ForwardRange0, class ForwardRange1>
  typename result<
    search_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    )
  >::type
  operator()(ForwardRange0 const &r0, ForwardRange1 const &r1) const
  {
    return range_search(r0, r1, equal_to());
  }



  ////////// operator() with predicate //////////
  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    search_ const(
      typename boost::add_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_search(r0, r1, pred);
  }
  
  template<class ForwardRange0, class ForwardRange1, class BinaryPredicate>
  typename result<
    search_ const(
      typename add_const_reference<ForwardRange0>::type
    , typename add_const_reference<ForwardRange1>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange0 const &r0, ForwardRange1 const &r1, BinaryPredicate pred
  ) const
  {
    return range_search(r0, r1, pred);
  }
}; // struct search_

search_ const search = search_();



////////////////////////////////////////////////////////////////////////////////
//search_n
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class Size, class T, class BinaryPredicate>
typename iterator<ForwardRange>::type
range_search_n(
  ForwardRange &r, Size count, T const &value, BinaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::search_n(begin(r), end(r), count, value, pred);
}

struct search_n_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class Size, class T>
  struct result< search_n_ const(ForwardRange, Size, T) >
    : public iterator<typename boost::remove_reference<ForwardRange>::type>
  { };
  
  template<class ForwardRange, class Size, class T>
  struct result< search_n_(ForwardRange, Size, T) >
    : public result< search_n_ const(ForwardRange, Size, T) >
  { };
  
  template<class ForwardRange, class Size, class T, class BinaryPredicate>
  struct result< search_n_(ForwardRange, Size, T, BinaryPredicate) >
    : public result< search_n_ const(ForwardRange, Size, T) >
  { };
  
  template<class ForwardRange, class Size, class T, class BinaryPredicate>
  struct result< search_n_ const(ForwardRange, Size, T, BinaryPredicate) >
    : public result< search_n_ const(ForwardRange, Size, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class Size, class T>
  typename result<
    search_n_ const(
      typename boost::add_reference<ForwardRange>::type
    , Size
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange &r, Size count, T const &value) const
  {
    return range_search_n(r, count, value, equal_to());
  }
  
  template<class ForwardRange, class Size, class T>
  typename result<
    search_n_ const(
      typename add_const_reference<ForwardRange>::type
    , Size
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange const &r, Size count, T const &value) const
  {
    return range_search_n(r, count, value, equal_to());
  }



  ////////// operator() with predicate //////////
  template<class ForwardRange, class Size, class T, class BinaryPredicate>
  typename result<
    search_n_ const(
      typename boost::add_reference<ForwardRange>::type
    , Size
    , typename add_const_reference<T>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange &r, Size count, T const &value, BinaryPredicate pred
  ) const
  {
    return range_search_n(r, count, value, pred);
  }
  
  template<class ForwardRange, class Size, class T, class BinaryPredicate>
  typename result<
    search_n_ const(
      typename add_const_reference<ForwardRange>::type
    , Size
    , typename add_const_reference<T>::type
    , BinaryPredicate
    )
  >::type operator()(
    ForwardRange const &r, Size count, T const &value, BinaryPredicate pred
  ) const
  {
    return range_search_n(r, count, value, pred);
  }
}; // struct search_n_

search_n_ const search_n = search_n_();



////////////////////////////////////////////////////////////////////////////////
//copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator>
OutputIterator range_copy(SinglePassRange const &r, OutputIterator result)
{
  using boost::begin;
  using boost::end;
  
  return std::copy(begin(r), end(r), result);
}

struct copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator>
  struct result< copy_ const(SinglePassRange, OutputIterator) >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange, class OutputIterator>
  struct result< copy_(SinglePassRange, OutputIterator) >
    : public result< copy_ const(SinglePassRange, OutputIterator) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator>
  typename result<
    copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    )
  >::type
  operator()(SinglePassRange const &r, OutputIterator result) const
  {
    return range_copy(r, result);
  }
}; // struct copy_

copy_ const copy = copy_();



////////////////////////////////////////////////////////////////////////////////
//copy_backward
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class BidirWriteIterator>
BidirWriteIterator range_copy_backward(
  BidirectionalRange const &r, BidirWriteIterator result
)
{
  using boost::begin;
  using boost::end;
  
  return std::copy_backward(begin(r), end(r), result);
}

struct copy_backward_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange, class BidirWriteIterator>
  struct result< copy_backward_ const(BidirectionalRange, BidirWriteIterator) >
    : public plain<BidirWriteIterator>
  { };

  template<class BidirectionalRange, class BidirWriteIterator>
  struct result< copy_backward_(BidirectionalRange, BidirWriteIterator) >
    : public result<
        copy_backward_ const(BidirectionalRange, BidirWriteIterator)
      >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange, class BidirWriteIterator>
  typename result<
    copy_backward_ const(
      typename add_const_reference<BidirectionalRange>::type
    , BidirWriteIterator
    )
  >::type operator()(
    BidirectionalRange const &r, BidirWriteIterator result
  ) const
  {
    return range_copy_backward(r, result);
  }
}; // struct copy_backward_

copy_backward_ const copy_backward = copy_backward_();



////////////////////////////////////////////////////////////////////////////////
//copy_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
OutputIterator range_copy_if(
  SinglePassRange const &r, OutputIterator result, UnaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return cry::copy_if(begin(r), end(r), result, pred);
}

struct copy_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  struct result<
    copy_if_ const(SinglePassRange, OutputIterator, UnaryPredicate)
  >
    : public plain<OutputIterator>
  { };
  
  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  struct result< copy_if_(SinglePassRange, OutputIterator, UnaryPredicate) >
    : public result<
        copy_if_ const(SinglePassRange, OutputIterator, UnaryPredicate)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  typename result<
    copy_if_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    , UnaryPredicate
    )
  >::type operator()(
    SinglePassRange const &r, OutputIterator result, UnaryPredicate pred
  ) const
  {
    return range_copy_if(r, result, pred);
  }
}; // struct copy_if_

copy_if_ const copy_if = copy_if_();



////////////////////////////////////////////////////////////////////////////////
//transform
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class SPWriteIterator, class UnaryOperation>
SPWriteIterator range_transform(
  SinglePassRange const &r, SPWriteIterator result, UnaryOperation op
)
{
  using boost::begin;
  using boost::end;
  
  return std::transform(begin(r), end(r), result, op);
}

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class SPWriteIterator
, class BinaryOperation
>
SPWriteIterator range_transform(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, SPWriteIterator result
, BinaryOperation binary_op
)
{
  using boost::begin;
  using boost::end;
  
  return std::transform(begin(r0), end(r0), begin(r1), result, binary_op);
}

struct transform_
  : public cry::lambda::enable_sig
{
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class SPWriteIterator, class UnaryOperation>
  struct result<
    transform_ const(SinglePassRange, SPWriteIterator, UnaryOperation)
  >
    : public plain<SPWriteIterator>
  { };

  template<class SinglePassRange, class SPWriteIterator, class UnaryOperation>
  struct result< transform_(SinglePassRange, SPWriteIterator, UnaryOperation) >
    : public result<
        transform_ const(SinglePassRange, SPWriteIterator, UnaryOperation)
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class SPWriteIterator
  , class BinaryOperation
  >
  struct result<
    transform_ const(
      SinglePassRange0, SinglePassRange1, SPWriteIterator, BinaryOperation
    )
  >
    : public plain<SPWriteIterator>
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class SPWriteIterator
  , class BinaryOperation
  >
  struct result<
    transform_(
      SinglePassRange0, SinglePassRange1, SPWriteIterator, BinaryOperation
    )
  >
    : public result<
        transform_ const(
          SinglePassRange0, SinglePassRange1, SPWriteIterator, BinaryOperation
        )
      >
  { };



  ////////// operator() for unary form //////////
  template<class SinglePassRange, class OutputIterator, class UnaryOperation>
  typename result<
    transform_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    , UnaryOperation
    )
  >::type operator()(
    SinglePassRange const &r, OutputIterator result, UnaryOperation op
  ) const
  {
    return range_transform(r, result, op);
  }



  ////////// operator() for binary form //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class BinaryOperation
  >
  typename result<
    transform_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    , BinaryOperation
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, BinaryOperation op
  ) const
  {
    return range_transform(r0, r1, result, op);
  }
}; // struct transform_

transform_ const transform = transform_();



////////////////////////////////////////////////////////////////////////////////
//replace
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class T>
void range_replace(
  ForwardRange &r, T const &old_value, T const &new_value
)
{
  using boost::begin;
  using boost::end;
  
  return std::replace(begin(r), end(r), old_value, new_value);
}

struct replace_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class T>
  struct result< replace_ const(ForwardRange, T) >
  {
    typedef void type;
  };

  template<class ForwardRange, class T>
  struct result< replace_(ForwardRange, T) >
    : public result< replace_ const(ForwardRange, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class T>
  typename result<
    replace_ const(
      typename boost::add_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(
    ForwardRange &r, T const &old_value, T const &new_value
  ) const
  {
    range_replace(r, old_value, new_value);
  }
}; // struct replace_

replace_ const replace = replace_();



////////////////////////////////////////////////////////////////////////////////
//replace_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class UnaryPredicate, class T>
void range_replace_if(
  ForwardRange &r, UnaryPredicate pred, T const &new_value
)
{
  using boost::begin;
  using boost::end;
  
  return std::replace_if(begin(r), end(r), pred, new_value);
}

struct replace_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class UnaryPredicate, class T>
  struct result< replace_if_ const(ForwardRange, UnaryPredicate, T) >
  {
    typedef void type;
  };

  template<class ForwardRange, class UnaryPredicate, class T>
  struct result< replace_if_(ForwardRange, UnaryPredicate, T) >
    : public result< replace_if_ const(ForwardRange, UnaryPredicate, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class UnaryPredicate, class T>
  typename result<
    replace_if_ const(
      typename boost::add_reference<ForwardRange>::type
    , UnaryPredicate
    , typename add_const_reference<T>::type
    )
  >::type operator()(
    ForwardRange &r, UnaryPredicate pred, T const &new_value
  ) const
  {
    range_replace_if(r, pred, new_value);
  }
}; // struct replace_if_

replace_if_ const replace_if = replace_if_();



////////////////////////////////////////////////////////////////////////////////
//replace_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class OutputIterator, class T>
OutputIterator range_replace_copy(
  ForwardRange const &r, OutputIterator result
, T const &old_value, T const &new_value
)
{
  using boost::begin;
  using boost::end;
  
  return std::replace_copy(begin(r), end(r), result, old_value, new_value);
}

struct replace_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class OutputIterator, class T>
  struct result< replace_copy_ const(ForwardRange, OutputIterator, T) >
    : public plain<OutputIterator>
  { };

  template<class ForwardRange, class OutputIterator, class T>
  struct result< replace_copy_(ForwardRange, OutputIterator, T) >
    : public result< replace_copy_ const(ForwardRange, OutputIterator, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class OutputIterator, class T>
  typename result<
    replace_copy_ const(
      typename add_const_reference<ForwardRange>::type
    , OutputIterator
    , typename add_const_reference<T>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(
    ForwardRange const &r, OutputIterator result
  , T const &old_value, T const &new_value
  ) const
  {
    return range_replace_copy(r, result, old_value, new_value);
  }
}; // struct replace_copy_

replace_copy_ const replace_copy = replace_copy_();



////////////////////////////////////////////////////////////////////////////////
//replace_copy_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class ForwardRange, class OutputIterator, class UnaryPredicate, class T
>
OutputIterator range_replace_copy_if(
  ForwardRange const &r, OutputIterator result
, UnaryPredicate pred, T const &new_value
)
{
  using boost::begin;
  using boost::end;
  
  return std::replace_copy_if(begin(r), end(r), result, pred, new_value);
}

struct replace_copy_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<
    class ForwardRange, class OutputIterator, class UnaryPredicate, class T
  >
  struct result<
    replace_copy_if_ const(ForwardRange, OutputIterator, UnaryPredicate, T)
  >
    : public plain<OutputIterator>
  { };

  template<
    class ForwardRange, class OutputIterator, class UnaryPredicate, class T
  >
  struct result<
    replace_copy_if_(ForwardRange, OutputIterator, UnaryPredicate, T)
  >
    : public result<
        replace_copy_if_ const(ForwardRange, OutputIterator, UnaryPredicate, T)
      >
  { };



  ////////// operator() //////////
  template<
    class ForwardRange, class OutputIterator, class UnaryPredicate, class T
  >
  typename result<
    replace_copy_if_ const(
      typename add_const_reference<ForwardRange>::type
    , OutputIterator
    , UnaryPredicate
    , typename add_const_reference<T>::type
    )
  >::type operator()(
    ForwardRange const &r, OutputIterator result
  , UnaryPredicate pred, T const &new_value
  ) const
  {
    return range_replace_copy_if(r, result, pred, new_value);
  }
}; // struct replace_copy_if_

replace_copy_if_ const repalce_copy_if = replace_copy_if_();



////////////////////////////////////////////////////////////////////////////////
//fill
////////////////////////////////////////////////////////////////////////////////

// ADL hool point
template<class SinglePassRange, class T>
void range_fill(SinglePassRange &r, T const &value)
{
  using boost::begin;
  using boost::end;
  
  std::fill(begin(r), end(r), value);
}

struct fill_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class T>
  struct result< fill_ const(SinglePassRange, T) >
  {
    typedef void type;
  };

  template<class SinglePassRange, class T>
  struct result< fill_(SinglePassRange, T) >
    : public result< fill_ const(SinglePassRange, T) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class T>
  typename result<
    fill_ const(
      typename boost::add_reference<SinglePassRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(SinglePassRange &r, T const &value) const
  {
    range_fill(r, value);
  }
}; // struct fill_

fill_ const fill = fill_();



////////////////////////////////////////////////////////////////////////////////
//generate
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class Generator>
void range_generate(SinglePassRange &r, Generator gen)
{
  using boost::begin;
  using boost::end;
  
  std::generate(begin(r), end(r), gen);
}

struct generate_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class Generator>
  struct result< generate_ const(SinglePassRange, Generator) >
  {
    typedef void type;
  };

  template<class SinglePassRange, class Generator>
  struct result< generate_(SinglePassRange, Generator) >
    : public result< generate_ const(SinglePassRange, Generator) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class Generator>
  typename result<
    generate_ const(
      typename boost::add_reference<SinglePassRange>::type
    , Generator
    )
  >::type operator()(SinglePassRange &r, Generator gen) const
  {
    range_generate(r, gen);
  }
}; // struct generate_

generate_ const generate = generate_();



////////////////////////////////////////////////////////////////////////////////
//remove
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class T>
typename iterator<ForwardRange>::type range_remove(
  ForwardRange &r, T const &value
)
{
  using boost::begin;
  using boost::end;
  
  return std::remove(begin(r), end(r), value);
}

struct remove_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class T>
  struct result< remove_ const(ForwardRange, T) >
    : public iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange, class T>
  struct result< remove_(ForwardRange, T) >
    : public result< remove_ const(ForwardRange, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class T>
  typename result<
    remove_ const(
      typename boost::add_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange &r, T const &value) const
  {
    return range_remove(r, value);
  }
}; // struct remove_

remove_ const remove = remove_();



////////////////////////////////////////////////////////////////////////////////
//remove_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class UnaryPredicate>
typename iterator<ForwardRange>::type
range_remove_if(ForwardRange &r, UnaryPredicate pred)
{
  using boost::begin;
  using boost::end;
  
  return std::remove_if(begin(r), end(r), pred);
}

struct remove_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class UnaryPredicate>
  struct result< remove_if_ const(ForwardRange, UnaryPredicate) >
    : public iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange, class UnaryPredicate>
  struct result< remove_if_(ForwardRange, UnaryPredicate) >
    : public result< remove_if_ const(ForwardRange, UnaryPredicate) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class UnaryPredicate>
  typename result<
    remove_if_ const(
      typename boost::add_reference<ForwardRange>::type
    , UnaryPredicate
    )
  >::type operator()(ForwardRange &r, UnaryPredicate pred) const
  {
    return range_remove_if(r, pred);
  }
}; // struct remove_if_

remove_if_ const remove_if = remove_if_();



////////////////////////////////////////////////////////////////////////////////
//remove_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator, class T>
OutputIterator range_remove_copy(
  SinglePassRange const &r, OutputIterator result, T const &value
)
{
  using boost::begin;
  using boost::end;
  
  return std::remove_copy(begin(r), end(r), result, value);
}

struct remove_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator, class T>
  struct result< remove_copy_ const(SinglePassRange, OutputIterator, T) >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange, class OutputIterator, class T>
  struct result< remove_copy_(SinglePassRange, OutputIterator, T) >
    : public result< remove_copy_ const(SinglePassRange, OutputIterator, T) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator, class T>
  typename result<
    remove_copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    , typename add_const_reference<T>::type
    )
  >::type operator()(
    SinglePassRange const &r, OutputIterator result, T const &value
  ) const
  {
    return range_remove_copy(r, result, value);
  }
}; // struct remove_copy_

remove_copy_ const remove_copy = remove_copy_();



////////////////////////////////////////////////////////////////////////////////
//remove_copy_if
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
OutputIterator range_remove_copy_if(
  SinglePassRange const &r, OutputIterator result, UnaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::remove_copy_if(begin(r), end(r), result, pred);
}

struct remove_copy_if_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  struct result<
    remove_copy_if_ const(SinglePassRange, OutputIterator, UnaryPredicate)
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  struct result<
    remove_copy_if_(SinglePassRange, OutputIterator, UnaryPredicate)
  >
    : public result<
        remove_copy_if_ const(SinglePassRange, OutputIterator, UnaryPredicate)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator, class UnaryPredicate>
  typename result<
    remove_copy_if_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    , UnaryPredicate
    )
  >::type operator()(
    SinglePassRange const &r, OutputIterator result, UnaryPredicate pred
  ) const
  {
    return range_remove_copy_if(r, result, pred);
  }
}; // struct remove_copy_if_

remove_copy_if_ const remove_copy_if = remove_copy_if_();



////////////////////////////////////////////////////////////////////////////////
//unique
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class BinaryPredicate>
typename iterator<ForwardRange>::type range_unique(
  ForwardRange &r, BinaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::unique(begin(r), end(r), pred);
}

struct unique_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange>
  struct result< unique_ const(ForwardRange) >
    : public iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange>
  struct result< unique_(ForwardRange) >
    : public result< unique_ const(ForwardRange) >
  { };

  template<class ForwardRange, class BinaryPredicate>
  struct result< unique_(ForwardRange, BinaryPredicate) >
    : public result< unique_ const(ForwardRange) >
  { };

  template<class ForwardRange, class BinaryPredicate>
  struct result< unique_ const(ForwardRange, BinaryPredicate) >
    : public result< unique_ const(ForwardRange) >
  { };



  ////////// operator() //////////
  template<class ForwardRange>
  typename result<
    unique_ const(typename boost::add_reference<ForwardRange>::type)
  >::type operator()(ForwardRange &r) const
  {
    return range_unique(r, equal_to());
  }



  ////////// operator() with predicate //////////
  template<class ForwardRange, class BinaryPredicate>
  typename result<
    unique_ const(
      typename boost::add_reference<ForwardRange>::type
    , BinaryPredicate
    )
  >::type operator()(ForwardRange &r, BinaryPredicate pred) const
  {
    return range_unique(r, pred);
  }
}; // struct unique_

unique_ const unique = unique_();



////////////////////////////////////////////////////////////////////////////////
//unique_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator, class BinaryPredicate>
OutputIterator range_unique_copy(
  SinglePassRange const &r, OutputIterator result, BinaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::unique_copy(begin(r), end(r), result, pred);
}

struct unique_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator>
  struct result< unique_copy_ const(SinglePassRange, OutputIterator) >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange, class OutputIterator>
  struct result< unique_copy_(SinglePassRange, OutputIterator) >
    : public result< unique_copy_ const(SinglePassRange, OutputIterator) >
  { };

  template<class SinglePassRange, class OutputIterator, class BinaryPredicate>
  struct result<
    unique_copy_(SinglePassRange, OutputIterator, BinaryPredicate)
  >
    : public result< unique_copy_ const(SinglePassRange, OutputIterator) >
  { };

  template<class SinglePassRange, class OutputIterator, class BinaryPredicate>
  struct result<
    unique_copy_ const(SinglePassRange, OutputIterator, BinaryPredicate)
  >
    : public result< unique_copy_ const(SinglePassRange, OutputIterator) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator>
  typename result<
    unique_copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    )
  >::type operator()(SinglePassRange const &r, OutputIterator result) const
  {
    return range_unique_copy(r, result, equal_to());
  }



  ////////// operator() with predicate //////////
  template<class SinglePassRange, class OutputIterator, class BinaryPredicate>
  typename result<
    unique_copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    , BinaryPredicate
    )
  >::type operator()(
    SinglePassRange const &r, OutputIterator result, BinaryPredicate pred
  ) const
  {
    return range_unique_copy(r, result, pred);
  }
}; // struct unique_copy_

unique_copy_ const unique_copy = unique_copy_();



////////////////////////////////////////////////////////////////////////////////
//reverse
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange>
void range_reverse(BidirectionalRange &r)
{
  using boost::begin;
  using boost::end;
  
  std::reverse(begin(r), end(r));
}

struct reverse_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange>
  struct result< reverse_ const(BidirectionalRange) >
  {
    typedef void type;
  };

  template<class BidirectionalRange>
  struct result< reverse_(BidirectionalRange) >
    : public result< reverse_ const(BidirectionalRange) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange>
  typename result<
    reverse_ const(typename boost::add_reference<BidirectionalRange>::type)
  >::type operator()(BidirectionalRange &r) const
  {
    range_reverse(r);
  }
}; // struct reverse_

reverse_ const reverse = reverse_();



////////////////////////////////////////////////////////////////////////////////
//reverse_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class OutputIterator>
OutputIterator range_reverse_copy(
  BidirectionalRange const &r, OutputIterator result
)
{
  using boost::begin;
  using boost::end;
  
  return std::reverse_copy(begin(r), end(r), result);
}

struct reverse_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange, class OutputIterator>
  struct result< reverse_copy_ const(BidirectionalRange, OutputIterator) >
    : public plain<OutputIterator>
  { };

  template<class BidirectionalRange, class OutputIterator>
  struct result< reverse_copy_(BidirectionalRange, OutputIterator) >
    : public result< reverse_copy_ const(BidirectionalRange, OutputIterator) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange, class OutputIterator>
  typename result<
    reverse_copy_ const(
      typename add_const_reference<BidirectionalRange>::type
    , OutputIterator
    )
  >::type operator()(BidirectionalRange const &r, OutputIterator result) const
  {
    return range_reverse_copy(r, result);
  }
}; // struct reverse_copy_

reverse_copy_ const reverse_copy = reverse_copy_();



////////////////////////////////////////////////////////////////////////////////
//rotate
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange>
void range_rotate(
  ForwardRange &r
, typename iterator<ForwardRange>::type middle
)
{
  using boost::begin;
  using boost::end;
  
  std::rotate(begin(r), middle, end(r));
}

struct rotate_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  // CONSIDER: It is possible to put the restriction that
  //           `Iterator == iterator<ForwardRange>::type' by SFINAE.
  template<class ForwardRange, class Iterator>
  struct result< rotate_ const(ForwardRange, Iterator) >
  {
    typedef void type;
  };

  template<class ForwardRange, class Iterator>
  struct result< rotate_(ForwardRange, Iterator) >
    : public result< rotate_ const(ForwardRange, Iterator) >
  { };



  ////////// operator() ///////////
  template<class ForwardRange>
  typename result<
    rotate_ const(
      typename boost::add_reference<ForwardRange>::type
    , typename iterator<ForwardRange>::type
    )
  >::type operator()(
    ForwardRange &r, typename iterator<ForwardRange>::type middle
  ) const
  {
    range_rotate(r, middle);
  }
}; // struct rotate_

rotate_ const rotate = rotate_();



////////////////////////////////////////////////////////////////////////////////
//rotate_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class OutputIterator>
OutputIterator range_rotate_copy(
  ForwardRange const &r, typename iterator<ForwardRange const>::type middle
, OutputIterator result
)
{
  using boost::begin;
  using boost::end;
  
  return std::rotate_copy(begin(r), middle, end(r), result);
}

struct rotate_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  // CONSIDER: It is possible to put the restriction that
  //           `Iterator == iterator<ForwardRange const>::type' by SFINAE.
  template<class ForwardRange, class Iterator, class OutputIterator>
  struct result< rotate_copy_ const(ForwardRange, Iterator, OutputIterator) >
    : public plain<OutputIterator>
  { };
  
  template<class ForwardRange, class Iterator, class OutputIterator>
  struct result< rotate_copy_(ForwardRange, Iterator, OutputIterator) >
    : public result<
        rotate_copy_ const(ForwardRange, Iterator, OutputIterator)
      >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class OutputIterator>
  typename result<
    rotate_copy_ const(
      typename add_const_reference<ForwardRange>::type
    , typename iterator<typename add_const<ForwardRange>::type>::type
    , OutputIterator
    )
  >::type operator()(
    ForwardRange const &r, typename iterator<ForwardRange const>::type middle
  , OutputIterator result
  ) const
  {
    return range_rotate_copy(r, middle, result);
  }
}; // struct rotate_copy_

rotate_copy_ const rotate_copy = rotate_copy_();



////////////////////////////////////////////////////////////////////////////////
//random_shuffle
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class UniformRNG>
void range_random_shuffle(RandomAccessRange &r, UniformRNG &rng)
{
  using boost::begin;
  using boost::end;
  
  cry::random_shuffle(begin(r), end(r), rng);
}

struct random_shuffle_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< random_shuffle_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< random_shuffle_(RandomAccessRange) >
    : public result< random_shuffle_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class UniformRNG>
  struct result< random_shuffle_(RandomAccessRange, UniformRNG) >
    : public result< random_shuffle_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class UniformRNG>
  struct result<
    random_shuffle_ const(RandomAccessRange, UniformRNG)
  >
    : public result< random_shuffle_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    random_shuffle_ const(
      typename boost::add_reference<RandomAccessRange>::type
    )
  >::type operator()(RandomAccessRange &r) const
  {
    boost::mt19937 rng;
    range_random_shuffle(r, rng);
  }



  ////////// operator() with RNG //////////
  template<class RandomAccessRange, class UniformRNG>
  typename result<
    random_shuffle_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , typename boost::add_reference<UniformRNG>::type
    )
  >::type operator()(RandomAccessRange &r, UniformRNG &rng) const
  {
    range_random_shuffle(r, rng);
  }
}; // struct random_shuffle_

random_shuffle_ const random_shuffle = random_shuffle_();



////////////////////////////////////////////////////////////////////////////////
//partition
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class UnaryPredicate>
typename iterator<BidirectionalRange>::type range_partition(
  BidirectionalRange &r, UnaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::partition(begin(r), end(r), pred);
}

struct partition_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange, class UnaryPredicate>
  struct result< partition_ const(BidirectionalRange, UnaryPredicate) >
    : public iterator<
        typename boost::remove_reference<BidirectionalRange>::type
      >
  { };
  
  template<class BidirectionalRange, class UnaryPredicate>
  struct result< partition_(BidirectionalRange, UnaryPredicate) >
    : public result< partition_ const(BidirectionalRange, UnaryPredicate) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange, class UnaryPredicate>
  typename result<
    partition_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , UnaryPredicate
    )
  >::type operator()(BidirectionalRange &r, UnaryPredicate pred) const
  {
    return range_partition(r, pred);
  }
}; // struct partition_

partition_ const partition = partition_();



////////////////////////////////////////////////////////////////////////////////
//stable_partition
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class UnaryPredicate>
typename iterator<BidirectionalRange>::type range_stable_partition(
  BidirectionalRange &r, UnaryPredicate pred
)
{
  using boost::begin;
  using boost::end;
  
  return std::stable_partition(begin(r), end(r), pred);
}

struct stable_partition_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange, class UnaryPredicate>
  struct result< stable_partition_ const(BidirectionalRange, UnaryPredicate) >
    : public iterator<
        typename boost::remove_reference<BidirectionalRange>::type
      >
  { };

  template<class BidirectionalRange, class UnaryPredicate>
  struct result< stable_partition_(BidirectionalRange, UnaryPredicate) >
    : public result<
        stable_partition_ const(BidirectionalRange, UnaryPredicate)
      >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange, class UnaryPredicate>
  typename result<
    stable_partition_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , UnaryPredicate
    )
  >::type operator()(BidirectionalRange &r, UnaryPredicate pred) const
  {
    return range_stable_partition(r, pred);
  }
}; // struct stable_partition_

stable_partition_ const stable_partition = stable_partition_();



////////////////////////////////////////////////////////////////////////////////
//sort
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_sort(RandomAccessRange &r, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  std::sort(begin(r), end(r), comp);
}

struct sort_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< sort_ const(RandomAccessRange) >
  {
    typedef void type;
  };
  
  template<class RandomAccessRange>
  struct result< sort_(RandomAccessRange) >
    : public result< sort_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< sort_(RandomAccessRange, Compare) >
    : public result< sort_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< sort_ const(RandomAccessRange, Compare) >
    : public result< sort_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    sort_ const(typename boost::add_reference<RandomAccessRange>::type)
  >::type operator()(RandomAccessRange &r) const
  {
    range_sort(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    sort_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_sort(r, comp);
  }
}; // struct sort_

sort_ const sort = sort_();



////////////////////////////////////////////////////////////////////////////////
//stable_sort
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_stable_sort(RandomAccessRange &r, Compare comp)
{
  std::stable_sort(begin(r), end(r), comp);
}

struct stable_sort_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< stable_sort_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< stable_sort_(RandomAccessRange) >
    : public result< stable_sort_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< stable_sort_(RandomAccessRange, Compare) >
    : public result< stable_sort_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< stable_sort_ const(RandomAccessRange, Compare) >
    : public result< stable_sort_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    stable_sort_ const(
      typename boost::add_reference<RandomAccessRange>::type
    )
  >::type operator()(RandomAccessRange &r) const
  {
    range_stable_sort(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    stable_sort_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_stable_sort(r, comp);
  }
}; // struct stable_sort_

stable_sort_ const stable_sort = stable_sort_();



////////////////////////////////////////////////////////////////////////////////
//partial_sort
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_partial_sort(
  RandomAccessRange &r, typename iterator<RandomAccessRange>::type middle
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  std::partial_sort(begin(r), middle, end(r), comp);
}

struct partial_sort_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  // CONSIDER: It is possible to put the restriction that
  //           `Iterator == iterator<RandomAccessRange>::type' by SFINAE.
  template<class RandomAccessRange, class Iterator>
  struct result< partial_sort_ const(RandomAccessRange, Iterator) >
  {
    typedef void type;
  };

  template<class RandomAccessRange, class Iterator>
  struct result< partial_sort_(RandomAccessRange, Iterator) >
    : public result< partial_sort_ const(RandomAccessRange, Iterator) >
  { };

  template<class RandomAccessRange, class Iterator, class Compare>
  struct result< partial_sort_(RandomAccessRange, Iterator, Compare) >
    : public result< partial_sort_ const(RandomAccessRange, Iterator) >
  { };

  template<class RandomAccessRange, class Iterator, class Compare>
  struct result< partial_sort_ const(RandomAccessRange, Iterator, Compare) >
    : public result< partial_sort_ const(RandomAccessRange, Iterator) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    stable_sort_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , typename iterator<RandomAccessRange>::type
    )
  >::type operator()(
    RandomAccessRange &r, typename iterator<RandomAccessRange>::type middle
  ) const
  {
    range_partial_sort(r, middle, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    stable_sort_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , typename iterator<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(
    RandomAccessRange &r, typename iterator<RandomAccessRange>::type middle
  , Compare comp
  ) const
  {
    range_partial_sort(r, middle, comp);
  }
}; // struct partial_sort_

partial_sort_ const partial_sort = partial_sort_();



////////////////////////////////////////////////////////////////////////////////
//partial_sort_copy
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class RandomAccessRange, class Compare>
typename iterator<RandomAccessRange>::type range_partial_sort_copy(
  SinglePassRange const &r0, RandomAccessRange &r1, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::partial_sort_copy(begin(r0), end(r0), begin(r1), end(r1), comp);
}

struct partial_sort_copy_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class RandomAccessRange>
  struct result< partial_sort_copy_ const(SinglePassRange, RandomAccessRange) >
    : public iterator<typename boost::remove_reference<RandomAccessRange>::type>
  { };

  template<class SinglePassRange, class RandomAccessRange>
  struct result< partial_sort_copy_(SinglePassRange, RandomAccessRange) >
    : public result<
        partial_sort_copy_ const(SinglePassRange, RandomAccessRange)
      >
  { };

  template<class SinglePassRange, class RandomAccessRange, class Compare>
  struct result<
    partial_sort_copy_(SinglePassRange, RandomAccessRange, Compare)
  >
    : public result<
        partial_sort_copy_ const(SinglePassRange, RandomAccessRange)
      >
  { };

  template<class SinglePassRange, class RandomAccessRange, class Compare>
  struct result<
    partial_sort_copy_ const(SinglePassRange, RandomAccessRange, Compare)
  >
    : public result<
        partial_sort_copy_ const(SinglePassRange, RandomAccessRange)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class RandomAccessRange>
  typename result<
    partial_sort_copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , typename boost::add_reference<RandomAccessRange>::type
    )
  >::type operator()(SinglePassRange const &r0, RandomAccessRange &r1) const
  {
    return range_partial_sort_copy(r0, r1, less());
  }



  ////////// operator() with Ordering //////////
  template<class SinglePassRange, class RandomAccessRange, class Compare>
  typename result<
    partial_sort_copy_ const(
      typename add_const_reference<SinglePassRange>::type
    , typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(
    SinglePassRange const &r0, RandomAccessRange &r1, Compare comp
  ) const
  {
    return range_partial_sort_copy(r0, r1, comp);
  }
}; // struct partial_sort_copy_

partial_sort_copy_ const partial_sort_copy = partial_sort_copy_();



////////////////////////////////////////////////////////////////////////////////
//is_sorted
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class Compare>
bool range_is_sorted(SinglePassRange const &r, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return cry::is_sorted(begin(r), end(r), comp);
}

struct is_sorted_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange>
  struct result< is_sorted_ const(SinglePassRange) >
  {
    typedef bool type;
  };

  template<class SinglePassRange>
  struct result< is_sorted_(SinglePassRange) >
    : public result< is_sorted_ const(SinglePassRange) >
  { };

  template<class SinglePassRange, class Compare>
  struct result< is_sorted_(SinglePassRange, Compare) >
    : public result< is_sorted_ const(SinglePassRange) >
  { };

  template<class SinglePassRange, class Compare>
  struct result< is_sorted_ const(SinglePassRange, Compare) >
    : public result< is_sorted_ const(SinglePassRange) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange>
  typename result<
    is_sorted_ const(
      typename add_const_reference<SinglePassRange>::type
    )
  >::type operator()(SinglePassRange const &r) const
  {
    return range_is_sorted(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class SinglePassRange, class Compare>
  typename result<
    is_sorted_ const(
      typename add_const_reference<SinglePassRange>::type
    , Compare
    )
  >::type operator()(SinglePassRange const &r, Compare comp) const
  {
    return range_is_sorted(r, comp);
  }
}; // struct is_sorted_

is_sorted_ const is_sorted = is_sorted_();



////////////////////////////////////////////////////////////////////////////////
//nth_element
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Size, class Compare>
void range_nth_element(RandomAccessRange &r, Size nth, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  std::nth_element(begin(r), begin(r) + nth, end(r), comp);
}

struct nth_element_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange, class Size>
  struct result< nth_element_ const(RandomAccessRange, Size) >
  {
    typedef void type;
  };

  template<class RandomAccessRange, class Size>
  struct result< nth_element_(RandomAccessRange, Size) >
    : public result< nth_element_ const(RandomAccessRange, Size) >
  { };

  template<class RandomAccessRange, class Size, class Compare>
  struct result< nth_element_(RandomAccessRange, Size, Compare) >
    : public result< nth_element_ const(RandomAccessRange, Size) >
  { };

  template<class RandomAccessRange, class Size, class Compare>
  struct result< nth_element_ const(RandomAccessRange, Size, Compare) >
    : public result< nth_element_ const(RandomAccessRange, Size) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange, class Size>
  typename result<
    nth_element_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Size
    )
  >::type operator()(RandomAccessRange &r, Size nth) const
  {
    range_nth_element(r, nth, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Size, class Compare>
  typename result<
    nth_element_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Size
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Size nth, Compare comp) const
  {
    range_nth_element(r, nth, comp);
  }
}; // struct nth_element_

nth_element_ const nth_element = nth_element_();



////////////////////////////////////////////////////////////////////////////////
//equal_range
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class T, class Compare>
std::pair<
  typename iterator<ForwardRange>::type, typename iterator<ForwardRange>::type
> range_equal_range(ForwardRange &r, T const &value, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return std::equal_range(begin(r), end(r), value, comp);
}

struct equal_range_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class T>
  struct result< equal_range_ const(ForwardRange, T) >
  {
    typedef std::pair<
        typename iterator<
          typename boost::remove_reference<ForwardRange>::type
        >::type
      , typename iterator<
          typename boost::remove_reference<ForwardRange>::type
        >::type
      > type;
  };

  template<class ForwardRange, class T>
  struct result< equal_range_(ForwardRange, T) >
    : public result< equal_range_ const(ForwardRange, T) >
  { };

  template<class ForwardRange, class T, class Compare>
  struct result< equal_range_(ForwardRange, T, Compare) >
    : public result< equal_range_ const(ForwardRange, T) >
  { };

  template<class ForwardRange, class T, class Compare>
  struct result< equal_range_ const(ForwardRange, T, Compare) >
    : public result< equal_range_ const(ForwardRange, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class T>
  typename result<
    equal_range_ const(
      typename boost::add_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange &r, T const &value) const
  {
    return range_equal_range(r, value, less());
  }

  template<class ForwardRange, class T>
  typename result<
    equal_range_ const(
      typename add_const_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange const &r, T const &value) const
  {
    return range_equal_range(r, value, less());
  }



  ////////// operator() with Ordering //////////
  template<class ForwardRange, class T, class Compare>
  typename result<
    equal_range_ const(
      typename boost::add_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    , Compare
    )
  >::type operator()(ForwardRange &r, T const &value, Compare comp) const
  {
    return range_equal_range(r, value, comp);
  }

  template<class ForwardRange, class T, class Compare>
  typename result<
    equal_range_ const(
      typename add_const_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    , Compare
    )
  >::type operator()(ForwardRange const &r, T const &value, Compare comp) const
  {
    return range_equal_range(r, value, comp);
  }
}; // struct equal_range_

equal_range_ const equal_range = equal_range_();



////////////////////////////////////////////////////////////////////////////////
//binary_search
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class T, class Compare>
bool range_binary_search(ForwardRange const &r, T const &value, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return std::binary_search(begin(r), end(r), value, comp);
}

struct binary_search_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange, class T>
  struct result< binary_search_ const(ForwardRange, T) >
  {
    typedef bool type;
  };

  template<class ForwardRange, class T>
  struct result< binary_search_(ForwardRange, T) >
    : public result< binary_search_ const(ForwardRange, T) >
  { };

  template<class ForwardRange, class T, class Compare>
  struct result< binary_search_(ForwardRange, T, Compare) >
    : public result< binary_search_ const(ForwardRange, T) >
  { };

  template<class ForwardRange, class T, class Compare>
  struct result< binary_search_ const(ForwardRange, T, Compare) >
    : public result< binary_search_ const(ForwardRange, T) >
  { };



  ////////// operator() //////////
  template<class ForwardRange, class T>
  typename result<
    binary_search_ const(
      typename add_const_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    )
  >::type operator()(ForwardRange const &r, T const &value) const
  {
    return range_binary_search(r, value, less());
  }

  template<class ForwardRange, class T, class Compare>
  typename result<
    binary_search_ const(
      typename add_const_reference<ForwardRange>::type
    , typename add_const_reference<T>::type
    , Compare
    )
  >::type operator()(ForwardRange const &r, T const &value, Compare comp) const
  {
    return range_binary_search(r, value, comp);
  }
}; // struct equal_range_

binary_search_ const binary_search = binary_search_();



////////////////////////////////////////////////////////////////////////////////
//merge
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class OutputIterator
, class Compare
>
OutputIterator range_merge(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, OutputIterator result
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::merge(begin(r0), end(r0), begin(r1), end(r1), result, comp);
}

struct merge_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    merge_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result< merge_(SinglePassRange0, SinglePassRange1, OutputIterator) >
    : public result<
        merge_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    merge_(SinglePassRange0, SinglePassRange1, OutputIterator, Compare)
  >
    : public result<
        merge_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    merge_ const(SinglePassRange0, SinglePassRange1, OutputIterator, Compare)
  >
    : public result<
        merge_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  typename result<
    merge_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result
  ) const
  {
    return range_merge(r0, r1, result, less());
  }



  ////////// operator() with Ordering //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  typename result<
    merge_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator, Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, Compare comp
  ) const
  {
    return range_merge(r0, r1, result, comp);
  }
}; // struct merge_

merge_ const merge = merge_();



////////////////////////////////////////////////////////////////////////////////
//inplace_merge
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class Compare>
void range_inplace_merge(
  BidirectionalRange &r, typename iterator<BidirectionalRange>::type middle
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::inplace_merge(begin(r), middle, end(r), comp);
}

struct inplace_merge_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  // CONSIDER: It is possible to put the restriction that
  //           `Iterator == iterator<BidirectionalRange>::type' by SFINAE.
  template<class BidirectionalRange, class Iterator>
  struct result< inplace_merge_ const(BidirectionalRange, Iterator) >
  {
    typedef void type;
  };

  template<class BidirectionalRange, class Iterator>
  struct result< inplace_merge_(BidirectionalRange, Iterator) >
    : public result< inplace_merge_ const(BidirectionalRange, Iterator) >
  { };

  template<class BidirectionalRange, class Iterator, class Compare>
  struct result< inplace_merge_(BidirectionalRange, Iterator, Compare) >
    : public result< inplace_merge_ const(BidirectionalRange, Iterator) >
  { };

  template<class BidirectionalRange, class Iterator, class Compare>
  struct result< inplace_merge_ const(BidirectionalRange, Iterator, Compare) >
    : public result< inplace_merge_ const(BidirectionalRange, Iterator) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange>
  typename result<
    inplace_merge_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , typename iterator<BidirectionalRange>::type
    )
  >::type operator()(
    BidirectionalRange &r, typename iterator<BidirectionalRange>::type middle
  ) const
  {
    range_inplace_merge(r, middle, less());
  }



  ////////// operator() with Ordering //////////
  template<class BidirectionalRange, class Compare>
  typename result<
    inplace_merge_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , typename iterator<BidirectionalRange>::type
    , Compare
    )
  >::type operator()(
    BidirectionalRange &r, typename iterator<BidirectionalRange>::type middle
  , Compare comp
  ) const
  {
    range_inplace_merge(r, middle, comp);
  }
}; // struct inplace_merge_

inplace_merge_ const inplace_merge = inplace_merge_();



////////////////////////////////////////////////////////////////////////////////
//includes
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange0, class SinglePassRange1, class Compare>
bool range_includes(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::includes(begin(r0), end(r0), begin(r1), end(r1), comp);
}

struct includes_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< includes_ const(SinglePassRange0, SinglePassRange1) >
  {
    typedef bool type;
  };

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< includes_(SinglePassRange0, SinglePassRange1) >
    : public result< includes_ const(SinglePassRange0, SinglePassRange1) >
  { };

  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  struct result< includes_(SinglePassRange0, SinglePassRange1, Compare) >
    : public result< includes_ const(SinglePassRange0, SinglePassRange1) >
  { };

  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  struct result< includes_ const(SinglePassRange0, SinglePassRange1, Compare) >
    : public result< includes_ const(SinglePassRange0, SinglePassRange1) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    includes_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  ) const
  {
    return range_includes(r0, r1, less());
  }



  ////////// operator() with Ordering //////////
  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  typename result<
    includes_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1, Compare comp
  ) const
  {
    return range_includes(r0, r1, comp);
  }
}; // struct includes_

includes_ const includes = includes_();



////////////////////////////////////////////////////////////////////////////////
//set_union
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class OutputIterator
, class Compare
>
OutputIterator range_set_union(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, OutputIterator result
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::set_union(begin(r0), end(r0), begin(r1), end(r1), result, comp);
}

struct set_union_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_union_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_union_(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public result<
        set_union_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_union_(SinglePassRange0, SinglePassRange1, OutputIterator, Compare)
  >
    : public result<
        set_union_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_union_ const(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_union_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  typename result<
    set_union_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result
  ) const
  {
    return range_set_union(r0, r1, result, less());
  }



  ////////// operator() with Ordering //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  typename result<
    set_union_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator, Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, Compare comp
  ) const
  {
    return range_set_union(r0, r1, result, comp);
  }
}; // struct set_union_

set_union_ const set_union = set_union_();



////////////////////////////////////////////////////////////////////////////////
//set_intersection
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class OutputIterator
, class Compare
>
OutputIterator range_set_intersection(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, OutputIterator result
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::set_intersection(
      begin(r0), end(r0), begin(r1), end(r1), result, comp
    );
}

struct set_intersection_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_intersection_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_intersection_(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public result<
        set_intersection_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_intersection_(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_intersection_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_intersection_ const(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_intersection_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  typename result<
    set_intersection_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result
  ) const
  {
    return range_set_intersection(r0, r1, result, less());
  }



  ////////// operator() with Ordering //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  typename result<
    set_intersection_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator, Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, Compare comp
  ) const
  {
    return range_set_intersection(r0, r1, result, comp);
  }
}; // struct set_intersection_

set_intersection_ const set_intersection = set_intersection_();



////////////////////////////////////////////////////////////////////////////////
//set_difference
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class OutputIterator
, class Compare
>
OutputIterator range_set_difference(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, OutputIterator result
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::set_difference(
      begin(r0), end(r0), begin(r1), end(r1), result, comp
    );
}

struct set_difference_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_difference_ const(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_difference_(SinglePassRange0, SinglePassRange1, OutputIterator)
  >
    : public result<
        set_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_difference_(SinglePassRange0, SinglePassRange1, OutputIterator, Compare)
  >
    : public result<
        set_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_difference_ const(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  typename result<
    set_difference_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result
  ) const
  {
    return range_set_difference(r0, r1, result, less());
  }



  ////////// operator() with Ordering //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  typename result<
    set_difference_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator, Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, Compare comp
  ) const
  {
    return range_set_difference(r0, r1, result, comp);
  }
}; // struct set_difference_

set_difference_ const set_difference = set_difference_();



////////////////////////////////////////////////////////////////////////////////
//set_symmetric_difference
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<
  class SinglePassRange0, class SinglePassRange1, class OutputIterator
, class Compare
>
OutputIterator range_set_symmetric_difference(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, OutputIterator result
, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::set_symmetric_difference(
      begin(r0), end(r0), begin(r1), end(r1), result, comp
    );
}

struct set_symmetric_difference_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_symmetric_difference_ const(
      SinglePassRange0, SinglePassRange1, OutputIterator
    )
  >
    : public plain<OutputIterator>
  { };

  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  struct result<
    set_symmetric_difference_(
      SinglePassRange0, SinglePassRange1, OutputIterator
    )
  >
    : public result<
        set_symmetric_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_symmetric_difference_(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_symmetric_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };

  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  struct result<
    set_symmetric_difference_ const(
      SinglePassRange0, SinglePassRange1, OutputIterator, Compare
    )
  >
    : public result<
        set_symmetric_difference_ const(
          SinglePassRange0, SinglePassRange1, OutputIterator
        )
      >
  { };
  
  
  
  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1, class OutputIterator>
  typename result<
    set_symmetric_difference_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result
  ) const
  {
    return range_set_symmetric_difference(r0, r1, result, less());
  }



  ////////// operator() with Ordering //////////
  template<
    class SinglePassRange0, class SinglePassRange1, class OutputIterator
  , class Compare
  >
  typename result<
    set_symmetric_difference_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , OutputIterator, Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  , OutputIterator result, Compare comp
  ) const
  {
    return range_set_symmetric_difference(r0, r1, result, comp);
  }
}; // struct set_symmetric_difference_

set_symmetric_difference_ const set_symmetric_difference
  = set_symmetric_difference_();



////////////////////////////////////////////////////////////////////////////////
//push_heap
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_push_heap(RandomAccessRange &r, Compare comp)
{
  std::push_heap(r, comp);
}

struct push_heap_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< push_heap_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< push_heap_(RandomAccessRange) >
    : public result< push_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< push_heap_(RandomAccessRange, Compare) >
    : public result< push_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< push_heap_ const(RandomAccessRange, Compare) >
    : public result< push_heap_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    push_heap_ const(typename boost::add_reference<RandomAccessRange>::type)
  >::type operator()(RandomAccessRange &r) const
  {
    range_push_heap(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    push_heap_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_push_heap(r, comp);
  }
}; // struct push_heap_

push_heap_ const push_heap = push_heap_();



////////////////////////////////////////////////////////////////////////////////
//pop_heap
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_pop_heap(RandomAccessRange &r, Compare comp)
{
  std::pop_heap(r, comp);
}

struct pop_heap_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< pop_heap_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< pop_heap_(RandomAccessRange) >
    : public result< pop_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< pop_heap_(RandomAccessRange, Compare) >
    : public result< pop_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< pop_heap_ const(RandomAccessRange, Compare) >
    : public result< pop_heap_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    pop_heap_ const(typename boost::add_reference<RandomAccessRange>::type)
  >::type operator()(RandomAccessRange &r) const
  {
    range_pop_heap(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    pop_heap_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_pop_heap(r, comp);
  }
}; // struct pop_heap_

pop_heap_ const pop_heap = pop_heap_();



////////////////////////////////////////////////////////////////////////////////
//make_heap
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_make_heap(RandomAccessRange &r, Compare comp)
{
  std::make_heap(r, comp);
}

struct make_heap_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< make_heap_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< make_heap_(RandomAccessRange) >
    : public result< make_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< make_heap_(RandomAccessRange, Compare) >
    : public result< make_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< make_heap_ const(RandomAccessRange, Compare) >
    : public result< make_heap_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result<
    make_heap_ const(typename boost::add_reference<RandomAccessRange>::type)
  >::type operator()(RandomAccessRange &r) const
  {
    range_make_heap(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    make_heap_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_make_heap(r, comp);
  }
}; // struct make_heap_

make_heap_ const make_heap = make_heap_();



////////////////////////////////////////////////////////////////////////////////
//sort_heap
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class RandomAccessRange, class Compare>
void range_sort_heap(RandomAccessRange &r, Compare comp)
{
  std::sort_heap(r, comp);
}

struct sort_heap_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class RandomAccessRange>
  struct result< sort_heap_ const(RandomAccessRange) >
  {
    typedef void type;
  };

  template<class RandomAccessRange>
  struct result< sort_heap_(RandomAccessRange) >
    : public result< sort_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< sort_heap_(RandomAccessRange, Compare) >
    : public result< sort_heap_ const(RandomAccessRange) >
  { };

  template<class RandomAccessRange, class Compare>
  struct result< sort_heap_ const(RandomAccessRange, Compare) >
    : public result< sort_heap_ const(RandomAccessRange) >
  { };



  ////////// operator() //////////
  template<class RandomAccessRange>
  typename result< sort_heap_ const(RandomAccessRange &) >::type
  operator()(RandomAccessRange &r) const
  {
    range_sort_heap(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class RandomAccessRange, class Compare>
  typename result<
    sort_heap_ const(
      typename boost::add_reference<RandomAccessRange>::type
    , Compare
    )
  >::type operator()(RandomAccessRange &r, Compare comp) const
  {
    range_sort_heap(r, comp);
  }
}; // struct sort_heap_

sort_heap_ const sort_heap = sort_heap_();



////////////////////////////////////////////////////////////////////////////////
//min_element
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class Compare>
typename iterator<ForwardRange>::type
range_min_element(ForwardRange &r, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return std::min_element(begin(r), end(r), comp);
}

struct min_element_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange>
  struct result< min_element_ const(ForwardRange) >
    : iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange>
  struct result< min_element_(ForwardRange) >
    : result< min_element_ const(ForwardRange) >
  { };

  template<class ForwardRange, class Compare>
  struct result< min_element_(ForwardRange, Compare) >
    : result< min_element_ const(ForwardRange) >
  { };

  template<class ForwardRange, class Compare>
  struct result< min_element_ const(ForwardRange, Compare) >
    : result< min_element_ const(ForwardRange) >
  { };



  ////////// operator() //////////
  template<class ForwardRange>
  typename result<
    min_element_ const(typename boost::add_reference<ForwardRange>::type)
  >::type operator()(ForwardRange &r) const
  {
    return range_min_element(r, less());
  }

  template<class ForwardRange>
  typename result<
    min_element_ const(
      typename add_const_reference<ForwardRange>::type
    )
  >::type operator()(ForwardRange const &r) const
  {
    return range_min_element(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class ForwardRange, class Compare>
  typename result<
    min_element_ const(
      typename boost::add_reference<ForwardRange>::type
    , Compare
    )
  >::type operator()(ForwardRange &r, Compare comp) const
  {
    return range_min_element(r, comp);
  }

  template<class ForwardRange, class Compare>
  typename result<
    min_element_ const(
      typename add_const_reference<ForwardRange>::type
    , Compare
    )
  >::type operator()(ForwardRange const &r, Compare comp) const
  {
    return range_min_element(r, comp);
  }
}; // struct min_element_

min_element_ const min_element = min_element_();



////////////////////////////////////////////////////////////////////////////////
//max_element
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class ForwardRange, class Compare>
typename iterator<ForwardRange>::type range_max_element(
  ForwardRange &r, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::max_element(begin(r), end(r), comp);
}

struct max_element_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class ForwardRange>
  struct result< max_element_ const(ForwardRange) >
    : iterator<typename boost::remove_reference<ForwardRange>::type>
  { };

  template<class ForwardRange>
  struct result< max_element_(ForwardRange) >
    : result< max_element_ const(ForwardRange) >
  { };

  template<class ForwardRange, class Compare>
  struct result< max_element_(ForwardRange, Compare) >
    : result< max_element_ const(ForwardRange) >
  { };

  template<class ForwardRange, class Compare>
  struct result< max_element_ const(ForwardRange, Compare) >
    : result< max_element_ const(ForwardRange) >
  { };



  ////////// operator() //////////
  template<class ForwardRange>
  typename result<
    max_element_ const(typename boost::add_reference<ForwardRange>::type)
  >::type operator()(ForwardRange &r) const
  {
    return range_max_element(r, less());
  }

  template<class ForwardRange>
  typename result<
    max_element_ const(
      typename add_const_reference<ForwardRange>::type
    )
  >::type operator()(ForwardRange const &r) const
  {
    return range_max_element(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class ForwardRange, class Compare>
  typename result<
    max_element_ const(
      typename boost::add_reference<ForwardRange>::type
    , Compare
    )
  >::type operator()(ForwardRange &r, Compare comp) const
  {
    return range_max_element(r, comp);
  }

  template<class ForwardRange, class Compare>
  typename result<
    max_element_ const(
      typename add_const_reference<ForwardRange>::type
    , Compare
    )
  >::type operator()(ForwardRange const &r, Compare comp) const
  {
    return range_max_element(r, comp);
  }
}; // struct max_element_

max_element_ const max_element = max_element_();



////////////////////////////////////////////////////////////////////////////////
//lexicographical_compare
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange0, class SinglePassRange1, class Compare>
bool range_lexicographical_compare(
  SinglePassRange0 const &r0, SinglePassRange1 const &r1, Compare comp
)
{
  using boost::begin;
  using boost::end;
  
  return std::lexicographical_compare(
      begin(r0), end(r0), begin(r1), end(r1), comp
    );
}

struct lexicographical_compare_
  : public cry::lambda::enable_sig
{
  ////////// Return type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange0, class SinglePassRange1>
  struct result<
    lexicographical_compare_ const(SinglePassRange0, SinglePassRange1)
  >
  {
    typedef bool type;
  };

  template<class SinglePassRange0, class SinglePassRange1>
  struct result< lexicographical_compare_(SinglePassRange0, SinglePassRange1) >
    : public result<
        lexicographical_compare_ const(SinglePassRange0, SinglePassRange1)
      >
  { };

  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  struct result<
    lexicographical_compare_(SinglePassRange0, SinglePassRange1, Compare)
  >
    : public result<
        lexicographical_compare_ const(SinglePassRange0, SinglePassRange1)
      >
  { };

  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  struct result<
    lexicographical_compare_ const(SinglePassRange0, SinglePassRange1, Compare)
  >
    : public result<
        lexicographical_compare_ const(SinglePassRange0, SinglePassRange1)
      >
  { };



  ////////// operator() //////////
  template<class SinglePassRange0, class SinglePassRange1>
  typename result<
    lexicographical_compare_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1
  ) const
  {
    return range_lexicographical_compare(r0, r1, less());
  }



  ////////// operator() with Ordering //////////
  template<class SinglePassRange0, class SinglePassRange1, class Compare>
  typename result<
    lexicographical_compare_ const(
      typename add_const_reference<SinglePassRange0>::type
    , typename add_const_reference<SinglePassRange1>::type
    , Compare
    )
  >::type operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1, Compare comp
  ) const
  {
    return range_lexicographical_compare(r0, r1, comp);
  }
}; // struct lexicographical_compare_

lexicographical_compare_ const lexicographical_compare
  = lexicographical_compare_();



////////////////////////////////////////////////////////////////////////////////
//next_permutation
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class Compare>
bool range_next_permutation(BidirectionalRange &r, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return std::next_permutation(begin(r), end(r), comp);
}

struct next_permutation_
  : public cry::lambda::enable_sig
{
  ////////// Result type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange>
  struct result< next_permutation_ const(BidirectionalRange) >
  {
    typedef bool type;
  };

  template<class BidirectionalRange>
  struct result< next_permutation_(BidirectionalRange) >
    : public result< next_permutation_ const(BidirectionalRange) >
  { };

  template<class BidirectionalRange, class Compare>
  struct result< next_permutation_(BidirectionalRange, Compare) >
    : public result< next_permutation_ const(BidirectionalRange) >
  { };

  template<class BidirectionalRange, class Compare>
  struct result< next_permutation_ const(BidirectionalRange, Compare) >
    : public result< next_permutation_ const(BidirectionalRange) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange>
  typename result<
    next_permutation_ const(
      typename boost::add_reference<BidirectionalRange>::type
    )
  >::type operator()(BidirectionalRange &r) const
  {
    return range_next_permutation(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class BidirectionalRange, class Compare>
  typename result<
    next_permutation_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , Compare
    )
  >::type operator()(BidirectionalRange &r, Compare comp) const
  {
    return range_next_permutation(r, comp);
  }
}; // struct next_permutation_

next_permutation_ const next_permutation = next_permutation_();



////////////////////////////////////////////////////////////////////////////////
//prev_permutation
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class BidirectionalRange, class Compare>
bool range_prev_permutation(BidirectionalRange &r, Compare comp)
{
  using boost::begin;
  using boost::end;
  
  return std::prev_permutation(begin(r), end(r), comp);
}

struct prev_permutation_
  : public cry::lambda::enable_sig
{
  ////////// Result type calculation //////////
  template<class FArgs>
  struct result;

  template<class BidirectionalRange>
  struct result< prev_permutation_ const(BidirectionalRange) >
  {
    typedef bool type;
  };

  template<class BidirectionalRange>
  struct result< prev_permutation_(BidirectionalRange) >
    : public result< prev_permutation_ const(BidirectionalRange) >
  { };

  template<class BidirectionalRange, class Compare>
  struct result< prev_permutation_(BidirectionalRange, Compare) >
    : public result< prev_permutation_ const(BidirectionalRange) >
  { };

  template<class BidirectionalRange, class Compare>
  struct result< prev_permutation_ const(BidirectionalRange, Compare) >
    : public result< prev_permutation_ const(BidirectionalRange) >
  { };



  ////////// operator() //////////
  template<class BidirectionalRange>
  typename result<
    prev_permutation_ const(
      typename boost::add_reference<BidirectionalRange>::type
    )
  >::type operator()(BidirectionalRange &r) const
  {
    return range_prev_permutation(r, less());
  }



  ////////// operator() with Ordering //////////
  template<class BidirectionalRange, class Compare>
  typename result<
    prev_permutation_ const(
      typename boost::add_reference<BidirectionalRange>::type
    , Compare
    )
  >::type operator()(BidirectionalRange &r, Compare comp) const
  {
    return range_prev_permutation(r, comp);
  }
}; // struct prev_permutation_

prev_permutation_ const prev_permutation = prev_permutation_();



////////////////////////////////////////////////////////////////////////////////
// joint
////////////////////////////////////////////////////////////////////////////////

// ADL hook point
template<class SinglePassRange, class OutputIterator>
OutputIterator range_joint(SinglePassRange const &r, OutputIterator result)
{
  // View a range of ranges as a single range. 
  using boost::begin;
  using boost::end;

  typedef typename iterator<typename add_const<SinglePassRange>::type>::type
    outer_const_iterator;
  
  for(outer_const_iterator it0(begin(r)); it0 != end(r); ++it0)
  {
    typedef typename boost::range_value<outer_const_iterator>::type
      outer_value_type;
    typedef typename iterator<outer_value_type>::type
      inner_const_iterator;
    for(inner_const_iterator it1(begin(*it0)); it1 != end(*it0); ++it1){
      *result++ = *it1;
    }
  }

  return result;
}

struct joint_
  : public cry::lambda::enable_sig
{
  ////////// Result type calculation //////////
  template<class FArgs>
  struct result;

  template<class SinglePassRange, class OutputIterator>
  struct result< joint_ const(SinglePassRange, OutputIterator) >
    : public plain<OutputIterator>
  { };
  
  template<class SinglePassRange, class OutputIterator>
  struct result< joint_(SinglePassRange, OutputIterator) >
    : public result< joint_ const(SinglePassRange, OutputIterator) >
  { };



  ////////// operator() //////////
  template<class SinglePassRange, class OutputIterator>
  typename result<
    joint_ const(
      typename add_const_reference<SinglePassRange>::type
    , OutputIterator
    )
  >::type operator()(SinglePassRange const &r, OutputIterator result) const
  {
    return range_joint(r, result);
  }
}; // struct joint_

joint_ const joint = joint_();

}} // namespace cry::range


#endif // !defiend(CRY_RANGE_ALGORITHM_HPP_INCLUDE_GUARD)
