// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_RANGE_COUNTED_RANGE_HPP_INCLUDE_GUARD)
#define CRY_RANGE_COUNTED_RANGE_HPP_INCLUDE_GUARD

#include <cstddef>
#include <limits>

#include <boost/config.hpp>

#include <boost/mpl/bool.hpp>
#include <boost/mpl/less.hpp>
#include <boost/mpl/sizeof.hpp>
#include <boost/mpl/eval_if.hpp>

#include <boost/type_traits/is_fundamental.hpp>

#include <boost/tuple/tuple.hpp>

#include <boost/iterator/iterator_categories.hpp>
#include <boost/iterator/iterator_traits.hpp>
#include <boost/iterator/iterator_facade.hpp>

#include <cry/type_traits/plain.hpp>
#include <cry/type_traits/function_traits.hpp>

#include <cry/move/is_move_convertible.hpp>
#include <cry/container/container_traits.hpp>
#include <cry/lambda/enable_sig.hpp>

#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable:4675)
#endif


namespace cry{ namespace range{

namespace counted_range_detail{

template<class Incrementable>
struct traversal
  : public boost::mpl::eval_if_c<
      std::numeric_limits<Incrementable>::is_specialized
    , boost::mpl::identity<boost::random_access_traversal_tag>
    , boost::iterator_traversal<Incrementable>
    >
{};

template<class Incrementable>
struct difference
  : public boost::mpl::eval_if_c<
      std::numeric_limits<Incrementable>::is_specialized
    , boost::mpl::eval_if<
        boost::mpl::less<
          typename boost::mpl::sizeof_<Incrementable>
        , typename boost::mpl::sizeof_<int>
        >
      , boost::mpl::identity<int>
      , boost::mpl::identity<Incrementable>
      >
    , boost::mpl::identity<Incrementable>
    >
{};

} // namespace counted_range_detail

template<class Incrementable>
class counted_range
{
public:
  typedef Incrementable value_type;
  typedef value_type const &reference;
  typedef reference const_reference;
  typedef std::size_t size_type; // CONSIDER: size_type
  typedef typename counted_range_detail::difference<
      Incrementable
    >::type difference_type;

  class iterator
    : public boost::iterator_facade<
        iterator
      , value_type
      , typename counted_range_detail::traversal<value_type>::type
      , reference
      , difference_type
      >
  {
    friend class boost::iterator_core_access;

  public:
    iterator()
      : value_()
    {}

    iterator(value_type const &value)
      : value_(value)
    {}

    iterator(iterator const &r)
      : value_(r.value_)
    {}

    iterator &operator=(iterator const &r)
    {
      iterator tmp(r);
      tmp.swap(*this);
      return *this;
    }

    void swap(iterator &r)
    {
      using std::swap;
      swap(value_, r.value_);
    }

    friend void swap(iterator &x, iterator &y)
    {
      x.swap(y);
    }
    
  private:
    reference dereference() const
    {
      return value_;
    }

    bool equal(iterator const &rhs) const
    {
      return value_ == rhs.value_;
    }

    void increment()
    {
      ++value_;
    }

    void decrement()
    {
      --value_;
    }

    void advance(difference_type n)
    {
      value_ += n;
    }

    difference_type distance_to(iterator const &r) const
    {
      return r.value_ - value_;
    }

  private:
    value_type value_;
  }; // class iterator
  
  typedef iterator const_iterator;
  
  counted_range()
    : first_(), last_()
  { }
  
  counted_range(Incrementable const &first, Incrementable const &last)
    : first_(first), last_(last)
  { }

  counted_range(counted_range const &r)
    : first_(r.first_)
    , last_(r.last_)
  { }

  const_iterator begin() const
  {
    return const_iterator(first_);
  }

  const_iterator end() const
  {
    return const_iterator(last_);
  }

  size_type size() const
  {
    return static_cast<size_type>(
        std::distance(const_iterator(first_), const_iterator(last_))
      );
  }

  void swap(counted_range &r)
  {
    using std::swap;

    swap(first_, r.first_);
    swap(last_, r.last_);
  }
  
  friend void swap(counted_range &lhs, counted_range &rhs)
  {
    lhs.swap(rhs);
  }

private:
  Incrementable first_;
  Incrementable last_;
}; // class counted_range



struct counted_
  : public cry::lambda::enable_sig
{
  ////////// Result type calculation //////////
  template<class FArgs>
  struct result
  {
    BOOST_STATIC_ASSERT((cry::arity<FArgs>::value == 2));
    typedef typename cry::plain<typename arg1<FArgs>::type>::type arg1_plain;
    typedef typename cry::plain<typename arg2<FArgs>::type>::type arg2_plain;
    BOOST_STATIC_ASSERT((boost::is_same<arg1_plain, arg2_plain>::value));
    
    typedef counted_range<arg1_plain> type;
  };


  
  /////////// operator() //////////
  template<class Incrementable>
  counted_range<typename cry::plain<Incrementable>::type>
  operator()(Incrementable first, Incrementable last) const
  {
    return counted_range<
        typename cry::plain<Incrementable>::type
      >(first, last);
  }
}; // struct counted_

counted_ const counted = counted_();

}} // namespace cry::range


#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // !defined(CRY_RANGE_COUNTED_RANGE_HPP_INCLUDE_GUARD)
