// (C) Copyright Cryolite 2004.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#if !defined(CRY_RANGE_TRANSFORMED_RANGE_HPP_INCLUDE_GUARD)
#define CRY_RANGE_TRANSFORMED_RANGE_HPP_INCLUDE_GUARD

#include <boost/type_traits/is_const.hpp>
#include <boost/type_traits/is_class.hpp>
#include <boost/type_traits/remove_reference.hpp>
#include <boost/type_traits/remove_cv.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/static_assert.hpp>

#include <boost/utility/result_of.hpp>

#include <boost/tuple/tuple.hpp>

#include <boost/iterator/iterator_facade.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/iterator/iterator_traits.hpp>

// Boost.Range meta-functions
#include <boost/range/result_iterator.hpp>
#include <boost/range/const_iterator.hpp>
#include <boost/range/difference_type.hpp>
#include <boost/range/size_type.hpp>

// Boost.Range functions
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>
#include <boost/range/size.hpp>
#include <boost/range/empty.hpp>

#include <cry/type_traits/plain.hpp>
#include <cry/type_traits/add_const.hpp>
#include <cry/lambda/enable_sig.hpp>


namespace cry{ namespace range{

// Note: The constness of "SinglePassRange" affects the writability of
//       "transform_range."
template<class UnaryFunction, class SinglePassRange>
class transformed_range
{
  typedef typename boost::range_result_iterator<SinglePassRange>::type
    base_iterator;
  typedef typename boost::range_const_iterator<SinglePassRange>::type
    base_const_iterator;
  typedef typename boost::iterator_reference<base_iterator>::type
    base_reference;
  typedef typename boost::iterator_reference<base_const_iterator>::type
    base_const_reference;

public:
  typedef typename boost::result_of< UnaryFunction(base_reference) >::type
    reference;
  typedef typename boost::result_of< UnaryFunction(base_const_reference) >::type
    const_reference;
  typedef typename boost::remove_cv<
      typename boost::remove_reference<reference>::type
    >::type value_type;
  typedef boost::transform_iterator<UnaryFunction, base_iterator, reference>
    iterator;
  typedef boost::transform_iterator<
      UnaryFunction, base_const_iterator, const_reference
    > const_iterator;
  typedef typename boost::range_difference<SinglePassRange>::type
    difference_type;
  typedef typename boost::range_size<SinglePassRange>::type
    size_type;

  transformed_range(SinglePassRange &r, UnaryFunction f)
    : first_(boost::begin(r), f)
    , last_(boost::end(r), f)
  { }

  explicit transformed_range(SinglePassRange &r)
    : first_(boost::begin(r))
    , last_(boost::end(r))
  {
    // This assertion avoids the case where "UnaryFunction" is a pointer to a
    // ordinary function.
    BOOST_STATIC_ASSERT((boost::is_class<UnaryFunction>::value));
  }

  transformed_range(transformed_range const &r)
    : first_(r.first_)
    , last_(r.last_)
  {}

  transformed_range &operator=(transformed_range const &r)
  {
    transformed_range tmp(r);
    tmp.swap(*this);
    return *this;
  }
  
  iterator begin()
  {
    return first_;
  }

  const_iterator begin() const
  {
    return first_;
  }

  iterator end()
  {
    return last_;
  }

  const_iterator end() const
  {
    return last_;
  }

  size_type size() const
  {
    return std::distance(first_, last_);
  }

  UnaryFunction functor() const
  {
    return first_.functor();
  }

  void swap(transformed_range &x)
  {
    using std::swap;
    swap(first_, x.first_);
    swap(last_, x.last_);
  }

  friend void swap(transformed_range &x, transformed_range &y)
  {
    x.swap(y);
  }
  
private:
  iterator first_;
  iterator last_;
}; // class transformed_range





// Note: The constnesses of "SinglePassRange0" and "SinglePassRange1" affect
//       the writability of "binary_transform_range."
template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
class binary_transformed_range
{
  typedef typename boost::range_result_iterator<SinglePassRange0>::type
    base_iterator0;
  typedef typename boost::range_const_iterator<SinglePassRange0>::type
    base_const_iterator0;
  typedef typename boost::iterator_reference<base_iterator0>::type
    base_reference0;
  typedef typename boost::iterator_reference<base_const_iterator0>::type
    base_const_reference0;

  typedef typename boost::range_result_iterator<SinglePassRange1>::type
    base_iterator1;
  typedef typename boost::range_const_iterator<SinglePassRange1>::type
    base_const_iterator1;
  typedef typename boost::iterator_reference<base_iterator1>::type
    base_reference1;
  typedef typename boost::iterator_reference<base_const_iterator1>::type
    base_const_reference1;

public:
  typedef typename boost::result_of<
      BinaryFunction(base_reference0, base_reference1)
    >::type reference;
  typedef typename boost::result_of<
      BinaryFunction(base_const_reference0, base_const_reference1)
    >::type const_reference;
  typedef typename boost::remove_cv<
      typename boost::remove_reference<reference>::type
    >::type value_type;
  typedef typename boost::range_difference<SinglePassRange0>::type
    difference_type; // TODO: Stricter difference_type.
  typedef typename boost::range_size<SinglePassRange0>::type
    size_type; // TODO: Stricter size_type.

  class const_iterator; // Forward declaration.
  
  class iterator
    : public boost::iterator_facade<
        iterator
      , value_type
      , boost::single_pass_traversal_tag // TODO: Stricter traversal tag.
      , reference
      , difference_type
      >
  {
    friend class boost::iterator_core_access;
    friend class const_iterator;

    typedef binary_transformed_range enclosing;

  public:
    typedef typename enclosing::reference reference;
    typedef typename enclosing::difference_type difference_type;
    
    iterator()
      : it0_(), it1_(), pred_()
    { }
    
    iterator(
      base_iterator0 const &it0, base_iterator1 const &it1, BinaryFunction pred
    )
      : it0_(it0), it1_(it1), pred_(pred)
    { }

    // Compiler-generated cctor, dtor and assignment are fine.

  private:
    reference dereference() const
    {
      return pred_(*it0_, *it1_);
    }

    template<class Iterator>
    bool equal(Iterator const &rhs) const
    {
      return it0_ == rhs.it0_ && it1_ == rhs.it1_;
    }

    void increment()
    {
      ++it0_;
      ++it1_;
    }

    void decrement()
    {
      --it0_;
      --it1_;
    }

    void advance(difference_type n)
    {
      it0_ += n;
      it1_ += n;
    }

    // TODO: Restrict to "iterator" and "const_iterator."
    template<class Iterator>
    difference_type distance_to(Iterator const &rhs) const
    {
      // CONSIDER: How to treat the case "diff0 != diff1" ?
      difference_type diff0 = rhs.it0_ - it0_;
      difference_type diff1 = rhs.it1_ - it1_;
      return diff0 == diff1 ?
          diff0 :
          std::numeric_limits<difference_type>::max();
    }

    base_iterator0 it0_;
    base_iterator1 it1_;
    BinaryFunction pred_;
  }; // class iterator
  
  class const_iterator
    : public boost::iterator_facade<
        const_iterator
      , value_type
      , boost::single_pass_traversal_tag // TODO: Stricter traversal tag.
      , const_reference
      , difference_type
      >
  {
    friend class boost::iterator_core_access;
    friend class iterator;

    typedef binary_transformed_range enclosing;

  public:
    typedef typename enclosing::const_reference reference;
    typedef typename enclosing::difference_type difference_type;
    
    const_iterator()
      : it0_(), it1_(), pred_()
    { }
    
    const_iterator(
      base_iterator0 const &it0, base_iterator1 const &it1, BinaryFunction pred
    )
      : it0_(it0), it1_(it1), pred_(pred)
    { }

    // Compiler-generated cctor, dtor and assignment are fine.

  private:
    reference dereference() const
    {
      return pred_(*it0_, *it1_);
    }

    template<class Iterator>
    bool equal(Iterator const &rhs) const
    {
      return it0_ == rhs.it0_ && it1_ == rhs.it1_;
    }

    void increment()
    {
      ++it0_;
      ++it1_;
    }

    void decrement()
    {
      --it0_;
      --it1_;
    }

    void advance(difference_type n)
    {
      it0_ += n;
      it1_ += n;
    }

    // TODO: Restrict to "iterator" and "const_iterator."
    template<class Iterator>
    difference_type distance_to(Iterator const &rhs) const
    {
      // CONSIDER: How to treat the case "diff0 != diff1" ?
      difference_type diff0 = rhs.it0_ - it0_;
      difference_type diff1 = rhs.it1_ - it1_;
      return diff0 == diff1 ?
          diff0 :
          std::numeric_limits<difference_type>::max();
    }

    base_iterator0 it0_;
    base_iterator1 it1_;
    BinaryFunction pred_;
  }; // class const_iterator
  
  binary_transformed_range(
    SinglePassRange0 &r0, SinglePassRange1 &r1, BinaryFunction f
  )
    : r0_(r0), r1_(r1), f_(f)
  { }

  binary_transformed_range(SinglePassRange0 &r0, SinglePassRange1 &r1)
    : r0_(r0), r1_(r1), f_()
  {
    // This assertion avoids the case where "UnaryFunction" is a pointer to a
    // ordinary function.
    BOOST_STATIC_ASSERT((boost::is_class<BinaryFunction>::value));
  }

  // Compiler-generated cctor and dtor are fine.
  // dctor is not declared. Compiler-generated assignment is implicitly declared
  // but never defined because of a reference member.
  
  iterator begin()
  {
    using boost::begin;
    
    return iterator(begin(r0_), begin(r1_), f_);
  }

  const_iterator begin() const
  {
    using boost::const_begin;
    
    return const_iterator(const_begin(r0_), const_begin(r1_), f_);
  }

  iterator end()
  {
    // CONSIDER: Algorithm.
    using boost::begin;
    using boost::end;
    using boost::size;

    base_iterator0 e0(begin(r0_));
    base_iterator1 e1(begin(r1_));
    if(size(r1_) < size(r0_)){
      e1 = end(r1_);
      std::advance(e0, size(r1_));
    }
    else{
      e0 = end(r0_);
      std::advance(e1, size(r0_));
    }
    
    return iterator(e0, e1, f_);
  }

  const_iterator end() const
  {
    // CONSIDER: Algorithm.
    using boost::const_begin;
    using boost::const_end;
    using boost::size;

    base_iterator0 e0(const_begin(r0_));
    base_iterator1 e1(const_begin(r1_));
    if(size(r1_) < size(r0_)){
      e1 = const_end(r1_);
      std::advance(e0, size(r1_));
    }
    else{
      e0 = const_end(r0_);
      std::advance(e1, size(r0_));
    }
    
    return const_iterator(e0, e1, f_);
  }

  size_type size() const
  {
    using boost::size;
    
    return std::min(size(r0_), size(r1_));
  }

  BinaryFunction functor() const
  {
    return f_;
  }

private:
  SinglePassRange0 &r0_;
  SinglePassRange1 &r1_;
  BinaryFunction f_;
}; // class binary_transformed_range





struct transformed_
  : public ::cry::lambda::enable_sig
{
  // Result type calculation.
  template<class FArgs>
  struct result;

  template<class UnaryFunction, class SinglePassRange>
  struct result< transformed_ const(SinglePassRange, UnaryFunction) >
  {
    typedef transformed_range<
        typename ::cry::plain<UnaryFunction>::type
      , typename ::boost::remove_reference<SinglePassRange>::type
      > type;
  };

  template<class UnaryFunction, class SinglePassRange>
  struct result< transformed_(SinglePassRange, UnaryFunction) >
    : public result< transformed_ const(SinglePassRange, UnaryFunction) >
  {};
  
  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  struct result<
    transformed_ const(SinglePassRange0, SinglePassRange1, BinaryFunction)
  >
  {
    typedef binary_transformed_range<
        typename ::cry::plain<BinaryFunction>::type
      , typename ::boost::remove_reference<SinglePassRange0>::type
      , typename ::boost::remove_reference<SinglePassRange1>::type
      > type;
  };

  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  struct result<
    transformed_(SinglePassRange0, SinglePassRange1, BinaryFunction)
  >
    : public result<
        transformed_ const(SinglePassRange0, SinglePassRange1, BinaryFunction)
      >
  {};



  // operator() for unary version.
  template<class UnaryFunction, class SinglePassRange>
  transformed_range<UnaryFunction, SinglePassRange>
  operator()(SinglePassRange &r, UnaryFunction pred) const
  {
    return transformed_range<UnaryFunction, SinglePassRange>(r, pred);
  }

  template<class UnaryFunction, class SinglePassRange>
  transformed_range<
    UnaryFunction, typename ::cry::add_const<SinglePassRange>::type
  >
  operator()(SinglePassRange const &r, UnaryFunction pred) const
  {
    return transformed_range<
        UnaryFunction, typename ::cry::add_const<SinglePassRange>::type
      >(r, pred);
  }



  // operator() for binary version.
  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  binary_transformed_range<BinaryFunction, SinglePassRange0, SinglePassRange1>
  operator()(
    SinglePassRange0 &r0, SinglePassRange1 &r1, BinaryFunction pred
  ) const
  {
    return binary_transformed_range<
        BinaryFunction, SinglePassRange0, SinglePassRange1
      >(r0, r1, pred);
  }

  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  binary_transformed_range<
    BinaryFunction
  , SinglePassRange0, typename ::cry::add_const<SinglePassRange1>::type
  >
  operator()(
    SinglePassRange0 &r0, SinglePassRange1 const &r1, BinaryFunction pred
  ) const
  {
    return binary_transformed_range<
        BinaryFunction
      , SinglePassRange0, typename ::cry::add_const<SinglePassRange1>::type
      >(r0, r1, pred);
  }

  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  binary_transformed_range<
    BinaryFunction
  , typename ::cry::add_const<SinglePassRange0>::type, SinglePassRange1
  >
  operator()(
    SinglePassRange0 const &r0, SinglePassRange1 &r1, BinaryFunction pred
  ) const
  {
    return binary_transformed_range<
        BinaryFunction
      , typename ::cry::add_const<SinglePassRange0>::type, SinglePassRange1
      >(r0, r1, pred);
  }

  template<class BinaryFunction, class SinglePassRange0, class SinglePassRange1>
  binary_transformed_range<
    BinaryFunction
  , typename ::cry::add_const<SinglePassRange0>::type
  , typename ::cry::add_const<SinglePassRange1>::type
  >
  operator()(
    SinglePassRange0 const &r0, SinglePassRange1 const &r1, BinaryFunction pred
  ) const
  {
    return binary_transformed_range<
        BinaryFunction
      , typename ::cry::add_const<SinglePassRange0>::type
      , typename ::cry::add_const<SinglePassRange1>::type
      >(r0, r1, pred);
  }
}; // struct transformed_

transformed_ const transformed = transformed_();

}} // namespace cry::range


#endif // !defined(CRY_RANGE_TRANSFORMED_RANGE_HPP_INCLUDE_GUARD)
