// (C) Copyright Cryolite 2004, 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// CONSIDER:  What is a "top level cv-qualification?"
//            In C++ type system, there are some kinds of types, which looks to
//           be cv-qualified, but in fact cannot be qualified directly in
//           generic context.
//            For example, consider the function type. Cv-qualified function
//           types can be defined in typedef. But it is not possible to
//           cv-qualify a template parameter which is substituted by a function
//           type.
//
//             typedef void const_type() const; // This is O.K.
//             
//             template<class T>
//             void f()
//             {
//               typedef T const type; // `const' does not cause any effect on
//                                     // `T' when `T' is a function type.
//             }
//
//             void f<void ()>();
//
//           In the above code, `T' is instantiated as `void ()', but
//           qualification for T does not cause any effect. So `type' is still
//           same to `void ()', not `void () const.' Then, we need some
//           meta-functions which manipulate
//           "pseudo-top-level cv-qualifications" of function types.
//
//             template<class T>
//             void f()
//             {
//               // Add a const qualification even if `T' is a function type.
//               typedef typename add_const<T>::type type;
//             }
//
//           Unfortunately, `boost::add_const' is not enough because it
//           manipulates only top level ones. We need other new traits.
//            For another example, some implementations cannot treat
//           "array to const type" as top-level qualified const.
//
//             template<class T>
//             void f(T &);       // (A)
//
//             template<class T>
//             void f(T const &); // (B)
//
//             int cosnt a[4] = {0,1,2,3};
//
//             f(a); // MSVC calls (A). GCC calls(B).
//
//           What is worse is that MSVC creats an invalid type when
//           "an array to const type" is qualified.
//
//             template<class T>
//             typename test<T const>::type f(T const &);
//
//             int const a[4] = {0,1,2,3};
//
//             f(a);
//
//           In the above code, `T const' is NOT VALID AT ALL in MSVC.
//           So I need some meta-functions as work-arounds for such a case,
//           which can treat "array to const type" as a top level qualified type
//           in all the implementations.
//            Then, some meta-functions are defined in cry::type_traits.
//
//              - is_const
//              - is_volatile
//              - add_const
//              - add_volatile
//              - add_cv
//              - add_const_reference
//              - add_volatile_reference
//              - add_cv_reference
//              - TODO: remove_const
//              - TODO: remove_volatile
//              - TODO: remove_cv
//
//           They can evaluate and manipulate "pseudo-top-level qualifications."
//           Here, the "pseudo-top-level qualification" can be defined as below;
//
//             - top level qualifications, or
//             - cv-qualifications for function types, or
//             - cv-qualifications for the value type of arrays, or
//             - CONSIDER: other possibilities?
//           
// TODO: Specializations for function types.
// CONSIDER: Other possibilities of "pseado-top-level qualification?"

#if !defined(CRY_TYPE_TRAITS_IS_CONST_HPP_INCLUDE_GUARD)
#define CRY_TYPE_TRAITS_IS_CONST_HPP_INCLUDE_GUARD

#include <boost/mpl/bool.hpp>
#include <boost/type_traits/is_const.hpp>


namespace cry{

template<class T>
struct is_const
  : public boost::is_const<T>
{};

template<class T, int sz>
struct is_const<T const [sz]>
  : public boost::mpl::true_
{};

} // namespace cry


#endif // !defined(CRY_TYPE_TRAITS_IS_CONST_HPP_INCLUDE_GUARD)
