// (C) Copyright Cryolite 2004, 2005.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//  This header provides a replacement for `std::vector'. The class `vector'
// defined here has complete compatibility with `std::vector'. Main advantages
// of this vector class over `std::vector' are;
//
//   1. it has move semantics, and
//   2. it is "move-aware." That is, this vector treats movable types
//      efficiently and safely, and
//   3. it can interact with models of Range.

#if !defined(CRY_VECTOR_HPP_INCLUDE_GUARD)
#define CRY_VECTOR_HPP_INCLUDE_GUARD

#include <iterator>
#include <algorithm>
#include <stdexcept>

#include <boost/mpl/if.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>
#include <boost/type_traits/is_integral.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/is_convertible.hpp>

#include <boost/utility/enable_if.hpp>

#include <boost/move.hpp> // TODO: Replace with `cry::move'.

#include <boost/iterator/iterator_categories.hpp>
#include <boost/iterator/iterator_traits.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include <boost/iterator/reverse_iterator.hpp>

#include <boost/tuple/tuple.hpp>
#include <boost/compressed_pair.hpp>

#include <cry/utility/enable_if_same.hpp>
#include <cry/utility/enable_if_convertible.hpp>
#include <cry/scope_guard.hpp>
#include <cry/memory.hpp>
#include <cry/utility.hpp>
#include <cry/move/enable_if_movable.hpp>
#include <cry/move_ptr.hpp>
#include <cry/deleter/deallocator.hpp>
#include <cry/container/container_traits.hpp>
#include <cry/range/value_type.hpp>
#include <cry/range/insert_to_container.hpp>
#include <cry/lambda/bind.hpp> // CONSIDER: This looks too heavy.

#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable:4675) // `swap' is found via ADL. This is by design.
#endif


namespace cry{

template<class T, class Allocator = cry::allocator<T> >
class vector
{
  template<class AnotherValue, class AnotherAllocator>
  friend class vector;
  typedef deallocator<Allocator> deallocator_type;
  typedef static_move_ptr<T[], deallocator_type> buffer_type;
  
public:
  typedef T value_type;
  typedef Allocator allocator_type;
  typedef typename Allocator::reference reference;
  typedef typename Allocator::const_reference const_reference;
  typedef typename Allocator::size_type size_type;
  typedef typename Allocator::difference_type difference_type;
  typedef typename Allocator::pointer pointer;
  typedef typename Allocator::const_pointer const_pointer;

  class const_iterator; // Forward declaration
  
  // The definitions of both `iterator' and `const_iterator' can be unified by
  // a single class template. However, such a implementation makes error message
  // more complicated. So I decide to provide explicitly separated definitions.
  class iterator
    : public boost::iterator_facade<
        iterator
      , value_type
      , boost::random_access_traversal_tag
      , reference
      , difference_type
      >
  {
    // All the member functions never throw.
    friend class boost::iterator_core_access;
    friend class const_iterator;
    typedef vector enclosing;

  public:
    typedef typename enclosing::reference reference;
    typedef typename enclosing::difference_type difference_type;
    
    iterator()
      : p_(0)
    { }

    iterator(value_type *p)
      : p_(p)
    { }

    iterator(iterator const &r)
      : p_(r.p_)
    {}

    iterator &operator=(iterator const &rhs)
    {
      p_ = rhs.p_;
      return *this;
    }
    
  private:
    reference dereference() const
    {
      return *p_;
    }

    bool equal(iterator rhs) const
    {
      return p_ == rhs.p_;
    }

    bool equal(const_iterator rhs) const
    {
      return p_ == rhs.p_;
    }
    
    void increment()
    {
      ++p_;
    }

    void decrement()
    {
      --p_;
    }

    void advance(difference_type n)
    {
      p_ += n;
    }

    difference_type distance_to(iterator rhs) const
    {
      return rhs.p_ - p_;
    }

    difference_type distance_to(const_iterator rhs) const
    {
      return rhs.p_ - p_;
    }

    value_type *p_;
  }; // class iterator

  class const_iterator
    : public boost::iterator_facade<
        const_iterator
      , value_type
      , boost::random_access_traversal_tag
      , const_reference
      , difference_type
      >
  {
    // All the member functions never throw.
    friend class boost::iterator_core_access;
    typedef vector enclosing;

  public:
    typedef typename enclosing::const_reference reference;
    typedef typename enclosing::difference_type difference_type;
    
    const_iterator()
      : p_(0)
    { }

    const_iterator(value_type const *p)
      : p_(p)
    { }
    
    // Constructor from a mutable iterator.
    const_iterator(iterator it)
      : p_(it.p_)
    { }

    const_iterator(const_iterator const &r)
      : p_(r.p_)
    {}

    const_iterator &operator=(const_iterator const &rhs)
    {
      p_ = rhs.p_;
      return *this;
    }
    
  private:
    reference dereference() const
    {
      return *p_;
    }

    bool equal(const_iterator rhs) const
    {
      return p_ == rhs.p_;
    }

    bool equal(iterator rhs) const
    {
      return p_ == rhs.p_;
    }
    
    void increment()
    {
      ++p_;
    }

    void decrement()
    {
      --p_;
    }

    void advance(difference_type n)
    {
      p_ += n;
    }

    difference_type distance_to(const_iterator rhs) const
    {
      return rhs.p_ - p_;
    }

    difference_type distance_to(iterator rhs) const
    {
      return rhs.p_ - p_;
    }

    value_type const *p_;
  }; // class const_iterator

  typedef boost::reverse_iterator<iterator> reverse_iterator;
  typedef boost::reverse_iterator<const_iterator> const_reverse_iterator;

  explicit vector(allocator_type const &a = allocator_type()) // throw()
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, a)
  { }

  // Strong gurantee.
  explicit vector(
    size_type n, value_type const &value = value_type()
  , allocator_type const &a = allocator_type()
  )
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, a)
  {
    insert(end(), n, value);
  }

  // In some cases, this constructor wrongly deduce the arguments for
  // non-iterator types. For example,
  //
  //   vector<int> u(1, 2);
  //
  // this syntax obviously intends to call
  // `vector(size_type, value_type const &)', but in fact it calls this
  // constructor because of overload resolution priolity. See also (23.1.1/9),
  // (23.1.1/10) and (23.1.1/11).
  //
  // Strong guarantee.
  template<class InputIterator>
  vector(
    InputIterator first, InputIterator last
  , allocator_type const &a = allocator_type()
  )
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, a)
  {
    insert(end(), first, last);
  }

  // Strong guarantee.
  template<class SinglePassRange>
  explicit vector(
    SinglePassRange const &r, allocator_type const &a = allocator_type()
  , typename cry::disable_if_same<SinglePassRange, vector>::type * = 0
  )
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, a)
  {
    cry::range::insert_to_container(r, *this, begin());
  }
  
  // Copy constructor from a non-const lvalue.
  // Strong guarantee.
  vector(vector &r)
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, r.allocator_reference())
  {
    insert(end(), r.begin(), r.end());
  }

  // Copy constructor from a const lvalue.
  // Strong guarantee.
  template<class ConstType>
  vector(
    ConstType &r
  , typename enable_if_same<ConstType, vector const>::type * = 0
  )
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, r.allocator_reference())
  {
    insert(end(), r.begin(), r.end());
  }
  
  // Move constructor.
  vector(boost::move_from<vector> src) // throw()
    : p_(0, deallocator_type(0))
    , size_and_allocator_(0, src->allocator_reference())
  {
    using boost::move;
    
    p_ = move(src->p_);
    size_reference() = src->size();

    src->size_reference() = 0;
  }

  // Destructor
  ~vector() // throw()
  {
    clear();
  }
  
  operator boost::move_from<vector>() // throw()
  {
    return boost::move_from<vector>(this);
  }

  operator boost::const_lvalue<vector>() const // throw()
  {
    return boost::const_lvalue<vector>(this);
  }
  
  // Assignment from a non-const lvalue.
  // Strong guarantee.
  vector &operator=(vector &rhs)
  {
    *this = cry::as_const(rhs);
    return *this;
  }

  // Assignment from a const lvalue.
  // Strong guarantee.
  template<class ConstType>
  typename enable_if_same<ConstType, vector const, vector &>::type
  operator=(ConstType &rhs)
  {
    vector tmp(rhs);
    swap(tmp);
    return *this;
  }

private:
  // `value_type' is non-movable.
  void move_assignment_dispatch(
    boost::move_from<vector> src, boost::mpl::false_ is_movable
  )
  {
    vector tmp(src);
    tmp.swap(*this);
  }

  // `value_type' is movable.
  void move_assignment_dispatch(
    boost::move_from<vector> src, boost::mpl::true_ is_movable
  )
  {
    // When `value_type' is movable and `capacity() >= rhs.size()', elements of
    // `rhs' can be moved and an additional allocation is no longer needed.
    if(capacity() >= src->size()){
      clear();
      uninitialized_move(
          src->begin(), src->end(), p_.get(), allocator_reference()
        );
      size_reference() = src->size();
    }
    else{
      vector tmp(src);
      tmp.swap(*this);
    }
  }

public:
  // Move assignment.
  //
  //   -No-fail if `value_type' is movable and `capacity() >= src->size()',
  //    otherwise
  //   -strong guarantee.
  vector &operator=(boost::move_from<vector> src)
  {
    move_assignment_dispatch(
        src
      , typename boost::is_movable<value_type>::type()
      );
    return *this;
  }

  // Strong gurantee.
  //
  // Note: Read range constructor's note.
  // 
  // CONSIDER: When this assignment is move operation, an optimization as in
  //           the copy assignment would be applied.
  template<class SPTravReadIter>
  void assign(SPTravReadIter first, SPTravReadIter last)
  {
    vector tmp(first, last);
    swap(tmp);
  }

  // Strong gurantee.
  void assign(size_type n, value_type const &u)
  {
    vector tmp(n, u);
    swap(tmp);
  }

  // Assignment from a range. This is safe even for the case where `rhs' is
  // dependent on `*this'.
  // Strong guarantee.
  template<class SinglePassRange>
  typename cry::disable_if_same<SinglePassRange, vector, vector &>::type
  operator=(SinglePassRange const &r)
  {
    vector tmp(r);
    tmp.swap(*this);
    return *this;
  }

  allocator_type get_allocator() const // throw()
  {
    return allocator_reference();
  }

  iterator begin() // throw()
  {
    return iterator(p_.get());
  }

  const_iterator begin() const // throw()
  {
    return const_iterator(p_.get());
  }

  iterator end() // throw()
  {
    return iterator(p_.get() + size());
  }

  const_iterator end() const // throw()
  {
    return const_iterator(p_.get() + size());
  }

  reverse_iterator rbegin() // throw()
  {
    return reverse_iterator(end());
  }

  const_reverse_iterator rbegin() const // throw()
  {
    return const_reverse_iterator(end());
  }

  reverse_iterator rend() // throw()
  {
    return reverse_iterator(begin());
  }

  const_reverse_iterator rend() const // throw()
  {
    return const_reverse_iterator(begin());
  }

  size_type size() const // throw()
  {
    return size_reference();
  }

  size_type max_size() const // throw()
  {
    return allocator_reference().max_size();
  }

  // Strong guarantee
  void resize(size_type sz, value_type const &c = value_type())
  {
    if(sz < size()){
      erase(begin() + sz, end());
    }
    else if(sz > size()){
      insert(end(), sz - size(), c);
    }
  }

  size_type capacity() const // throw()
  {
    return p_.get_deleter().size();
  }

  bool empty() const // throw()
  {
    return size() == 0;
  }

  // Strong guarantee.
  void reserve(size_type n)
  {
    using boost::move;
    
    if(n > capacity()){
      size_type old_size = size();
      buffer_type new_buffer = get_grown_buffer(n);
      uninitialized_move(
          begin(), end(), new_buffer.get(), allocator_reference()
        );

      // commit
      clear(); // size() == 0
      p_ = move(new_buffer);
      size_reference() = old_size;
    }
  }

  reference operator[](size_type n) // throw()
  {
    return p_[n];
  }

  const_reference operator[](size_type n) const // throw()
  {
    return p_[n];
  }

  // Strong guarantee.
  reference at(size_type n)
  {
    if(n >= size()){
      throw std::out_of_range("range error in `cry::vector::at'");
    }
    return p_[n];
  }

  // Strong guarantee.
  const_reference at(size_type n) const
  {
    if(n >= size()){
      throw std::out_of_range("range error in `cry::vector::at'");
    }
    return p_[n];
  }

  reference front() // throw()
  {
    return *begin();
  }

  const_reference front() const // throw()
  {
    return *begin();
  }

  reference back() // throw()
  {
    return *--end();
  }

  const_reference back() const // throw()
  {
    return *--end();
  }

private:
  template<class U>
  void do_push_back(U &x)
  {
    // CONSIDER: Code bloat. It is possible to separate the part independent
    // from `U'.
    
    // `U' is one of followings.
    //   - `value_type'
    //   - `boost::move_from<value_type>'
    //   - a type convertible to `value_type'

    using boost::move;
    
    if(capacity() > size()){
      // If `U' is `boost::move_from<value_type>', use move. Otherwise use copy.
      allocator_reference().construct(p_.get() + size(), x);

      // commit
      ++size_reference();
    }
    else{
      buffer_type new_buffer = get_grown_buffer(size() + 1);
      size_type new_size = size() + 1;
      
      pointer pos = uninitialized_move(
          begin(), end(), new_buffer.get(), allocator_reference()
        ); 
      // If `value_type' is movable, This scope guard moves the values in new
      // buffers to old buffer and destroy values in new buffer. Otherwise it
      // only destroy values in new buffer.
      scope_guard sg = make_guard(
          boost::lambda::bind(
            uninitialized_move_rollback
          , new_buffer.get(), pos, begin(), allocator_reference()
          )
        );
    
      // If `U' is `boost::move_from<value_type>', use move. Otherwise use copy.
      allocator_reference().construct(pos, x);

      // commit
      clear();
      p_ = move(new_buffer);
      size_reference() = new_size;
      sg.dismiss();
    }
  }
  
public:
  // push_back
  // Strong guarantee.

  // For non-movable types, except non-const lvalue.
  template<class U>
  void push_back(
    U const &x
  , typename cry::disable_if_movable<value_type, U>::type * = 0
  , typename cry::enable_if_convertible<U, value_type>::type * = 0
  )
  {
    do_push_back(x);
  }

  // For non-const lvalue.
  void push_back(value_type &x)
  {
    do_push_back(x);
  }
  
  // For const lvalue.
  template<class U>
  void push_back(
    U &x
  , typename cry::enable_if_movable<value_type, U>::type * = 0
  , typename cry::enable_if_same<U, value_type const>::type * = 0
  )
  {
    do_push_back(x);
  }

  // Via conversion from different type to movable type.
  template<class U>
  void push_back(
    U const &x
  , typename cry::enable_if_movable<value_type, U>::type * = 0
  , typename cry::disable_if_same<U, value_type>::type * = 0
  //, typename enable_if_convertible<U, value_type>::type * = 0
  )
  {
    do_push_back(x);
  }
  
  // For move_from or rvalue of movable type.
  void push_back(boost::move_from<value_type> x)
  {
    do_push_back(x);
  }

  void pop_back() // throw()
  {
    allocator_reference().destroy(&*(--end()));
    --size_reference();
  }

private:
  template<class U>
  void do_insert_impl(iterator position, U &x)
  {
    // CONSIDER: Code bloat. It is possible to separate the part independent
    // from the template argument `U'.
    
    // `U' is one of followings.
    //   - `value_type'
    //   - `boost::move_from<value_type>'
    //   - a type convertible to `value_type'
    
    using boost::move;

    if(capacity() > size()){
      if(position == end()){
        // If `U' is movable, use move. Otherwise use copy.
        allocator_reference().construct(&*position, x);

        // Commit
        ++size_reference();
      }
      else{
        pointer pos = p_.get() + size();
        
        // If `U' is movable, use move. Otherwise use copy.
        allocator_reference().construct(pos, move(back()));
        // If `value_type' is movable, this scope guard moves `pos' to `pos - 1'
        // and destroy `pos'. Otherwise it only destroys `pos'.
        scope_guard sg0 = make_guard(
            boost::lambda::bind(
              uninitialized_move_rollback, pos, pos - 1, allocator_reference()
            )
          );
        
        // If `value_type' is movable, use move. Otherwise copy. If `value_type'
        // is not a movable type, this operation cannot be rolled back.
        boost::move_backward(&*position, pos - 1, pos);
        // If `value_type' is a movable type, this scope guard moves
        // [position + 1, pos) to [position, pos - 1). Otherwise it does not do
        // anything.
        scope_guard sg1 = make_guard(
            boost::lambda::bind(move_rollback, &*position + 1, pos, position)
          );
        
        // If `U' is `boost::move_from<value_type>', use move. Otherwise copy.
        *position = x;

        // Commit.
        ++size_reference();
        sg0.dismiss();
        sg1.dismiss();
      }
    }
    else{
      buffer_type new_buffer = get_grown_buffer(size() + 1);
      size_type new_size = size() + 1;

      pointer pos = uninitialized_move(
          begin(), position, new_buffer.get(), allocator_reference()
        );
      scope_guard sg0 = make_guard(
          boost::lambda::bind(
            uninitialized_move_rollback
          , new_buffer.get(), pos, begin(), allocator_reference()
          )
        );
      
      allocator_reference().construct(pos, x);
      scope_guard sg1 = make_guard(
          boost::lambda::bind(destroy, pos, allocator_reference())
        ); // TODO: Is this O.K.? Or move_rollback should be used?
      ++pos;
      
      uninitialized_move(position, end(), pos, allocator_reference());

      // commit
      clear();
      p_ = move(new_buffer);
      size_reference() = new_size;
      sg0.dismiss();
      sg1.dismiss();
    }
  }

  template<class U>
  iterator do_insert(iterator position, U &x)
  {
    // "do_insert_dispatch" invalidates the iterator, but `offset' is still
    // stable.
    difference_type offset = position - begin();
    do_insert_impl(position, x);
    return begin() + offset;
  }
  
public:
  // insert (single value version)
  //   
  //   - strong guarantee if `value_type' is movable, otherwise
  //   - strong guarantee if position == end(), otherwise
  //   - basic guarantee.

  // For non-movable type, except non-const lvalue.
  template<class ValueType>
  iterator insert(
    iterator position, ValueType const &x
  , typename disable_if_movable<value_type, ValueType>::type * = 0
  , typename enable_if_convertible<ValueType, value_type>::type * = 0
  )
  {
    return do_insert(position, x);
  }
  
  // For non-const lvalue.
  iterator insert(iterator position, value_type &x)
  {
    return do_insert(position, x);
  }

  // For const lvalue.
  template<class U>
  iterator
  insert(
    iterator position, U &x
  , typename enable_if_movable<value_type, U>::type * = 0
  , typename enable_if_same<U, value_type const>::type * = 0
  )
  {
    return do_insert(position, x);
  }

  // Via conversion from different type rvalue to movable type.
  template<class U>
  iterator
  insert(
    iterator position, U const &x
  , typename enable_if_movable<value_type, U>::type * = 0
  , typename disable_if_same<U, value_type>::type * = 0
  )
  {
    return insert(position, x);
  }
  
  // For rvalue or move source of movable type.
  iterator insert(iterator position, boost::move_from<value_type> x)
  {
    return do_insert(position, x);
  }

private:
  void insert_n(iterator position, size_type n, value_type const &x)
  {
    using boost::move;
    
    // GCC warns the comparison of unsigned values. `as_signed' puts it down.
    if(as_signed(capacity()) >= as_signed(size() + n)){
      if(as_signed(n) > end() - position){
        pointer p = uninitialized_move(
            position, end(), &*position + n, allocator_reference()
          );
        scope_guard sg0 = make_guard(
            boost::lambda::bind(
              uninitialized_move_rollback
            , &*position + n, p, position, allocator_reference()
            )
          );
        
        uninitialized_fill(
            &*end(), &*position + n, x, allocator_reference()
          );
        scope_guard sg1 = make_guard(
            boost::lambda::bind(
              destroy, &*end(), &*position + n, allocator_reference()
            )
          );
        
        std::fill(position, end(), x);

        // commit
        size_reference() += n;
        sg0.dismiss();
        sg1.dismiss();
      }
      else{
        pointer p = uninitialized_move(
            end() - n, end(), &*end(), allocator_reference()
          );
        scope_guard sg0 = make_guard(
            boost::lambda::bind(
              uninitialized_move_rollback
            , &*end(), p, end() - n, allocator_reference()
            )
          );
        
        iterator pos = move_backward(position, end() - n, end());
        scope_guard sg1 = make_guard(
            boost::lambda::bind(move_rollback, pos, end(), position)
          );
        
        std::fill(position, position + n, x);

        // commit
        size_reference() += n;
        sg0.dismiss();
        sg1.dismiss();
      }
    }
    else{
      buffer_type new_buffer = get_grown_buffer(size() + n);
      size_type new_size = size() + n;
      
      pointer p = uninitialized_move(
          begin(), position, new_buffer.get(), allocator_reference()
        );
      scope_guard sg0 = make_guard(
          boost::lambda::bind(
            uninitialized_move_rollback
          , new_buffer.get(), p, begin(), allocator_reference()
          )
        );
      
      uninitialized_fill(p, p + n, x, allocator_reference());
      scope_guard sg1 = make_guard(
          boost::lambda::bind(destroy, p, p + n, allocator_reference())
        );
      p += n;

      uninitialized_move(position, end(), p, allocator_reference());

      // commit
      clear();
      p_ = move(new_buffer);
      size_reference() = new_size;
      sg0.dismiss();
      sg1.dismiss();
    }
  }

public:
  // insert (filling version)
  // 
  //   - strong guarantee if `value_type' is movable, otherwise
  //   - strong guarantee if position == end(), otherwise
  //   - basic guarantee.
  void insert(iterator position, size_type n, value_type const &x)
  {
    insert_n(position, n, x);
  }
  
private:
  // For SinglePassTraversalIterator.
  template<class SPTravReadIterator>
  void insert_range_dispatch(
    iterator position, SPTravReadIterator first, SPTravReadIterator last
  , boost::single_pass_traversal_tag
  )
  {
    if(position == end()){
      std::copy(first, last, std::back_inserter(*this));
    }
    else{
      std::copy(first, last, std::inserter(*this, position));
    }
  }

  // For ForwardTraversalIterator.
  template<class FwdTrvRdIterator>
  void insert_range_dispatch(
    iterator position, FwdTrvRdIterator first, FwdTrvRdIterator last
  , boost::forward_traversal_tag
  )
  {
    using boost::move;
    
    difference_type n
      = static_cast<difference_type>(std::distance(first, last));
    if(capacity() >= size() + n){
      if(n > end() - position){
        pointer p = uninitialized_move(
            position, end(), &*position + n, allocator_reference()
          );
        scope_guard sg0 = make_guard(
            boost::lambda::bind(
              uninitialized_move_rollback
            , &*position + n, p, position, allocator_reference()
            )
          );

        p = uninitialized_copy(
            first + (end() - position), last, &*end(), allocator_reference()
          );
        scope_guard sg1 = make_guard(
            boost::lambda::bind(destroy, &*end(), p, allocator_reference())
          );
        
        std::copy(first, first + (end() - position), position);
        
        // commit
        size_reference() += n;
        sg0.dismiss();
        sg1.dismiss();
      }
      else{
        pointer p = uninitialized_move(
            end() - n, end(), &*end(), allocator_reference()
          );
        scope_guard sg0 = make_guard(
            boost::lambda::bind(
              uninitialized_move_rollback
            , &*end(), p, end() - n, allocator_reference()
            )
          );
        
        iterator pos = move_backward(position, end() - n, end());
        scope_guard sg1 = make_guard(
            boost::lambda::bind(move_rollback, pos, end(), position)
          );
        
        std::copy(first, last, position);

        // commit
        size_reference() += n;
        sg0.dismiss();
        sg1.dismiss();
      }
    }
    else{
      buffer_type new_buffer = get_grown_buffer(size() + n);
      size_type new_size = size() + n;
      
      pointer p = uninitialized_move(
          begin(), position, new_buffer.get(), allocator_reference()
        );
      scope_guard sg0 = make_guard(
          boost::lambda::bind(
            uninitialized_move_rollback
          , new_buffer.get(), p, begin(), allocator_reference()
          )
        );

      uninitialized_copy(first, last, p, allocator_reference());
      scope_guard sg1 = make_guard(
          boost::lambda::bind(destroy, p, p + n, allocator_reference())
        );
      p += n;

      uninitialized_move(position, end(), p, allocator_reference());

      // commit
      clear();
      p_ = move(new_buffer);
      size_reference() = new_size;
      sg0.dismiss();
      sg1.dismiss();
    }
  }

  // Sure it is iterator. Dispatch to iterator range version.
  template<class SPTravReadIterator>
  void insert_corrected(
    iterator position, SPTravReadIterator first, SPTravReadIterator last
  , typename boost::disable_if<
      boost::is_integral<SPTravReadIterator>
    >::type * = 0
  )
  {
    insert_range_dispatch(
        position, first, last
      , typename boost::iterator_traversal<SPTravReadIterator>::type()
      );
  }

  // It is an integer type. Dispatch to filling version.
  template<class Integer>
  void insert_corrected(
    iterator position, Integer n, Integer value
  , typename boost::enable_if<boost::is_integral<Integer> >::type * = 0
  )
  {
    insert_n(position, n, value);
  }

public:
  // If SPTravReadIterator is integral, it should be dispatched to `insert_n'.
  // See also the note of `vector(Iterator, Iterator, Allocator)'.
  //
  //   - strong guarantee if `value_type' is movable, otherwise
  //   - strong guarantee if position == end(), otherwise
  //   - basic guarantee.
  template<class SPTravReadIterator>
  void insert(
    iterator position, SPTravReadIterator first, SPTravReadIterator last
  )
  {
    // TODO: Dispatch for movable type to provide strong guarantee.
    //       (Has this been already done?)
    insert_corrected(position, first, last);
  }

  // Strong guarantee.
  // CONSIDER: Exception guarantee.
  template<class SinglePassRange>
  void insert(
    iterator position, SinglePassRange const &r
  , typename cry::disable_if_convertible<
      typename cry::range::value<SinglePassRange>::type, value_type
    >::type * = 0
  )
  {
    cry::range::insert_to_container(r, *this, position);
  }
  
  // - No-fail if `value_type' is movable, otherwise
  // - no-fail if `position == --end()', otherwise
  // - basic guarantee.
  iterator erase(iterator position)
  {
    difference_type offset = position - begin();
    boost::move(position + 1, end(), position);
    allocator_reference().destroy(&*end() - 1);
    --size_reference();

    return begin() + offset;
  }

  // - No-fail if `value_type' is movable, otherwise
  // - no-fail if `last == end()', otherwise
  // - basic guarantee.
  iterator erase(iterator first, iterator last)
  {
    difference_type offset = first - begin();
    difference_type n = last - first;
    iterator pos = boost::move(last, end(), first);
    destroy(pos, end(), allocator_reference());
    size_reference() -= n;

    return begin() + offset;
  }

  void swap(vector &x) // throw()
  {
    using std::swap;
    swap(p_, x.p_);
    // boost::compressed_pair's free-standing `swap' is ADL-aware.
    swap(size_and_allocator_, x.size_and_allocator_);
  }

  void clear() // throw()
  {
    destroy(begin(), end(), allocator_reference());
    size_reference() = 0;
  }

  friend
  bool operator==(vector const &lhs, vector const &rhs)
  {
    return lhs.size() == rhs.size()
      && std::equal(lhs.begin(), lhs.end(), rhs.begin());
  }
  
  friend
  bool operator<(vector const &lhs, vector const &rhs)
  {
    return std::lexicographical_compare(
        lhs.begin(), lhs.end(), rhs.begin(), rhs.end()
      );
  }
  
  friend
  bool operator!=(vector const &lhs, vector const &rhs)
  {
    return !(lhs == rhs);
  }
  
  friend
  bool operator>(vector const &lhs, vector const &rhs)
  {
    return rhs < lhs;
  }
  
  friend
  bool operator>=(vector const &lhs, vector const &rhs)
  {
    return !(lhs < rhs);
  }
  
  friend
  bool operator<=(vector const &lhs, vector const &rhs)
  {
    return !(lhs > rhs);
  }
  
  friend
  void swap(vector &lhs, vector &rhs) // throw()
  {
    lhs.swap(rhs);
  }
  
private:
  buffer_type get_grown_buffer(size_type n) const
  {
    // Exponentially grown with growth coefficient of 1.5.
    size_type new_capacity = capacity() + capacity() / 2;
    new_capacity = std::max(new_capacity, std::max(n, capacity() + 1));

    return buffer_type(
        get_allocator().allocate(new_capacity) // allocator_reference() cannot
                                               // be used, because allocate
                                               // member function is non-const.
      , deallocator_type(new_capacity)
      );
  }

  size_type &size_reference() // throw()
  {
    return size_and_allocator_.first();
  }

  size_type const &size_reference() const // throw()
  {
    return  size_and_allocator_.first();
  }
  
  allocator_type &allocator_reference() // throw()
  {
    return size_and_allocator_.second();
  }

  allocator_type const &allocator_reference() const // throw()
  {
    return size_and_allocator_.second();
  }
  
private:
  buffer_type p_;
  boost::compressed_pair<size_type, allocator_type> size_and_allocator_;
}; // class vector



namespace container{

template<class T, class Allocator>
struct container_traits<vector<T, Allocator> >
  : public cry::container::container_traits<>
{
  typedef boost::mpl::true_                  is_specialized;
  typedef back_insertable_sequential_tag     container_category;
  typedef boost::random_access_traversal_tag traversal_tag;
  typedef unstable_iterator_tag              iterator_stability;
  typedef boost::mpl::true_                  has_reserve;
  typedef boost::mpl::true_                  has_capacity;
};

} // namespace container

} // namespace cry


#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // !defined(CRY_VECTOR_HPP_INCLUDE_GUARD)
